/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 *
 *  @file ScaFES.hpp
 *
 *  @brief Main include file for applications using 'ScaFES'.
 *
 *  ScaFES is an abbreviation for 'Scalable Framework for Explicit Solvers'.
 *
 *  \n The framework provides interfaces for solving time-dependent problems
 *  like the Heat equation or the Maxwell equations.
 *
 *  \n The numerical solutions of such problems can be (parallely) written 
 *  to output files via an NetCDF interface and afterwards graphically 
 *  displayed via 'VisIt'.
 *
 *  \n Via the PAPI interface, it is possible to count hardware events. 
 */

/**
 *  \mainpage ScaFES Documentation
 *
 *  \section introduction Introduction
 *
 *  \section Problems Problems
 *
 *  Applications are controlled via command line parameters.
 * <ol>
 * <li>
 * Decompose the MPI domain.
 *  Grid partitions are created and stored in the file indicated.
 *  The decomposition of the domain uses an algorithm that delivers minimal
 *  cutting faces. If one do not want to cut in a certain dimension, one has
 *  to disable the cutting via the variable
 *  \code Index divideGrid(a,b,c) \endcode
 * with \code a,b,c \in {0,1}'. \endcode
 *  The value 0 disables cutting in this direction.
 * The numerical grid can be given via the command line parameter
 *  '--nNodes=width x depth x height'.
 * This will create a grid with w*d*h nodes.
 * \remarks
 * Until now: The domain decomposition works only sequential!
 * </li>
 *
 * </li>
 *  Compute the solution of a problem based on the partition in 1)
 *  It is assumed that the domain is already decomposed and that its name is
 *  declared via the command line parameter '--partfile'.
 * </li>
 *
 *
 * \subsection Implementation How to implement a new problem in ScaFES:
 * <ul>
 * <li> Create a new class and derive it from the class \c Problem. </li>
 * <li> Add data fields to the problem constructor via the method
 *      \c Problem::addDataField(). </li>
 *      The data fields will has to be initialized at all grid points
 *      via a given function. 
 *
 * <li> Implement the virtual method "Problem::computeTimestep()"
 *      in the derived class.
 *      Compute DF_new as a function of DF_old and other fields
 *      and parameters. </li>
 * <li> Synchronize values at all ghost nodes. </li>
 * <li> Change the meaning of DF_new and DF_old. </li>
 * </ul>
 * 
 * \remarks The master MPI process reads in the command line arguments.
 *       
 * \subsection Discretization Discretization in time and space:
 * Grid size        \code h_l := (b_p-a_p)/(nNodes_p-1) for p=0,1,2. \endcode
 * Grid coordinates \code x_i := i*h_0 for i=0,1,2,...,nNodes_0-1. \endcode
 * Grid coordinates \code y_j := j*h_1 for j=0,1,2,...,nNodes_1-1. \endcode
 * Grid coordinates \code z_k := k*h_2 for k=0,1,2,...,nNodes_2-1. \endcode
 * Time step size   \code tau := (tend-tstart)/nTimesteps. \endcode
 * Time coordinate  \code t_l := l*tau for l=0,1,2,...,nTimesteps. \endcode
 *
 * Example:
 * \code
 * Function    f_0(x,y,z) := x/h_0 + y/h_1 + z/h_2 
 * ==> f_0(x_i,y_j,z_k) = x_i/h_0 + y_j/h_1 + z_k/h_2 
 *                    = i*h_0/h_0 + j*h_1/h_1 + k*h_2/h_2 = i + j + k.
 *
 * Function    fana(x,y,z,t) := x/hx + y/hy + z/hz + 2*t/tau
 * ==> fana(x_i,y_j,z_k,t_l) = x_i/h_0 + y_j/h_1 + z_k/h_2 + 2*t_l/tau
 *                           = i*h_0/h_0 + j*h_1/h_1 + k*h_2/h_2 + 2*l*tau/tau 
 *                           = i + j + k + 2*l.
 * \endcode
 *
 * 
 * \n Design of test classes:
 * If there exists an ordering, then the order of the objects is as follows:
 * NULL < A < scalNonZero < B.
 * Less < Greater
 * Undetermined cannot be orderd between Less and Greater.
 *
 * REMARK: Many tests of this class are based on the getter method 'elem()'
 *         'EXPECT_TRUE(this->elemV0 == TripX.elem(0));'
 *
 * Choose characteristic values of equivalence classes
 * (so called 'representative').
 * Positive Tests (with valid inputs)    = Correctness
 * Negative Tests (with non-valid input) = Robustness
 *
 * Test order:
 * 1) test getter methods,
 * 2) test equals and unequals operator (if implemented)
 *    - Is it suffcient to check the equals operator on the basis of the getter
 *      methods of all "independent" member variables of the class
 *      because the dependent methods will be set in an own methods?!
 *    - "independent member variables"
 *        = usually, all input parameters of the class constructor.
 *    - Test "operatorEqualsComponentwise" is not necessary because it
 *      contains the (nearly) same code as the method operator==().
 * 3) test life cycle methods.
 *
 * -Do not use default values for the test values.
 * -Set different numbers for each element of independent members, if possible.
 * -Compute values of dependent members by hand!
 *
 * Some abbreviations:
 * -------------------
 * TT = TRACETYPE = type for tracing with ADOL-C (usually "adouble").
 * CT = COMPTYPE  = type for computing (usually "double").
 ******************************************************************************/

#ifndef SCAFES_HPP_
#define SCAFES_HPP_

#include "ScaFES_Config.hpp"

#include "ScaFES_Problem.hpp"

#endif
