/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_BUFFERDUMMY_HPP_
#define SCAFES_BUFFERDUMMY_HPP_

#include "ScaFES_Config.hpp"

#include <vector>

#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
#include <boost/serialization/vector.hpp>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class \c Buffer provides buffers for sending and receiving values
 * from one grid partition to a neighboured one.
 * The values in the buffers are exchanged by using the method \c exchange.
 *
 * \n The Boost.MPI skeleton concept can be used in order to improve the
 * performance of the data exchange.
 *
 * \remarks
 * Boost.MPI cannot send and receive data of type std::complex<double>.
 * Thus, one should use the class ScaFES::Complex<double>.
 */
template<typename TT>
class Buffer
{
    public:
        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor. */
        Buffer();

        /** Creates own constructor. */
        Buffer(boost_mpi::communicator const& myWorld,
               int const& idNeighPartition,
               int const& sizeBuffer,
               bool const& useSkeletonConcept);

        /** Creates own copy constructor. */
        Buffer(Buffer<TT> const&);

        /** Creates own assignment operator. */
        Buffer& operator=(Buffer<TT> rhs);

        /** Creates own destructor. */
        ~Buffer();

        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the MPI communicator. */
        boost_mpi::communicator const& myWorld() const;

        /** Returns the identifier of the neighboured grid partition. */
        int const& idNeighPartition() const;

        /** Returns the number of the elements within the buffers. */
        int const& sizeBuffer() const;

        /** Returns if the Boost.MPI skeleton concept will be used or not. */
        bool const& useSkeletonConcept() const;

        /** Returns the buffer for storing the send values. */
        std::vector<TT> const& dataReceived() const;

        /** Returns the buffer for storing the receive values. */
        std::vector<TT> const& dataSent() const;

        /** Returns the value of the buffer at a given index. */
        TT const& elemData(int const& idx) const;

        /*----------------------------------------------------------------------
        | SETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Sets a given value to a given position in the send buffer.
         * \param memoryPos Given position.
         * \param elemData Given value.
         */
        void setElemData(unsigned long int const& memoryPos,
                         TT const& elemData);

        /*----------------------------------------------------------------------
        | COMPARISON METHODS.
        ----------------------------------------------------------------------*/
        /** Tests if two buffers are equal. */
        bool operator==(Buffer<TT> const& rhs) const;

        /** Tests if two buffers are not equal. */
        bool operator!=(Buffer<TT> const& rhs) const;

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Sends the values stored in the send buffer to the given neighbour
         * partition. */
        void sendValues();

        /** Receives the values of send buffer from the given neighbour */
        void receiveValues();

        /** Waits until all communication calls have bee finished. */
        void waitAll();

        /** Serializes this class. */
        template<class Archive>
        void serialize(Archive& ar, unsigned int const version);

        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Swaps two objecty by applying the copy-and-swap idiom. */
        template<typename SS>
        friend void swap(Buffer<SS>& first,
                         Buffer<SS>& second);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Boost.MPI communicator. */
        boost_mpi::communicator mMyWorld;

        /** Identifiers of the (neighboured) grid partitions
        * to which data are sent and received. */
        int mIdNeighPartition;

        /** Size of underlying grid of memory = Number of elements for buffer.*/
        int mSizeBuffer;

        /** Uses the Boost.MPI skeleton concept? */
        bool mUseSkeletonConcept;

        /** Buffer for storing the send values. */
        std::vector<TT> mDataSent;

        /** Buffer for storing the receive values. */
        std::vector<TT> mDataReceived;

        /** Content of the send buffer that is separated from the structure. */
        boost_mpi::content mvContSend;

        /** Content of receive buffer that is separated from the structure. */
        boost_mpi::content mvContRecv;

        /** List of interfaces to the neighbours. */
        // std::vector<Communicator> vCommunicator;

        /** List of request states of sending and receiving. */
        std::vector<boost_mpi::request> mReqs;
}; // End of class. //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c Buffer. */
template<typename TT>
void swap(Buffer<TT>& first, Buffer<TT>& second);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename TT>
inline Buffer<TT>::Buffer()
    : mMyWorld()
    , mIdNeighPartition(-1)
    , mSizeBuffer(0)
    , mUseSkeletonConcept(false)
    , mDataSent()
    , mDataReceived()
    , mvContSend()
    , mvContRecv()
    , mReqs()
{
    mReqs.resize(2);
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Buffer<TT>::Buffer(boost_mpi::communicator const& myWorld,
                          int const& idNeighPartition,
                          int const& sizeBuffer,
                          bool const& useSkeletonConcept)
    : mMyWorld(myWorld)
    , mIdNeighPartition(idNeighPartition)
    , mSizeBuffer(sizeBuffer)
    , mUseSkeletonConcept(useSkeletonConcept)
    , mDataSent()
    , mDataReceived()
    , mvContSend()
    , mvContRecv()
    , mReqs()
{
    // At the beginning, the content of the buffers should be empty.
    mDataSent.resize(sizeBuffer);
    mDataReceived.resize(sizeBuffer);
    mReqs.resize(2);

#ifdef VTRACE
    VT_TRACER("Buffer()");
#endif
    if (this->useSkeletonConcept()) {
        boost_mpi::request reqsSend[1];
        boost_mpi::request reqsRecv[1];
        reqsSend[0] = this->myWorld().isend(this->idNeighPartition(),
                                            0, boost_mpi::skeleton(mDataSent));
        mvContSend = boost_mpi::get_content(mDataSent);
        reqsRecv[0] = this->myWorld().irecv(this->idNeighPartition(),
                                            0, boost_mpi::skeleton(mDataReceived));
        mvContRecv = boost_mpi::get_content(mDataReceived);
        boost_mpi::wait_all(reqsRecv, reqsRecv + 1);
        boost_mpi::wait_all(reqsSend, reqsSend + 1);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Buffer<TT>::Buffer(Buffer<TT> const& rhs)
    : mMyWorld(rhs.myWorld())
    , mIdNeighPartition(rhs.mIdNeighPartition)
    , mSizeBuffer(rhs.sizeBuffer())
    , mUseSkeletonConcept(rhs.useSkeletonConcept())
    , mDataSent(rhs.dataSent())
    , mDataReceived(rhs.dataReceived())
    , mvContSend(rhs.mvContSend)
    , mvContRecv(rhs.mvContRecv)
    , mReqs(rhs.mReqs)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Buffer<TT>& Buffer<TT>::operator=(Buffer<TT> rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Buffer<TT>::~Buffer()
{
}


/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename TT>
inline boost_mpi::communicator const& Buffer<TT>::myWorld() const
{
    return mMyWorld;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline int const& Buffer<TT>::idNeighPartition() const
{
    return mIdNeighPartition;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline int const& Buffer<TT>::sizeBuffer() const
{
    return mSizeBuffer;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool const& Buffer<TT>::useSkeletonConcept() const
{
    return mUseSkeletonConcept;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline std::vector<TT> const& Buffer<TT>::dataSent() const
{
    return mDataSent;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline std::vector<TT> const& Buffer<TT>::dataReceived() const
{
    return mDataReceived;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline TT const& Buffer<TT>::elemData(int const& idx) const
{
    return mDataReceived[idx];
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<typename TT>
inline void Buffer<TT>::setElemData(unsigned long int const& memoryPos,
                                    TT const& elemData)
{
    mDataSent[memoryPos] = elemData;
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
template<typename TT>
inline bool Buffer<TT>::operator==(Buffer<TT> const& rhs) const
{
    bool isEqual = true;

    if (idNeighPartition() != rhs.idNeighPartition()) {
        isEqual = false;
    }
    if (sizeBuffer() != rhs.sizeBuffer()) {
        isEqual = false;
    }
    if (dataSent() != rhs.dataSent()) {
        isEqual = false;
    }
    if (dataReceived() != rhs.dataReceived()) {
        isEqual = false;
    }
    return isEqual;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Buffer<TT>::operator!=(Buffer<TT> const& rhs) const
{
    return !(*this == rhs);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT>
void Buffer<TT>::sendValues()
{
#ifdef VTRACE
    VT_TRACER("Buffer::send()");
#endif
    if (mUseSkeletonConcept) {
        mvContSend = boost_mpi::get_content(mDataSent);
        mReqs[0] = myWorld().isend(mIdNeighPartition, 0, mvContSend);
    } else {
        mReqs[0] = myWorld().isend(mIdNeighPartition, 0, mDataSent);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT>
void Buffer<TT>::receiveValues()
{
#ifdef VTRACE
    VT_TRACER("Buffer::receive()");
#endif
    if (mUseSkeletonConcept) {
        mvContRecv = boost_mpi::get_content(mDataReceived);
        mReqs[1] = myWorld().irecv(mIdNeighPartition, 0, mvContRecv);
    } else {
        mReqs[1] = myWorld().irecv(mIdNeighPartition, 0, mDataReceived);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT>
void Buffer<TT>::waitAll()
{
#ifdef VTRACE
    VT_TRACER("Buffer::waitAll()");
#endif
    boost_mpi::wait_all(mReqs.begin(), mReqs.end());
}
/*----------------------------------------------------------------------------*/
template<typename TT>
template<class Archive>
inline void Buffer<TT>::serialize(Archive& ar,
                                  unsigned int const version)
{
    if (1 <= version) {
        ar& mMyWorld;
        ar& mIdNeighPartition;
        ar& mSizeBuffer;
        ar& mUseSkeletonConcept;
        ar& mDataSent;
        ar& mDataReceived;
        ar& mvContSend;
        ar& mvContRecv;
        ar& mReqs;
    }
}

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c Buffer. */
template<typename TT>
inline void swap(Buffer<TT>& first, Buffer<TT>& second)
{
    std::swap(first.mMyWorld, second.mMyWorld);
    std::swap(first.mIdNeighPartition, second.mIdNeighPartition);
    std::swap(first.mSizeBuffer, second.mSizeBuffer);
    std::swap(first.mUseSkeletonConcept, second.mUseSkeletonConcept);
    std::swap(first.mDataSent, second.mDataSent);
    std::swap(first.mDataReceived, second.mDataReceived);
    std::swap(first.mvContSend, second.mvContSend);
    std::swap(first.mvContRecv, second.mvContRecv);
    std::swap(first.mReqs, second.mReqs);
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
namespace boost
{
namespace serialization
{
/** Designed to set the boost serialization version of a class template. */
template<typename TT>
struct version< ScaFES::Buffer<TT> > {
    /** Sets the version number for serialization. */
    BOOST_STATIC_CONSTANT(unsigned BOOST_PFTO int, value = 2);
};
} // namespace serialization
} // namespace boost

#endif
