/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_COMPLEX_HPP_
#define SCAFES_COMPLEX_HPP_

//#include <cstdint>
#include <iostream>
#include <iomanip>
#include <ios>
#include <cmath>
#include <cstdlib>
//#include <tr1/unordered_map>

#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
namespace boost
{
namespace serialization
{
class access;
}
}
#include <boost/serialization/version.hpp>
#include <boost/serialization/pfto.hpp>
//#include <boost/serialization/complex.hpp>
#endif

// #ifdef SCAFES_HAVE_BOOST
// #include <boost/type_traits.hpp>
// #endif

namespace ScaFES
{
/*******************************************************************************
 ******************************************************************************/
/** The class template \code Complex<TT> \endcode provides methods for
 * handling complex number of a given type \c TT (usually \c double).
 */
template<typename TT>
class Complex
{
    public:
        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor.
         *  All elements are initialized by zero.*/
        Complex<TT>();

        /** Creates a special constructor for threedimensional vectors:
         *  Both parts are initialized by given values. */
        Complex<TT>(TT const&, TT const&);

        /** Creates own constructor:
         *  Only The REAL PART is initalized by the given value,
         *  the imaginary part remains ZERO! */
        Complex<TT>(TT const& re);

        /** Creates copy constructor. */
        Complex<TT>(Complex<TT> const& rhs);

        /** Creates copy assignment operator using the copy-and-swap idiom. */
        Complex<TT>& operator=(Complex<TT> rhs);

        /** Creates the destructor. */
        ~Complex();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the real part of this complex number. */
        const TT& real() const;

        /** Returns the imaginary part of this complex number. */
        const TT& imag() const;

        /** Returns the element at position \c idx in the complex number. */
        const TT& operator[] (std::size_t const& idx) const;

    public:
        /*----------------------------------------------------------------------
        | SETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns a reference to the element at position \c idx in the complex
         * number.
         */
        TT& operator[] (std::size_t const& idx);

    public:
        /*----------------------------------------------------------------------
        | COMPARISON METHODS.
        ----------------------------------------------------------------------*/
        /** Compares elementwise if this complex number is smaller
         *  than a given rhs complex number. */
        bool operator<(Complex<TT> const& rhs) const;

        /** Compares elementwise if this complex number is greater or equal
         *  than a a given rhs complex number. */
        bool operator>=(Complex<TT> const& rhs) const;

        /** Compares elementwise if this complex number is smaller or equal
         *  than a a given rhs complex number. */
        bool operator<=(Complex<TT> const& rhs) const;

        /** Compares elementwise if this complex number is greater than
         *  a given rhs complex number. */
        bool operator>(Complex<TT> const& rhs) const;

        /** Compares elementwise if this complex number is equal to a given rhs
         *  complex number. */
        bool operator==(Complex<TT> const& rhs) const;

        /** Compares elementwise if this complex number is unequal to
         * a given rhs complex number. */
        bool operator!=(Complex<TT> const& rhs) const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS
        ----------------------------------------------------------------------*/
        /** Computes the dimension of the complex number,
         * i.e., all elements will be multiplied.
         */
        TT size() const;

        /** Returns the absolute value. */
        double fabs() const;

         /** Serializes this class. */
        template<class Archive>
        void serialize(Archive& ar, unsigned int const version);

    public:
        /*----------------------------------------------------------------------
        | ARITHMETIC METHODS.
        ----------------------------------------------------------------------*/
        /** Creates an assignment operator.
         * Only the REAL part will be assigned to the given scalar.
         */
        Complex<TT>& operator=(TT const& re);

        /** Computes the additive inverse of this complex number. */
        Complex<TT>& operator-();

        /** Adds this complex number to a given scalar (elementwise). */
        Complex<TT>& operator+=(TT const& sca);

        /** Subtracts this complex number from a given scalar (elementwise). */
        Complex<TT>& operator-=(TT const& sca);


        /** Multiplies this complex number by a given scalar (elementwise). */
        Complex<TT>& operator*=(TT const& sca);

        /** Divides this complex number by a given scalar (elementwise). */
        Complex<TT>& operator/=(TT const& sca);

        /** Adds a given rhs complex number to this complex number. */
        Complex<TT>& operator+=(Complex<TT> const& rhs);

        /** Subtracts a given rhs complex number from this complex number. */
        Complex<TT>& operator-=(Complex<TT> const& rhs);

        /** Multiplies this complex number by a given rhs complex number. */
        Complex<TT>& operator*=(Complex<TT> const& rhs);

        /** Divides this complex number by a given rhs complex number. */
        Complex<TT>& operator/=(Complex<TT> const& rhs);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Overloads the output operator \c operator<<.
         * A complex number is printed this way: '(re, im})' */
        template<typename S>
        friend std::ostream& operator<< (std::ostream& output,
                                         Complex <S> const& t);

        /** Method to swap members of two complexs numbers. */
        template<typename S>
        friend void swap(Complex<S>& first, Complex<S>& second);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Real part of complex number. */
        TT mReal;

        /** Imaginary part of complex number. */
        TT mImag;
}; // End of class //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c Complex. */
template<typename TT>
void swap(Complex<TT>& first, Complex<TT>& second);
/*----------------------------------------------------------------------------*/
/** Preprares writing a complex number to output. */
template<typename TT>
std::ostream& operator<<(std::ostream& output, Complex<TT> const& rhs);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename TT>
inline Complex<TT>::Complex()
    : mReal(0)
    , mImag(0)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>::Complex(TT const& re, TT const& im)
    : mReal(re)
    , mImag(im)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>::Complex(TT const& re)
    : mReal(re)
    , mImag(0)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator=(Complex<TT> rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>::Complex(Complex<TT> const& rhs)
    : mReal(rhs.mReal)
    , mImag(rhs.mImag)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>::~Complex() { }

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename TT>
inline const TT& Complex<TT>::real() const
{
    return mReal;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline const TT& Complex<TT>::imag() const
{
    return mImag;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline const TT& Complex<TT>::operator[] (std::size_t const& position) const
{
    if (0 == position) {
        return mReal;
    } else {
        return mImag;
    }
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<typename TT>
inline TT& Complex<TT>::operator[] (std::size_t const& pos)
{
    return const_cast<TT&>(static_cast<const Complex<TT>&>(*this)[pos]);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT>
inline TT Complex<TT>::size() const
{
    return (mReal * mImag);
}
/*----------------------------------------------------------------------------*/
template <typename TT>
inline double Complex<TT>::fabs() const
{
    return ( mReal * mReal + mImag * mImag );
}
/*----------------------------------------------------------------------------*/
template<typename TT>
template<class Archive>
inline void Complex<TT>::serialize(Archive& ar, unsigned int const version)
{
    if (1 <= version) {
        ar & (mReal);
        ar & (mImag);
    }
}

/*******************************************************************************
 * ARITHMETIC METHODS.
 ******************************************************************************/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator-()
{
    mReal *= static_cast<TT>(-1.0);
    mImag *= static_cast<TT>(-1.0);
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator=(TT const& sca)
{
    mReal = sca;
    mImag = sca;
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator+=(TT const& sca)
{
    mReal += sca;
    mImag += sca;
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator-=(TT const& sca)
{
    mReal -= sca;
    mImag -= sca;
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator*=(TT const& sca)
{
    mReal *= sca;
    mImag *= sca;
    return *this;
}
/*----------------------------------------------------------------------------*/
// TODO: Throw an exception if the given scalar is near to zero.
template<typename TT>
inline Complex<TT>& Complex<TT>::operator/=(TT const& sca)
{
    mReal /= sca;
    mImag /= sca;
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator+=(Complex<TT> const& rhs)
{
    mReal += rhs.real();
    mImag += rhs.imag();
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator-=(Complex<TT> const& rhs)
{
    mReal -= rhs.real();
    mImag -= rhs.imag();
    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline Complex<TT>& Complex<TT>::operator*=(Complex<TT> const& rhs)
{
    TT tmpRe = mReal * rhs.real() - mImag * rhs.imag();
    mReal = tmpRe;
    TT tmpIm = mReal * rhs.imag() + mImag * rhs.real();
    mImag = tmpIm;
    return *this;
}
/*----------------------------------------------------------------------------*/
// TODO: Throw exception if 'sca' is near to zero.
template<typename TT>
inline Complex<TT>& Complex<TT>::operator/=(Complex<TT> const& rhs)
{
    TT denom = rhs.imag() * rhs.imag() + rhs.real() * rhs.real();
    TT tmpRe = (mReal * rhs.real() + mImag * rhs.imag()) / denom;
    mReal = tmpRe;
    TT tmpIm = (mImag * rhs.real() - mReal * rhs.imag()) / denom;
    mImag = tmpIm;
    return *this;
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
template<typename TT>
inline bool Complex<TT>::operator==(Complex<TT> const& rhs) const
{
    bool res = true;

    if (!(::fabs(mReal - rhs.real()) < 2.2e-15)) {
        res = false;
    }

    if (!(::fabs(mImag - rhs.imag()) < 2.2e-15)) {
        res = false;
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Complex<TT>::operator!=(Complex<TT> const& rhs) const
{
    return !(*this == rhs);
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Complex<TT>::operator<(Complex<TT> const& rhs) const
{
    bool res = true;

    if (!(mReal < rhs.real())) {
        res = false;
    }

    if (!(mImag < rhs.imag())) {
        res = false;
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Complex<TT>::operator<=(Complex<TT> const& rhs) const
{
    bool res = true;

    if ((mReal > rhs.real())) {
        res = false;
    }

    if ((mImag > rhs.imag())) {
        res = false;
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Complex<TT>::operator>=(Complex<TT> const& rhs) const
{
    bool res = true;

    if ((mReal < rhs.real())) {
        res = false;
    }

    if ((mImag < rhs.imag())) {
        res = false;
    }

    return res;
    //return !(*this < rhs); Does not work because we are working elementwise!
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline bool Complex<TT>::operator>(Complex<TT> const& rhs) const
{
    bool res = true;

    if (!(mReal > rhs.real())) {
        res = false;
    }

    if (!(mImag > rhs.imag())) {
        res = false;
    }

    return res;
    //return !(*this <= rhs); Does not work because we are working elementwise!
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
template<typename TT>
inline void swap(Complex<TT>& first, Complex<TT>& second)
{
    std::swap(first.mReal, second.mReal);
    std::swap(first.mImag, second.mImag);
}
/*----------------------------------------------------------------------------*/
template<typename TT>
inline std::ostream& operator<<(std::ostream& output, Complex<TT> const& rhs)
{
    TT epsilon = static_cast<TT>(2.2e-15);

    if (-epsilon < rhs.real() && rhs.real() < epsilon) {
        output << "[    0,";
    } else {
        output << "[" << ::std::setw(4) << ::std::right << rhs.real() << ";";
    }

    if (-epsilon < rhs.imag() && rhs.imag() < epsilon) {
        output << "    0i]";
    } else {
        output << ::std::setw(4) << ::std::right << rhs.imag() << "i]";
    }

    return output;
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
namespace boost
{
namespace serialization
{
/** Designed to set the boost serialization version of a class template. */
template<typename TT>
struct version< ScaFES::Complex<TT> > {
    /** Sets the version number for serialization. */
    BOOST_STATIC_CONSTANT(unsigned long int, value = 2);
};
}
}

#endif
