/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_DATAFIELD_HPP_
#define SCAFES_DATAFIELD_HPP_

#include "ScaFES_Config.hpp"

// #include <cassert>
#include <iostream>
#include <algorithm>
#include <utility>
// #include <ios>
#include <vector>
#include <string>

// #ifdef SCAFES_HAVE_BOOST_SERIALIZATION
// namespace boost
// {
// namespace serialization
// {
// class access;
// }
// }
// #include <boost/serialization/version.hpp>
// #include <boost/serialization/pfto.hpp>
// #include <boost/serialization/vector.hpp>
// #include <boost/serialization/base_object.hpp>
// #include <boost/serialization/complex.hpp>
// #endif

// #include <boost/mpi.hpp>

#ifdef VTRACE
#include "vt_user.h"
#endif

#ifdef _OPENMP
extern "C" {
#include <omp.h>
}
#endif

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Complex.hpp"
#include "ScaFES_Complex_FreeFunctions.hpp"
#include "ScaFES_Parameters.hpp"
#include "ScaFES_Grid.hpp"
#include "ScaFES_GridSub.hpp"
#include "ScaFES_GridGlobal.hpp"
#include "ScaFES_Buffer.hpp"
#include "ScaFES_DataFile.hpp"

namespace ScaFES
{

/** Global constant: Used in checks if field values are equal. */
const double EPSILON = 2.2e-10;

/*******************************************************************************
 ******************************************************************************/
/** The class \c DataField is the key to the data management in
* ScaFES.
*
* \n Physical fields are partitioned into subfields accordingly to the domain
* decomposition. \c DataField represents the geometric description of these
* subfields.
* \n Functions, like the pressure (a real valued function) or the velocity
* (a vector valued function), will be modeled using this class.
* An object of this class stores the function values at all grid nodes.
*
* \n The grid nodes are numbered in two different ways:
* <ul>
* <li> globally, i.e. all grid nodes of the WHOLE MPI domain are numbered
*      lexicographically, </li>
* <li> locally, i.e. all grid nodes of ONE MPI partition are numbered
*      lexicographically. </li>
* </ul>
*
* \n As the global computational domain (= the global grid) is subdivided
* into subdomains (so-called partitions), physical fields are partitioned
* and mapped to the corresponding MPI processes, too. In this case, an object
* of type \c DataField can have an extended grid (so called ghost cells
* with ghost nodes) where neighbouring processes store their function values.
* This means, all function values at these ghost nodes are stored at multiple
* locations to enable and speed up the computation.
*
* \n The node numbering starts with zero at the first ghost node of the first
* ghost cell. If there are no ghost cells, then the first local node number
* \c will be zero.
*
* \n The class provides the memory for storing the function values at all grid
* nodes.
*
* \n Example: In order to access the elements of a three-dimensional data field
* named \c E, the following possibilities exist:
* <ul>
* <li> \code int i, j, k;           E(i,j,k)    = ...\endcode </li>
* <li> \code unsigned long int memoryPos; E(memoryPos)  = ...\endcode </li>
* <li> \code Index idxNode;        E(idxNode) = ...\endcode </li>
* </ul>
* Elements of a data field can be accesed via an iterator.
*
* \remarks \c memoryPos is the MPI process local memory position.
* This memory position is only the same for those data fields which have
* the same ghost cell width.
* \c idxNode returns the MPI global node number.
*  ==> Using \c memoryPos is better for performance reasons:
*  It avoids unnecessary recalculations (using the method
*  \c idxNode2MemoryPos()) for accessing the correct memory locations,
*  but: if one is unsure about indexing, one should use \c idxNode:
*  It always gives the global node number and thus, the correct memory location
*  can be accessed.
*
* \remarks There are two variants for each access method:
* According to the class \c std::vector<CT>, the \c operator(node) is overloaded
* and returns the addressed component. One the other hand, the method
* \code at(node) \endcode performs a range check and returns the addressed
* component.
*
* \remarks For the moment, only symmetric ghost layers with respect to the
* axis are considered.
*
* \n Assumption: The computational domain should be convex.
* -> There exists at most one common face.
*
* \n The synchronization is performed in three steps:
* a) Copy values to the send buffer,
* b) exchange values by storing the values of the send buffer
*    in the receive buffer of the neighbour and vice versa.
* c) copy values from the receive buffer.

* \n The synchronization is done if and only if the size of MPI communciator is
* greater than one or the border width of the data field is greater than zero.
* 
* The method \c addDataField inserts a new data field to the problem.
* If additional memory for the ghost nodes is necessary then the parameter
* \c stencilWidth must be set appropriately.
* \code addDataField(G, 2);  // G has 2 ghost node layers.\endcode

* \n The class possesses different predefined sub grids such that one can easily
* iterated over these special grid using the iterator of the class \c GridSub.
* \c memGhost Memory representation of all ghost nodes.
* \c memBorder Memory representation of all boundary nodes.
* \c memNormal Memory representation of all normal nodes
* \c memAll Memory representation of really all nodes (incl. ghost nodes).
*
* \n \code
* \begin{figure}[h!] \label{pic:wrongGhostInit}
* \centering
*   \includegraphics[width=70mm, keepaspectratio, trim=120mm 30mm 120mm 65mm,
* clip]{graphics/timedependend_noInit_ghostcells.png}
*   %\rule{2cm}{2cm}
*   \caption{The application Supra shows nPartitions strips in the first
*    iterations. These strips are related to a wrong initialisation of the
*   values at the ghost nodes (incorrect gradients at local domain boundaries).
* \end{figure}
* \endcode
*/
template<typename CT, std::size_t DIM>
class DataField
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

        /** Re-export typename CT STL-like. */
        typedef CT value_type;

    public:
        /*----------------------------------------------------------------------
        | INTERNAL CLASSES.
        ----------------------------------------------------------------------*/
        /** Class for iterating through the grid.*/
        class Iterator
            : public std::iterator<std::random_access_iterator_tag , Index>
        {

                /*--------------------------------------------------------------
                | FRIEND CLASSES.
                --------------------------------------------------------------*/
                friend class DataField<CT, DIM>;

            public:
                /*--------------------------------------------------------------
                | LIFE CYCLE METHODS.
                --------------------------------------------------------------*/
                /** Creates own constructor. */
                Iterator(const Grid<DIM>* gd, Index const& idxNode);

//                 /** Creates own copy constructor. */
//                 Iterator(Iterator const&); // = default;
// 
//                 /** Creates own assignment operator. */
//                 Iterator& operator=(Iterator const&); // = default;

            public:
                /*--------------------------------------------------------------
                | GETTER METHODS.
                --------------------------------------------------------------*/
                /** Returns the current global node number. */
                Index const& idxNode() const;

                /** Returns the current memory position. */
                unsigned long int const& memoryPos() const;

            public:
                /*--------------------------------------------------------------
                | COMPARISON METHODS.
                --------------------------------------------------------------*/
                /** Checks if the local node number of this iterator is unequal
                 * to the one of the given iterator. */
                bool operator!=(const Iterator& other) const;

                /** Checks if the local node number of this iterator is smaller
                 * than the one of the given iterator. */
                bool operator<(const Iterator& other) const;

            public:
                /*--------------------------------------------------------------
                | WORK METHODS.
                --------------------------------------------------------------*/
                /** Returns the current global node number. */
                Index const& operator*();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global node
                 * number.
                 * \remarks Prefix variant.
                 */
                Iterator& operator++();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global node
                 * number.
                 * \remarks Postfix variant.
                 */
                Iterator operator++(int incr);

            private:
                /*--------------------------------------------------------------
                | WORK METHODS.
                --------------------------------------------------------------*/
                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global node
                 * number iterating "columnwise" through the grid.
                 */
                Iterator& nextStyleC();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global node
                 * number iterating "rowwise" through the grid.
                 */
                Iterator& nextStyleFortran();

            private:
                /*--------------------------------------------------------------
                | MEMBER VARIABLES.
                --------------------------------------------------------------*/
                /** Grid over which should be iterated. */
                const Grid<DIM>* mGridIter;

                /** Current global node number. */
                Index mIdxNode;

                /** Current local node number. */
                unsigned long int mMemoryPos;
        };

    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for iterator class analogously to STL class. */
        typedef Iterator iterator;

        /** Type definition for iterator class analogously to STL class. */
        typedef const Iterator const_iterator;

        /** Type definition for iterator class analogously to STL class. */
        typedef Index& reference_type;

//         /*----------------------------------------------------------------------
//         | FRIEND CLASSES.
//         ----------------------------------------------------------------------*/
//         friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates default constructor.
         * All members variables are initialised by zero. */
        DataField(); // = default;

        /** Creates another own constructor.
         * Sets the name of the data field, the global grid, the border width,
         * the parameter set.
         */
        DataField(std::string const& nameDataField,
                  Parameters const& params,
                  GridGlobal<DIM> const& gg,
                  int const& stencilWidth);

        /** Creates another own constructor.
         * Sets the name of the data field, the global grid, the border width,
         * the parameter set and the grid of the hole (where the data field
         * is not defined).
         */
        DataField(std::string const& nameDataField,
                  Parameters const& params,
                  GridGlobal<DIM> const& gg,
                  int const& stencilWidth,
                  int const& nLayers);

        /** Creates own constructor: Data field will be initialised by a
         * given function. */
        template<typename PT>
        DataField(std::string const& nameDataField,
                  PT& prblm,
                  void (PT::*fptrInit)(CT&,
                                       Ntuple<double, DIM> const&,
                                       double const&),
                  Parameters const& params,
                  GridGlobal<DIM> const& gg,
                  int const& stencilWidth);

        /** Creates own constructor: Data field will be initialised by a
         * given function. */
        DataField(std::string const& nameDataField,
                  void (*fptrInit)(CT&,
                                   Ntuple<double, DIM> const&,
                                   double const&),
                  Parameters const& params,
                  GridGlobal<DIM> const& gg,
                  int const& stencilWidth);

        /** Creates own copy constructor. */
        DataField(DataField<CT, DIM> const& from);

        /** Creates own assignment operator using the copy and swap idiom. */
        DataField<CT, DIM>& operator=(DataField<CT, DIM> from);

        /** Creates own destructor. */
        ~DataField();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the parameter set. */
        std::string const& name() const;

        /** Returns the MPI rank of the communicator */
        Parameters const& params() const;

        /** Returns the global grid which includes all partitions. */
        GridGlobal<DIM> const& gridGlobal() const;

        /** Returns the number of ghost cells. */
        int const& stencilWidth() const;

        /** Returns the grid incl. ghost nodes of the data field. */
        Grid<DIM> const& memAll() const;

        /** Returns the normal grid of the data field. */
        GridSub<DIM> const& memNormal() const;

        /** Returns the grid where the data field is NOT defined. */
        GridSub<DIM> const& memHole() const;

        /** Returns the interior grid of the data field. */
        GridSub<DIM> const& memInner() const;

        /** Returns the element size of the underlying type. */
        std::size_t elementSize() const;

        /** Returns the memory. */
        CT* const& elemData() const;

        /** Returns the element \c idx of the memory. */
        CT const& elemData(int const& ii) const;

        /** Returns the border grid of the data field. */
        std::vector< GridSub<DIM> > const& memBorder() const;

        /** Returns the communication grid of the data field. */
        std::vector< GridSub<DIM> > const& memComm() const;

        /** Returns the ghost grid of the data field. */
        std::vector< GridSub<DIM> > const& memGhost() const;

    public:
        /*----------------------------------------------------------------------
        | SETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Address elements via local node numbers.
         * Returns a reference, what makes an assignment like \e A(0) = 2
         * possible.
         * \param memoryPos Local node number.
         */
        CT& operator()(unsigned long int const& memoryPos) const;

        /** Address elements via global node numbers.
         * \param idxNode Global node number.
         */
        CT& operator()(Index const& idxNode) const;

        /** Address elements via global node number give explicitly as (i,j,k).
         * \param i First entry of global node number.
         * \param j Second entry of global node number.
         * \param k Third entry of global node number.
         */
        CT& operator()(int const& i, int const& j, int const& k) const;

        /** Address elements via global node number give explicitly as (i,j).
         * \param i First entry of global node number.
         * \param j Second entry of global node number.
         */
        CT& operator()(int const& i, int const& j) const;

        /** Overloads ()-operator to address elements via local node numbers.
         * This method includes a range check.
         * \param memoryPos Local node number.
         */
        CT& at(unsigned long int const& memoryPos) const;

        /** Overloads ()-operator to address elements via global node numbers.
         * This method includes a range check.
         * \param idxNode Global node number.
         */
        CT& at(Index const& idxNode) const;

        /** Overloads ()-operator to address elements via global node number
         * (i,j,k). This method includes a range check.
         * \param i First entry of global node number.
         * \param j Second entry of global node number.
         * \param k Third entry of global node number.
         */
        CT& at(int const& i, int const& j, int const& k) const;

        /** Overloads the []-operator to address elements via the local
         * node number.
         * \remarks This method is forbidden for safety reasons. */
        //CT& operator[] (unsigned long int const& memoryPos) = delete;

    public:
        /*----------------------------------------------------------------------
         | COMPARISON METHODS.
         ---------------------------------------------------------------------*/
        /** Compares two data fields for equality.
         * First, geometrical sizes are compared.
         * In a second step, the method haveSameValues(DF1, DF2) compares
         * node values.
         */
        bool operator==(DataField<CT, DIM> const& second);

        /** Checks if this data field and the rhs data field have the same
         * geometrical size.
         */
        bool hasSameSize(DataField<CT, DIM> const& rhs) const;

        /** Checks if this data field and the rhs data field have the same
         *  values.
         */
        bool hasSameValues(DataField<CT, DIM> const& rhs,
                           double const& eps) const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Determines the memory location in the send buffer for a given
         * global node of a given grid. */
        unsigned long int idxNode2Buffer(Index const& idxNode,
                                          Grid<DIM> const& gg) const;

        /** Synchronizes the data exchange of the ghost cell values with all
         *  neighbours.
         */
        void sync(int const& timestep);

        /** Sums up and synchronizes the data exchange of the ghost cell values
         *  with all neighbours.
         */
        void collectValuesAtMemComm(int const& timestep);

        /** Copies the values of the memory of the communication nodes
         *  of this data field to the send buffer.
         */
        void copyValuesFromMemCommToSendBuffer(int const& timestep);

        /** Copies the values of the memory of the ghost nodes
         *  of this data field to the send buffer.
         */
        void copyValuesFromMemGhostToSendBuffer(int const& timestep);

        /** Communicates asynchronously with direct neighbours.
         * Manages the sending and receiving values of the ghost
         * nodes of a data field to the corresponding neighboured data field.
         */
        void exchangeValuesInBuffers(int const& timestep);

        /** Waits until all communication calls have been finished. */
        void waitAll();

        /** Copies the values of the receive buffer to the memory
         *  of the ghost nodes of this data field.
         */
        void copyValuesFromReceiveBufferToMemGhost(int const& timestep);

        /** Copies the values of the receive buffer to the memory
         *  of the communication nodes of this data field.
         */
        void copyValuesFromReceiveBufferToMemComm(int const& timestep);

        /** Copies and sums the values of the receive buffer to the memory
         *  of the communication nodes of this data field.
         */
        void copyAndSumValuesFromReceiveBufferToMemComm(int const& timestep);

        /** Writes the data field to a NetCDF file. */
        void write(int const& timestep);

        /** Reads the data field from a NetCDF file. */
        void read(int const& timestep);

        /** Adds the values at all grid nodes of a given data field
         *  to the existing one. */
        DataField<CT, DIM>& operator+= (DataField<CT, DIM> const& rhs);

        /** Subtracts the values at all grid nodes of a given data field
         *  to the existing one. */
        DataField<CT, DIM>& operator-= (DataField<CT, DIM> const& rhs);

        /** Sets the elements of a data field at all "normal" nodes
         * (except those nodes at the ghost cells) to a given value \c val.
         * \param val Given value.
         */
        void operator=(CT const& val);

        /** Returns the iterator corresponding to the underlying memory grid,
         * starting with the first node of the grid. */
        iterator begin() const;

        /** Returns the iterator corresponding to the underlying memory grid,
         * starting with the last node of the grid. */
        iterator end() const;

        /** Computes the corresponding memory position from a given
         * global node number.
         * \param idxNode Given global node number.
         * \return Corresponding memory position.
         */
        unsigned long int idxNode2MemoryPos(Index const& idxNode) const;

        /** Computes the corresponding global node number from a given
         * local memory position.
         * \param memoryPos Given memory position.
         * \return Corresponding global node number.
         */
        Index memoryPos2IdxNode(unsigned long int const& memoryPos) const;

        /** Swaps the values of two data fields.
         * \remarks For time dependent equations, it is necessary to change the
         * meaning of the new and the old data field.
         * The old data field will be reused as new data field in the next
         * timestep.
         */
        void swapWith(DataField<CT, DIM>& df);

        /** Assigns node values from one data field to another data field.
         * \remarks Copies values of all (normal + ghost) nodes!
         */
        void assignValues(DataField<CT, DIM> const& from);

        /** Checks if this data field and another given data field have the
         * same values. */
        void check(DataField<CT, DIM> const& df2);

        /** Computes the L^{inf} norm (maximum norm) of this data field. */
        double normLinf() const;

        /** Evaluates a given function at the whole grid partition
         * (Version using a member function pointer). */
        template<typename PT, typename TT>
        void evalAtGrid(PT& prblm,
                        void (PT::*fptrEvalFunc)(CT&, Ntuple<TT, DIM> const&,
                                                 TT const&),
                        int const& timestep);

        /** Evaluates a given function at the whole grid partition
         * (Version using a global function pointer). */
        template<typename TT>
        void evalAtGrid(void (*fptrEvalFunc)(CT&, Ntuple<TT, DIM> const&,
                                             TT const&),
                        int const& timestep);

        /** Computes the error of this and another given data field
         * using the maximum norm. */
        void compErrLinf(DataField<CT, DIM> const& dfAna);

//         /** Serializes this class. */
//         template<class Archive>
//         void serialize(Archive& ar, unsigned int const version);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Swaps two objects of this class. */
        template<typename SS, std::size_t DD>
        friend void swap(DataField<SS, DD>& first,
                         DataField<SS, DD>& second);

        /** Overloads the output operator \c operator<<. */
        template<typename ST, std::size_t DD>
        friend std::ostream& operator<<(std::ostream& output,
                                        DataField<ST, DD> const& df);

    protected:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Parameter set. */
        Parameters mParams;

        /** Global grid including all grid partitions. */
        GridGlobal<DIM> mGG;

        /** Underlying grid where the data field lives:
          Roughly spoken: grid partition + grid(stencilWidth).
         * A data field can be restricted to a partition of the global grid.
         * Thus, \c mMemAll.idxNodeFirst() indicates the first global node
         * number of the data field (numbering incl. nodes of ghost cells).
         */
        Grid<DIM> mMemAll;

        /** The normal grid: All grid nodes without the nodes at the ghost
         * cells. */
        GridSub<DIM> mMemNormal;

        /** (Global) grid where the data field is NOT defined
          (derived from the global grid). */
        GridSub<DIM> mMemHole;

        /** Grid of all border nodes. */
        std::vector< GridSub<DIM> > mMemBorder;

        /** Memory of all communication nodes. */
        std::vector< GridSub<DIM> > mMemComm;

        /** Memory of all ghost nodes. */
        std::vector< GridSub<DIM> > mMemGhost;

        /** Number of ghost cells (uniform in all directions). */
        int mStencilWidth;

        /** Name of data field. */
        std::string mName;

        /** Used in assertions to make sure that only elements in grid are
         * addressed. */
        std::size_t mMemoryPosFirst;

        /** Used in assertions to make sure that only elements in grid are
         * addressed. */
        std::size_t mMemoryPosLast;

        /** Linear continuous memory grid for the elements of a data field.
         * A data field can only store values of one type.
         * Moreover, the type remains constant during the life time
         * of the data field.
         */
        CT* mElemData;

        /** Buffer for all elements of the data field which should be sent
         * to and received from the neighbour.
         */
        std::vector< Buffer<CT> > mValuesToExchange;

        /** Data file. */
        DataFile<CT, DIM> mOutput;

        /** List of interfaces to the neighbours. */
        // std::vector<Communicator> vCommunicator;

    private:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Initializes dependent member variables. */
        void initDependentMembers();

        /** Performs range check. */
        void checkRange(unsigned long int const& memoryPos) const;

        /** Computes the corresponding memory position from a given
         * global node number using the C style numbering.
         * \param idxNode Given global node number.
         * \return Corresponding memory position.
         */
        unsigned long int idxNode2MemoryPosStyleC(Index const& idxNode) const;

        /** Computes the corresponding memory position from a given
         * global node number using the FORTRAN style numbering.
         * \param idxNode Given global node number.
         * \return Corresponding memory position.
         */
        unsigned long int idxNode2MemoryPosStyleFortran(Index const& idxNode) const;

        /** Computes the corresponding global node number from a given
         * local memory position using the C style numbering.
         * \param memoryPos Given memory position.
         * \return Corresponding global node number.
         */
        Index memoryPos2IdxNodeStyleC(unsigned long int const& memoryPos) const;

        /** Computes the corresponding global node number from a given
         * local memory position using the FORTRAN style numbering.
         * \param memoryPos Given memory position.
         * \return Corresponding global node number.
         */
        Index memoryPos2IdxNodeStyleFortran(unsigned long int const& memoryPos) const;

    private:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Computes the error in the L-inf norm (Internal method). */
        template<std::size_t DD>
        friend double compErrLinfFree(DataField<double, DD> const& df1,
                                      DataField<double, DD> const& df2);

        /** Computes the error in the L-inf norm (Internal method). */
        template<std::size_t DD>
        friend double compErrLinfFree(DataField<Complex<double>, DD> const& df1,
                                      DataField<Complex<double>, DD> const& df2);

        /** Computes the error in the L-inf norm (Internal method). */
        template<std::size_t DD>
        friend double compErrLinfFree(DataField<Ntuple<double, DD>, DD> const& df1,
                                      DataField<Ntuple<double, DD>, DD> const& df2);

        /** Checks if two data fields have the same values (Internal method). */
        template<std::size_t DD>
        friend bool haveSameValuesFree(DataField<double, DD> const& df1,
                                       DataField<double, DD> const& df2);

        /** Checks if two data fields have the same values (Internal method). */
        template<std::size_t DD>
        friend bool haveSameValuesFree(DataField<Complex<double>, DD> const& df1,
                                       DataField<Complex<double>, DD> const& df2);

        /** Checks if two data fields have the same values (Internal method). */
        template<std::size_t DD>
        friend bool haveSameValuesFree(DataField<Ntuple<double, DD>, DD> const& df1,
                                       DataField<Ntuple<double, DD>, DD> const& df2);

}; // End of class. //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c DataField. */
template<typename CT, std::size_t DIM>
void swap(DataField<CT, DIM>& first, DataField<CT, DIM>& second);
/*----------------------------------------------------------------------------*/
/** Preprares writing a data field to output. */
template<typename CT, std::size_t DIM>
std::ostream& operator<< (std::ostream& output,
                          DataField<CT, DIM> const& df);
/*----------------------------------------------------------------------------*/
/** Computes the error in the L-inf norm (Internal method). */
template<std::size_t DIM>
double compErrLinfFree(DataField<double, DIM> const& df1,
                       DataField<double, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Computes the error in the L-inf norm (Internal method). */
template<std::size_t DIM>
double compErrLinfFree(DataField<Complex<double>, DIM> const& df1,
                       DataField<Complex<double>, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Computes the error in the L-inf norm (Internal method). */
template<std::size_t DIM>
double compErrLinfFree(DataField<Ntuple<double, DIM>, DIM> const& df1,
                       DataField<Ntuple<double, DIM>, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Checks if two data fields have the same values (Internal method). */
template<std::size_t DIM>
bool haveSameValuesFree(DataField<double, DIM> const& df1,
                        DataField<double, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Checks if two data fields have the same values (Internal method). */
template<std::size_t DIM>
bool haveSameValuesFree(DataField<Complex<double>, DIM> const& df1,
                        DataField<Complex<double>, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Checks if two data fields have the same values (Internal method). */
template<std::size_t DIM>
bool haveSameValuesFree(DataField<Ntuple<double, DIM>, DIM> const& df1,
                        DataField<Ntuple<double, DIM>, DIM> const& df2);
/*----------------------------------------------------------------------------*/
/** Adds the elements of two data fields. */
template<typename CT, std::size_t DIM>
const DataField<CT, DIM> operator+(DataField<CT, DIM> const& lhs,
                                          DataField<CT, DIM> const& rhs);
/*----------------------------------------------------------------------------*/
/** Subtracts the elements of two data fields. */
template<typename CT, std::size_t DIM>
const DataField<CT, DIM> operator-(DataField<CT, DIM> const& lhs,
                                          DataField<CT, DIM> const& rhs);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::DataField()
    : mParams()
    , mGG()
    , mMemAll()
    , mMemNormal()
    , mMemHole()
    , mMemBorder()
    , mMemComm()
    , mMemGhost()
    , mStencilWidth(0)
    , mName("EMPTYNAME")
    , mMemoryPosFirst(0L)
    , mMemoryPosLast(0L)
    , mElemData(NULL)
    , mValuesToExchange()
    , mOutput()
{
    initDependentMembers();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::DataField(std::string const& nameDataField,
                                     Parameters const& params,
                                     GridGlobal<DIM> const& gg,
                                     int const& stencilWidth)
    : mParams(params)
    , mGG(gg)
    , mMemAll(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth),
              gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth)),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth)))
    , mMemNormal(gg.partition(params.myWorld().rank()).idxNodeFirstSub(),
                 gg.partition(params.myWorld().rank()).idxNodeLastSub(),
                 mMemAll)
    , mMemHole()
    , mStencilWidth(stencilWidth)
    , mName(nameDataField)
    , mMemoryPosFirst(0L)
    , mMemoryPosLast(512L)
    , mElemData(NULL)
    , mValuesToExchange()
{
    initDependentMembers();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::DataField(std::string const& nameDataField,
                                     Parameters const& params,
                                     GridGlobal<DIM> const& gg,
                                     int const& stencilWidth,
                                     int const& nLayers)
    : mParams(params)
    , mGG(gg)
    , mMemAll(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth),
              gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth)),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth)))
    , mMemNormal(gg.partition(params.myWorld().rank()).idxNodeFirstSub(),
                 gg.partition(params.myWorld().rank()).idxNodeLastSub(),
                 mMemAll)
    , mMemHole(gg.idxNodeFirst() + Index(nLayers),
               gg.idxNodeLast() - Index(nLayers),
               gg)
    , mStencilWidth(stencilWidth)
    , mName(nameDataField)
    , mMemoryPosFirst(0L)
    , mMemoryPosLast(512L)
    , mElemData(NULL)
    , mValuesToExchange()
{
    initDependentMembers();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename PT>
inline DataField<CT, DIM>::DataField(std::string const& nameDataField,
                                     PT& prblm,
                                     void (PT::*fptrInit)(CT&,
                                                          Ntuple<double, DIM> const&,
                                                          double const&),
                                     Parameters const& params,
                                     GridGlobal<DIM> const& gg,
                                     int const& stencilWidth)
    : mParams(params)
    , mGG(gg)
    , mMemAll(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth),
              gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth)),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth)))
    , mMemNormal(gg.partition(params.myWorld().rank()).idxNodeFirstSub(),
                 gg.partition(params.myWorld().rank()).idxNodeLastSub(),
                 mMemAll)
    , mMemHole()
    , mStencilWidth(stencilWidth)
    , mName(nameDataField)
{
    initDependentMembers();
    evalAtGrid(prblm, fptrInit, 0);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::DataField(std::string const& nameDataField,
                                     void (*fptrInit)(CT&,
                                                      Ntuple<double, DIM> const&,
                                                      double const&),
                                     Parameters const& params,
                                     GridGlobal<DIM> const& gg,
                                     int const& stencilWidth)
    : mParams(params)
    , mGG(gg)
    , mMemAll(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth),
              gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeFirstSub() - Index(stencilWidth)),
              gg.coordinates(gg.partition(params.myWorld().rank()).idxNodeLastSub() + Index(stencilWidth)))
    , mMemNormal(gg.partition(params.myWorld().rank()).idxNodeFirstSub(),
                 gg.partition(params.myWorld().rank()).idxNodeLastSub(),
                 mMemAll)
    , mMemHole()
    , mStencilWidth(stencilWidth)
    , mName(nameDataField)
    , mValuesToExchange()
{
    initDependentMembers();
    evalAtGrid(fptrInit, 0);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void DataField<CT, DIM>::initDependentMembers()
{
    assert(0 <= stencilWidth());
    assert(gridGlobal().idxNodeFirst()
           <= gridGlobal().partition(mParams.myWorld().rank()).idxNodeFirstSub());
    assert(gridGlobal().partition(mParams.myWorld().rank()).idxNodeLastSub()
           <= gridGlobal().idxNodeLast());

    mMemoryPosFirst = idxNode2MemoryPos(this->memAll().idxNodeFirst());
    mMemoryPosLast = idxNode2MemoryPos(this->memAll().idxNodeLast());
    assert(0 == mMemoryPosFirst);

    std::vector<std::string> tmpName;
    tmpName.push_back(this->name());
    mOutput = DataFile<CT, DIM>(params().nameDataFile(), tmpName,
                       params().myWorld(), gridGlobal().nNodes());

    unsigned long int nElems = this->memAll().nNodes().size();

//     GridSub<DIM> memIntersect(max(memNormal().idxNodeFirstSub(),
//                                   memHole().idxNodeFirstSub()),
//                               min(memNormal().idxNodeLastSub(),
//                                   memHole().idxNodeLastSub()),
//                               memAll());    
//     int nElemsIntersect = memIntersect.nNodesSub().size();
//     std::cout << " rank= " << params().myWorld().rank()
//               << " Nelems = " << nElems
//               << "  NelemsIntersect = " <<nElemsIntersect
//               << " gg.partition = "
//               << std::endl
//               << gridGlobal().partition(params().myWorld().rank())
//                << std::endl;

    // Map n-dimensional grid to one dimensional memory array.
    // This ensures that the one memory lump is allocated for the data field!
    // Alternative memory layout:
    // mElemData = new CT[dim0][dim1][dim2];
    // ...
    mElemData = NULL;

    if (0 < nElems) {
        mElemData = new CT[nElems];

        // Set data field at all nodes (especially at ghost nodes) to zero
        // in order to be sure that there is no dump in the memory.
        // Apply the first-touch-policy: 
        // All elements of a data field which will be updated by a given 
        // thread (later on), should be initialised by the same thread.
       int ii = 0;
       int nNodesTotal = memAll().nNodes().size();
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii) \
    shared(nNodesTotal)
#endif
        for (ii = 0; ii < nNodesTotal; ++ii) {
            mElemData[ii] = -3.1467;
        }
    }

    // TODO: Check if the value of border width is too large.
    // Is the grid too small to set all faces?
    // There should be at least 3*stencilWidth()() nodes in every direction.
    Index itStart = this->memNormal().idxNodeFirstSub();
    Index itEnd = this->memNormal().idxNodeLastSub();
    Index diff = this->memNormal().nNodesSub();
    int bw = std::max(this->stencilWidth(), 1);

    // Create iterators for all boundary faces of the current underlying
    // grid partition.
    for (std::size_t ii = 0; ii < DIM; ++ii) {
        // Does a face fit?
        if (diff.elem(ii) >= bw) {
            // Left, front, bottom, ...
            Index tmpIdxNodeFirst(itStart);
            Index tmpIdxNodeLast(itEnd);
            tmpIdxNodeLast[ii] = itStart.elem(ii) + bw;

            for (std::size_t jj = 0; jj < ii; ++jj) {
                tmpIdxNodeFirst[jj] += bw;
                tmpIdxNodeLast[jj]  -= bw;
            }

            GridSub<DIM> gridToAdd(tmpIdxNodeFirst,
                                   tmpIdxNodeLast,
                                   this->memAll());
            mMemBorder.push_back(gridToAdd);

            // Does an additional face fit?
            if (diff.elem(ii) > 2 * bw) {
                // Right, back, top,...
                Index tmpIdxNodeFirst2(itStart);
                Index tmpIdxNodeLast2(itEnd);
                tmpIdxNodeFirst2[ii] = itEnd.elem(ii) - bw;

                for (std::size_t jj = 0; jj < ii; ++jj) {
                    tmpIdxNodeFirst2[jj] += bw;
                    tmpIdxNodeLast2[jj]  -= bw;
                }

                GridSub<DIM> gridToAddFit(tmpIdxNodeFirst2,
                                          tmpIdxNodeLast2,
                                          this->memAll());
                mMemBorder.push_back(gridToAddFit);
            }
        }
    }

    // Add buffers for the communication.
    if (0 < this->stencilWidth()) {
        int dirNeighbour;
        int idNeighbour;

        GridSub<DIM> memFaceCurr;
        GridSub<DIM> memCommCurr;
        GridSub<DIM> memExtendedNeighbour;

        // Loop over all neighbours of the current grid partition.
        int nNeighbours = this->gridGlobal().neighbourId(this->params().myWorld().rank()).size();

        for (int ii = 0; ii < nNeighbours; ++ii) {

            // Create grid of neighbour.
            idNeighbour = this->gridGlobal().neighbourId(this->params().myWorld().rank(), ii);
            dirNeighbour = this->gridGlobal().neighbourDir(this->params().myWorld().rank(), ii);
            GridSub<DIM> memNormalNeighbour(this->gridGlobal().partition(idNeighbour).idxNodeFirstSub(),
                                            this->gridGlobal().partition(idNeighbour).idxNodeLastSub(),
                                            this->memAll());

            // The grid of the data field can be larger / smaller than the one
            // of the domain.
            // ==> The communication grid must be larger / smaller, too.
            memExtendedNeighbour = memNormalNeighbour.extend(dirNeighbour,
                                                             this->stencilWidth());
            memCommCurr = memExtendedNeighbour.intersectWith(this->memNormal());

            for (std::size_t idxFace = 0; idxFace < 2 * DIM; ++idxFace) {
                memFaceCurr = memExtendedNeighbour.intersectWith(this->memBorder().at(idxFace));

                if (memFaceCurr.valid()) {
                    // Check if memFaceCurr is equal to the global
                    // dimension of the data field in at least one direction.
                    // If the data field ends within a gird partition,
                    // then a new grid will be created, BUT:
                    // The corresponding neighbour grid partition would NOT create
                    // a grid!
                    int idx = dirNeighbour / 2;
                    int rem = dirNeighbour % 2;

                    // LEFT, BACK, BOTTOM...
                    if ( (0 == rem) &&
                            (this->gridGlobal().idxNodeFirst().elem(idx)
                             == memCommCurr.idxNodeFirstSub().elem(idx)) ) {
                        std::cout << __FILE__ << " " << __LINE__
                                  << " myRank = " << this->params().myWorld().rank()
                                  << "\t end(df) reached: "
                                  << " this->gridGlobal().idxNodeFirst()= "
                                  << this->gridGlobal().idxNodeFirst()
                                  << " memCommCurr.idxNodeFirstSub= "
                                  << memCommCurr.idxNodeFirstSub() << "\n";
                        break;
                    }
                    // RIGHT, FRONT, TOP...
                    else if ( (1 == rem) &&
                              (this->gridGlobal().idxNodeLast().elem(idx) + 1
                               == memCommCurr.idxNodeLastSub().elem(idx)) ) {
                        std::cout << __FILE__ << " " << __LINE__
                                  << "\t end(df) reached: "
                                  << " this->gridGlobal().idxNodeLast()= "
                                  << this->gridGlobal().idxNodeLast()
                                  << " memCommCurr.idxNodeLastSub= "
                                  << memCommCurr.idxNodeLastSub() << "\n";
                        break;
                    } else {
                        mMemComm.push_back(memCommCurr);
                        mValuesToExchange.push_back(
                            Buffer<CT>(params().myWorld(),
                                       idNeighbour,
                                       memCommCurr.nNodesSub().size(),
                                       true
                                      )
                        );
                        // Create ghost grids.
                        int idx;
                        Index idxNodeFirstGridGhost;
                        Index idxNodeLastGridGhost;
                        idxNodeFirstGridGhost = memCommCurr.idxNodeFirstSub();
                        idxNodeLastGridGhost  = memCommCurr.idxNodeLastSub();

                        idx = dirNeighbour / 2;

                        if (0 == dirNeighbour % 2) {
                            // Left, back, bottom.
                            idxNodeFirstGridGhost[idx] -= this->stencilWidth();
                            idxNodeLastGridGhost[idx]  -= this->stencilWidth();
                        } else {
                            // Right, front, top.
                            idxNodeFirstGridGhost[idx] += this->stencilWidth();
                            idxNodeLastGridGhost[idx]  += this->stencilWidth();
                        }

                        GridSub<DIM> memCurr(idxNodeFirstGridGhost,
                                             idxNodeLastGridGhost,
                                             this->memAll());
                        mMemGhost.push_back(memCurr);
                        break;
                    }
                }
            }
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::DataField(DataField const& from)
    : mParams(from.params())
    , mGG(from.gridGlobal())
    , mMemNormal(from.memNormal())
    , mMemHole(from.memHole())
    , mMemBorder(from.memBorder())
    , mMemComm(from.memComm())
    , mMemGhost(from.memGhost())
    , mStencilWidth(from.stencilWidth())
    , mName(from.name())
    , mMemoryPosFirst(from.mMemoryPosFirst)
    , mMemoryPosLast(from.mMemoryPosLast)
    , mValuesToExchange(from.mValuesToExchange)
    , mOutput(from.mOutput)
{
     int ii;
     mElemData = NULL;
     if (0 < from.memAll().nNodes().size()) {
         mElemData = new CT[from.memAll().nNodes().size()];
     }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
        private(ii) \
        shared(from)
#endif
    for (ii = 0; ii < from.memAll().nNodes().size(); ++ii) {
        mElemData[ii] = from.elemData(ii);
    }
    // memAll has to be set AFTER the for loops.
    mMemAll = from.memAll();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::~DataField()
{
    delete[] mElemData;
    mElemData = NULL;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>& DataField<CT, DIM>::operator=(DataField<CT, DIM> from)
{
    swap(*this, from);
    return *this;
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline std::string const& DataField<CT, DIM>::name() const
{
    return mName;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline Parameters const& DataField<CT, DIM>::params() const
{
    return mParams;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline GridGlobal<DIM> const& DataField<CT, DIM>::gridGlobal() const
{
    return mGG;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline int const& DataField<CT, DIM>::stencilWidth() const
{
    return mStencilWidth;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline Grid<DIM> const& DataField<CT, DIM>::memAll() const
{
    return mMemAll;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline GridSub<DIM> const& DataField<CT, DIM>::memNormal() const
{
    return mMemNormal;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline GridSub<DIM> const& DataField<CT, DIM>::memHole() const
{
    return mMemHole;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline std::vector< GridSub<DIM> > const& DataField<CT, DIM>::memBorder() const
{
    return mMemBorder;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline std::vector< GridSub<DIM> > const& DataField<CT, DIM>::memComm() const
{
    return mMemComm;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline std::vector< GridSub<DIM> > const& DataField<CT, DIM>::memGhost() const
{
    return mMemGhost;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT const& DataField<CT, DIM>::elemData(int const& idx) const
{
    return mElemData[idx];
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT* const& DataField<CT, DIM>::elemData() const
{
    return mElemData;
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::operator()(unsigned long int const& memoryPos) const
{
    return mElemData[memoryPos];
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::operator()(Index const& idxNode) const
{
    return mElemData[idxNode2MemoryPos(idxNode)];
}
/*----------------------------------------------------------------------------*/
// TODO: Only if DIM = 3:
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::operator()(int const& ii,
                                          int const& jj,
                                          int const& kk) const
{
//    Index idxNode(ii, jj, kk);
//    unsigned long int memoryPos = idxNode2MemoryPos(idxNode);
    unsigned long int memoryPos = kk - this->memAll().idxNodeFirst(2);
    memoryPos = memoryPos * this->memAll().nNodes(1) + jj
                 - this->memAll().idxNodeFirst(1);
    memoryPos = memoryPos * this->memAll().nNodes(0) + ii
                 - this->memAll().idxNodeFirst(0);
    checkRange(memoryPos);
    return mElemData[memoryPos];
}
/*----------------------------------------------------------------------------*/
// TODO: Only if DIM = 2:
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::operator()(int const& ii,
                                          int const& jj) const
{
    Index idxNode(ii, jj);
    unsigned long int memoryPos = idxNode2MemoryPos(idxNode);
    checkRange(memoryPos);
    return mElemData[memoryPos];
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::at(unsigned long int const& memoryPos) const
{
    assert(memoryPos <= mMemoryPosLast);
    assert(mMemoryPosFirst <= memoryPos);
    // throw(out-of-range)
    checkRange(memoryPos);
    return mElemData[memoryPos];
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::at(Index const& idxNode) const
{
    assert(idxNode <= this->memAll().idxNodeLast());
    assert(this->memAll().idxNodeFirst() <= idxNode);
    // throw(out-of-range)
    unsigned long int memoryPos = idxNode2MemoryPos(idxNode);
    checkRange(memoryPos);
    return mElemData[memoryPos];
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline CT& DataField<CT, DIM>::at(int const& i,
                                  int const& j,
                                  int const& k) const
{
    Index idxNode(i, j, k);
    assert(this->memAll().idxNodeFirst() <= idxNode);
    assert(idxNode <= this->memAll().idxNodeLast());
    return mElemData[idxNode2MemoryPos(idxNode)];
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline bool DataField<CT, DIM>::operator==(DataField<CT, DIM> const& second)
{
    double eps = 1E-10;
    return hasSameValues(second, eps);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline bool DataField<CT, DIM>::hasSameSize(DataField<CT, DIM> const& rhs) const
{
    if (this->stencilWidth() != rhs.stencilWidth()) {
        return false;
    }

    if (this->memAll() != rhs.memAll()) {
        return false;
    }

    return true;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline bool DataField<CT, DIM>::hasSameValues(DataField<CT, DIM> const& second,
                                              double const& eps) const
{
    if (hasSameSize(second)) {
        if (haveSameValuesFree(*this, second)) {
            return true;
        }

        return true;
    }

    return false;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline void DataField<CT, DIM>::checkRange(unsigned long int const& memoryPos) const
{
    assert(memoryPos <= mMemoryPosLast);
    assert(mMemoryPosFirst <= memoryPos);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void DataField<CT, DIM>::swapWith(DataField<CT, DIM>& df)
{
    std::swap(this->mElemData, df.mElemData);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void DataField<CT, DIM>::operator= (CT const& val)
{
//#pragma omp parallel for
    for (int ii = 0; ii < this->memAll().nNodes().size(); ++ii) {
        mElemData[ii] = val;
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void DataField<CT, DIM>::assignValues(DataField<CT, DIM> const& from)
{
    if (hasSameSize(from)) {
//#pragma omp parallel for
        for (int ii = 0; ii < this->memAll().nNodes().size(); ++ii) {
            mElemData[ii] = from.elemData(ii);
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline unsigned long int DataField<CT, DIM>::idxNode2MemoryPosStyleC(Index const& idxNode) const
{
    unsigned long int memoryPos = idxNode.elem(0)
                                  - this->memAll().idxNodeFirst(0);

    // NOT: #pragma omp parallel for
    // BETTER use SIMD vectorization
    for (std::size_t ii = 1; ii < DIM; ++ii) {
        memoryPos = memoryPos * this->memAll().nNodes(ii)
                    + idxNode.elem(ii) - this->memAll().idxNodeFirst(ii);
    }

    return memoryPos;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Index DataField<CT, DIM>::memoryPos2IdxNodeStyleC(unsigned long int const& memoryPos) const
{
    Index idxNode;
    int tmpOld;
    int tmpNew;
    tmpOld = memoryPos;
    for (std::size_t ii = DIM - 1; ii > 0; --ii) {
        tmpNew = tmpOld / this->memAll().nNodes(ii);
        idxNode[ii] = tmpOld % this->memAll().nNodes(ii)
                              + memAll().idxNodeFirst(ii); ;
        tmpOld = tmpNew;
    }
    idxNode[0] = tmpOld;
    return idxNode;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline unsigned long int DataField<CT, DIM>::idxNode2MemoryPosStyleFortran(Index const& idxNode) const
{
    unsigned long int memoryPos = idxNode.elem(DIM - 1)
                                  - this->memAll().idxNodeFirst(DIM - 1);
    const int dimLocal = DIM;

    // NOT: #pragma omp parallel for
    // BETTER use SIMD vectorization.
    for (int ii = dimLocal - 2; ii >= 0; --ii) {
        memoryPos = memoryPos * this->memAll().nNodes(ii)
                    + idxNode.elem(ii) - this->memAll().idxNodeFirst(ii);
    }

    return memoryPos;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Index DataField<CT, DIM>::memoryPos2IdxNodeStyleFortran(unsigned long int const& memoryPos) const
{
    typename DataField<CT, DIM>::Index idxNode;
    unsigned long int tmp = memoryPos;

    for (std::size_t ii = 0; ii < DIM; ++ii) {
        idxNode[ii] = tmp % memAll().nNodes(ii) + memAll().idxNodeFirst(ii);
        tmp /= this->memAll().nNodes(ii);
    }

    return idxNode;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline unsigned long int DataField<CT, DIM>::idxNode2MemoryPos(Index const& idxNode) const
{
    //return this->idxNode2MemoryPosStyleC(idxNode);
    return this->idxNode2MemoryPosStyleFortran(idxNode);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Index DataField<CT, DIM>::memoryPos2IdxNode(unsigned long int const& memoryPos) const
{
    //return this->memoryPos2IdxNodeStyleC(memoryPos);
    return this->memoryPos2IdxNodeStyleFortran(memoryPos);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline std::size_t DataField<CT, DIM>::elementSize() const
{
    // Cast from size_t to int.
    return sizeof(CT);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>& DataField<CT, DIM>::operator+=(DataField<CT, DIM> const& rhs)
{
    unsigned long int idxScalarNode;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = mMemNormal.begin(),
            et = mMemNormal.end(); it < et; ++it) {
        idxScalarNode = it.idxScalarNode();
        mElemData[idxScalarNode] += rhs.elemData(idxScalarNode);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>& DataField<CT, DIM>::operator-=(DataField<CT, DIM> const& rhs)
{
    unsigned long int idxScalarNode;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = mMemNormal.begin(),
            et = mMemNormal.end(); it < et; ++it) {
        idxScalarNode = it.idxScalarNode();
        mElemData[idxScalarNode] -= rhs.elemData(idxScalarNode);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::iterator DataField<CT, DIM>::begin() const
{
    return iterator(&(this->memAll()), this->memAll().idxNodeFirst());
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::iterator DataField<CT, DIM>::end() const
{
    return iterator(&(this->memAll()), this->memAll().idxNodeLast());
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline double DataField<CT, DIM>::normLinf() const
{
    double res = 0.0;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = mMemNormal.begin(),
            et = mMemNormal.end(); it < et; ++it) {
        if (res < fabs(elemData(it.idxScalarNode()))) {
            res = elemData(it.idxScalarNode());
        }
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename PT, typename TT>
inline void DataField<CT, DIM>::evalAtGrid(PT& prblm,
                                           void (PT::*fptrEvalFunc)(CT&,
                                                   Ntuple<TT, DIM> const&,
                                                   TT const&),
                                           int const& timestep)
{
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Evaluate data field " << this->name()
                  << " at grid at timestep " << timestep << "..."
                  << std::endl;
    }
    mParams.decreaseLevel();

    Index idxNode;
    Ntuple<TT, DIM> currCoord;
    CT evalFunc;
    TT currTime = params().timeIntervalStart() + params().tau() * timestep;

    //#pragma omp parallel for
    for (typename Grid<DIM>::iterator it = memAll().begin(),
            et = memAll().end(); it < et; ++it) {
        idxNode = it.idxNode();
        currCoord = this->gridGlobal().coordinates(idxNode);
        (prblm.*fptrEvalFunc)(evalFunc, currCoord, currTime);
        mElemData[it.idxScalarNode()] = evalFunc;
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Evaluated."
                  << std::endl;
    }

    sync(timestep);
}
/*----------------------------------------------------------------------------*/
// This covers the cases: CT=Ntuple<TT,DIM>, CT=TT, CT=Complex<TT>,
// especially: CT=adouble, TT=adouble
template<typename CT, std::size_t DIM>
template<typename TT>
inline void DataField<CT, DIM>::evalAtGrid(void (*fptrEvalFunc)(CT&,
                                           Ntuple<TT, DIM> const&,
                                           TT const&),
                                           int const& timestep)
{
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Evaluate data field " << this->name()
                  << " at grid at timestep " << timestep << "..."
                  << std::endl;
    }
    mParams.decreaseLevel();

    Index idxNode;
    Ntuple<TT, DIM> currCoord;
    CT evalFunc;
    TT currTime = params().timeIntervalStart() + params().tau() * timestep;

//#pragma omp parallel for
    for (typename Grid<DIM>::iterator it = memAll().begin(),
            et = memAll().end(); it < et; ++it) {
        idxNode = it.idxNode();
        currCoord = this->gridGlobal().coordinates(idxNode);
        (*fptrEvalFunc)(evalFunc, currCoord, currTime);
        mElemData[it.idxScalarNode()] = evalFunc;
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Evaluated."
                  << std::endl;
    }

    sync(timestep);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
void DataField<CT, DIM>::check(DataField<CT, DIM> const& df)
{
    if (haveSameValuesFree(*this, df)) {
        if (0 == df.params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << ": Check of (" << this->name() << ", "
                      <<  df.name() << ")  SUCCESSFUL."
                      << std::endl;
        }
    } else {
        if (0 == df.params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << ": Check of (" << this->name() << ", "
                      <<  df.name() << ")  FAILED."
                      << std::endl;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
void DataField<CT, DIM>::compErrLinf(DataField<CT, DIM> const& df)
{
    if ((0 == params().myWorld().rank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Compute Linf error(" << name()
                  << ", " << df.name() << ")... " << std::endl;
    }
    mParams.decreaseLevel();
    double errLinf = compErrLinfFree(*this, df);
    if ((0 == params().myWorld().rank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Linf error(" << name()
                  << ", " << df.name() << "): "
                  << std::scientific << errLinf << std::fixed << std::endl;
    }
    mParams.increaseLevel();
    if ((0 == params().myWorld().rank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline unsigned long int DataField<CT, DIM>::idxNode2Buffer(Index const& idxNode,
        Grid<DIM> const& gg) const
{
    const int dimLocal = DIM;
    unsigned long int posBuffer = idxNode.elem(DIM - 1) - gg.idxNodeFirst(DIM - 1);

    for (int ii = dimLocal - 2; ii >= 0; --ii) {
        posBuffer = posBuffer * gg.nNodes(ii)
                    + idxNode.elem(ii) - gg.idxNodeFirst(ii);
    }

    return posBuffer;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::exchangeValuesInBuffers(int const& timestep)
{
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Exchange buffer values of data field " << this->name()
                  << "..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    for (std::size_t ii = 0; ii < mValuesToExchange.size(); ++ii) {
        mValuesToExchange[ii].sendValues();
        mValuesToExchange[ii].receiveValues();
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Exchanged." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::waitAll()
{
    for (std::size_t ii = 0; ii < mValuesToExchange.size(); ++ii) {
        mValuesToExchange[ii].waitAll();
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::copyValuesFromMemCommToSendBuffer(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::copyValuesFromMemCommToSendBuffer()");
#endif
    unsigned long int posBuffer;

    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Copy values from mem comm. to send buffer of data field "
                  << this->name() << "..." << std::endl;
    }
    mParams.decreaseLevel();
//#pragma omp parallel for
    for (std::size_t ii = 0; ii < memComm().size(); ++ii) {
        posBuffer = 0;

        for (typename GridSub<DIM>::iterator it = memComm().at(ii).begin(),
                et = memComm().at(ii).end(); it < et; ++it) {
            mValuesToExchange.at(ii).setElemData(posBuffer,
                                                 this->elemData(it.idxScalarNode()));
            ++posBuffer;
        }
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Copied." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::copyValuesFromMemGhostToSendBuffer(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::copyValuesFromMemGhostToSendBuffer()");
#endif
    unsigned long int posBuffer;

    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Copy values from mem ghost to send buffer of data field "
                  << this->name() << "..." << std::endl;
    }
    mParams.decreaseLevel();
//#pragma omp parallel for
    for (std::size_t ii = 0; ii < memGhost().size(); ++ii) {
        posBuffer = 0;

        for (typename GridSub<DIM>::iterator it = memGhost().at(ii).begin(),
                et = memGhost().at(ii).end(); it < et; ++it) {
            mValuesToExchange.at(ii).setElemData(posBuffer,
                                                 this->elemData(it.idxScalarNode()));
            ++posBuffer;
        }
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Copied." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::copyValuesFromReceiveBufferToMemGhost(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::copyValuesFromReceiveBufferToMemGhost()");
#endif
    unsigned long int posBuffer;

    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Copy values from receive buffer to mem ghost of data field "
                  << this->name() << "..." << std::endl;
    }
    mParams.decreaseLevel();

//#pragma omp parallel for
    for (std::size_t ii = 0; ii < memGhost().size(); ++ii) {
        posBuffer = 0;

        for (typename GridSub<DIM>::iterator it = memGhost().at(ii).begin(),
                et = memGhost().at(ii).end(); it < et; ++it) {
            this->mElemData[it.idxScalarNode()]
                = mValuesToExchange.at(ii).elemData(posBuffer);
            ++posBuffer;
        }
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Copied." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::copyValuesFromReceiveBufferToMemComm(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::copyValuesFromReceiveBufferToMemComm()");
#endif
    unsigned long int posBuffer;

    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Copy values from receive buffer to mem comm. of data field "
                  << this->name() << "..." << std::endl;
    }
    mParams.decreaseLevel();

//#pragma omp parallel for
    for (std::size_t ii = 0; ii < memComm().size(); ++ii) {
        posBuffer = 0;

        for (typename GridSub<DIM>::iterator it = memComm().at(ii).begin(),
                et = memComm().at(ii).end(); it < et; ++it) {
            this->mElemData[it.idxScalarNode()]
                = mValuesToExchange.at(ii).elemData(posBuffer);
            ++posBuffer;
        }
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Copied." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::copyAndSumValuesFromReceiveBufferToMemComm(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::copyAndSumValuesFromReceiveBufferToMemComm()");
#endif
    unsigned long int posBuffer;

    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << " * Copy and sum values from receive buffer to mem comm. of data field "
                  << this->name() << "..." << std::endl;
    }
    mParams.decreaseLevel();

//#pragma omp parallel for
    for (std::size_t ii = 0; ii < memComm().size(); ++ii) {
        posBuffer = 0;

        for (typename GridSub<DIM>::iterator it = memComm().at(ii).begin(),
                et = memComm().at(ii).end(); it < et; ++it) {
            this->mElemData[it.idxScalarNode()]
                += mValuesToExchange.at(ii).elemData(posBuffer);
            ++posBuffer;
        }
    }
    mParams.increaseLevel();
    if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                  << "   Copied." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::sync(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::sync()");
#endif

    // For the following special cases, no communication is necessary.
    if (1 == this->mParams.myWorld().size()) {
        return;
    }

    if (0 == this->stencilWidth()) {
        return;
    }

    copyValuesFromMemCommToSendBuffer(timestep);
    exchangeValuesInBuffers(timestep);
    waitAll();
    copyValuesFromReceiveBufferToMemGhost(timestep);
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::collectValuesAtMemComm(int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("Df::collectValuesAtMemComm()");
#endif

    // For the following special cases, no communication is necessary.
    if (1 == this->mParams.myWorld().size()) {
        return;
    }

    if (0 == this->stencilWidth()) {
        return;
    }

    copyValuesFromMemGhostToSendBuffer(timestep);
    exchangeValuesInBuffers(timestep);
    waitAll();
    copyAndSumValuesFromReceiveBufferToMemComm(timestep);
}

/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::write(int const& timestep)
{
    // Check if file should be written at current time step.
    int intervalOfWrites = this->params().nTimesteps();

    if (0 < this->params().nSnapshots()) {
        if (this->params().nSnapshots() <= this->params().nTimesteps()) {
            intervalOfWrites = this->params().nTimesteps()
                               / this->params().nSnapshots();
        }
    }

    std::vector<TT*> tmpElemData;
    tmpElemData.push_back(this->elemData());
    if (0 == timestep % intervalOfWrites) {
        if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << " * Write data field " << name()
                      << " to NetCDF file..." << std::endl;
        }
        mParams.decreaseLevel();
        mOutput.write(tmpElemData, this->memNormal(), timestep);
        mParams.increaseLevel();
        if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << "   Written." << std::endl;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataField<TT, DIM>::read(int const& timestep)
{
    // Check if file should be written at current time step.
    int intervalOfWrites = this->params().nTimesteps();

    if (0 < this->params().nSnapshots()) {
        if (this->params().nSnapshots() <= this->params().nTimesteps()) {
            intervalOfWrites = this->params().nTimesteps()
                               / this->params().nSnapshots();
        }
    }

    std::vector<TT*> tmpElemData;
    tmpElemData.push_back(this->elemData());
    if (0 == timestep % intervalOfWrites) {
        if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << ": * Read data field " << name()
                      << "from NetCDF file..." << std::endl;
        }
        mParams.decreaseLevel();
        mOutput.read(tmpElemData, this->memNormal(), timestep);
        mParams.increaseLevel();
        if (0 == params().myWorld().rank() && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), params().myWorld().rank())
                      << ":   Read." << std::endl;
        }
    }
}

// /*----------------------------------------------------------------------------*/
// template<typename TT, std::size_t DIM>
// template<class Archive>
// inline void DataField<TT, DIM>::serialize(Archive& ar,
//                                           unsigned int const version)
// {
// //    if (1 <= version) {
//         ar & mParams;
//         ar & mGG;
//         ar & mMemAll;
//         ar & mMemNormal;
//         ar & mStencilWidth;
//         ar & mMemBorder;
//         ar & mMemComm;
//         ar & mMemGhost;
//         ar & mName;
//             std::cout << "memAllnodes = " << this->memAll().nNodes().size()
//             <<std::endl;
//             std::cout << "mElemData = " << mElemData << std::endl;
//         for (int ii = 0; ii < this->memAll().nNodes().size(); ++ii) {
//             ar & mElemData[ii];
//         }
//         ar & mMemoryPosFirst;
//         ar & mMemoryPosLast;
//         ar & mValuesToExchange;
//         ar & mOutput;
// //    }
// }

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline DataField<CT, DIM>::Iterator::Iterator(const Grid<DIM>* gd,
                                              Index const& currIdxNode)
    : mGridIter(gd)
    , mIdxNode(currIdxNode)
    , mMemoryPos(gd->idxNode2IdxScalarNode(currIdxNode))
{
    // TODO: Check if \c currIdxNode is inside underlying grid.
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: GETTER METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Index const& DataField<CT, DIM>::Iterator::idxNode() const
{
    return mIdxNode;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline unsigned long int const& DataField<CT, DIM>::Iterator::memoryPos() const
{
    return mMemoryPos;
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: COMPARISON METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline bool DataField<CT, DIM>::Iterator::operator!=(const Iterator& other) const
{
    return (*this < other) || (other < *this);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline bool DataField<CT, DIM>::Iterator::operator<(const Iterator& it) const
{
    return (memoryPos() <= it.memoryPos());
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: WORK METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Index const& DataField<CT, DIM>::Iterator::operator*()
{
    return idxNode();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Iterator& DataField<CT, DIM>::Iterator::nextStyleC()
{
    ++(mMemoryPos);
    ++(mIdxNode[DIM - 1]);

    for (std::size_t ii = DIM - 1; ii > 0; --ii) {
        if (mIdxNode.elem(ii) > this->mGridIter->idxNodeLast(ii)) {
            mIdxNode[ii] = this->mGridIter->idxNodeFirst(ii);
            ++(mIdxNode[ii - 1]);
        }
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Iterator& DataField<CT, DIM>::Iterator::nextStyleFortran()
{
    ++(mMemoryPos);
    ++(mIdxNode[0]);

    for (std::size_t ii = 0; ii < DIM - 1; ++ii) {
        if (mIdxNode.elem(ii) > this->mGridIter->idxNodeLast(ii)) {
            mIdxNode[ii] = this->mGridIter->idxNodeFirst(ii);
            ++(mIdxNode[ii + 1]);
        }
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Iterator& DataField<CT, DIM>::Iterator::operator++()
{
    if (mGridIter->isNumberedStyleC()) {
        return this->nextStyleC();
    }

    return this->nextStyleFortran();
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline typename DataField<CT, DIM>::Iterator DataField<CT, DIM>::Iterator::operator++(int)
{
    Iterator tmp = *this;
    ++(*this);
    return tmp;
}

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline void swap(DataField<CT, DIM>& first, DataField<CT, DIM>& second)
{
    ScaFES::swap(first.mParams, second.mParams);
    ScaFES::swap(first.mGG, second.mGG);
    ScaFES::swap(first.mMemAll, second.mMemAll);
    ScaFES::swap(first.mMemNormal, second.mMemNormal);
    ScaFES::swap(first.mMemHole, second.mMemHole);
    std::swap(first.mStencilWidth, second.mStencilWidth);
    std::swap(first.mMemBorder, second.mMemBorder);
    std::swap(first.mMemComm, second.mMemComm);
    std::swap(first.mMemGhost, second.mMemGhost);
    std::swap(first.mName, second.mName);
    //std::swap(first.mElemData, second.mElemData);
    //swapDataFields(first, second);
    CT* tmp2;
    tmp2 = first.mElemData;
    first.mElemData = second.mElemData;
    second.mElemData = tmp2;
    std::swap(first.mMemoryPosLast, second.mMemoryPosLast);
    std::swap(first.mMemoryPosFirst, second.mMemoryPosFirst);
    std::swap(first.mValuesToExchange, second.mValuesToExchange);
    ScaFES::swap(first.mOutput, second.mOutput);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline std::ostream& operator<<(std::ostream& output,
                                DataField<CT, DIM> const& df)
{
//#pragma omp parallel for
    for (typename Grid<DIM>::iterator it = df.memAll().begin(),
            et = df.memAll().end(); it < et; ++it) {
        output << it.idxScalarNode() << "   "
               << it.idxNode() << "   "
               << df(it.idxNode()) << std::endl;
    }

    return output;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline double compErrLinfFree(DataField<double, DIM> const& df1,
                              DataField<double, DIM> const& df2)
{
    typename DataField<double, DIM>::Index idxNode;
    Ntuple<int, DIM> idxNodeFirstSub = df1.gridGlobal().partition(df1.params().myWorld().rank()).idxNodeFirstSub();
    Ntuple<int, DIM> idxNodeLastSub = df1.gridGlobal().partition(df1.params().myWorld().rank()).idxNodeLastSub();

    double errLinf = 0.0;
    bool isNormal = true;

    //Ntuple<int, DIM> idxNode;
    int nNodesTotal = df1.memAll().nNodes().size();
    int ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * df1 READ access --> shared
    // * df2 READ access --> shared
    // Parallelize the region, assign portions of the loop to each thread 
    // and compute for each thread a local maximum. 
    // At the end, compute the global maximum of all threads.
//#pragma omp parallel default(none) private(ii, idxNode) shared(df1, df2, nNodesTotal, errLinf)
//{
//    double tmpMax = 0.0;
//#pragma omp for nowait schedule(static)
//    for (ii = 0; ii < nNodesTotal; ++ii) {
//        idxNode = df1.memoryPos2IdxNode(ii);
//        tmpMax = std::max(::fabs(df1(idxNode) - df2(idxNode)), tmpMax);
//    }       
//#pragma omp critical
//   {
//       errLinf = std::max(errLinf, tmpMax);
//   }
//}

//#pragma omp parallel for schedule(static) private(ii, idxNode) shared(df1, df2, nNodesTotal, idxNodeFirstGlobal, idxNodeLastGlobal) reduction(max: errLinf)
//    for (ii = 0; ii < nNodesTotal; ++ii) {
//        idxNode = df1.memoryPos2IdxNode(ii);
//        for (std::size_t jj = 0; jj < DIM; ++jj) {
//            if ( (idxNodeFirstGlobal.elem(jj) < idxNode.elem(jj))
//                || (idxNodeLastGlobal.elem(jj) > idxNode.elem(jj)) ) {
//                continue;
//            }
//        }

//        if (::fabs(df1(idxNode) - df2(idxNode)) > errLinf) {
//            errLinf = ::fabs(df1(idxNode) - df2(idxNode));
//        }
//    }


//    double tmpDiff = 0.0;
//    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
//            et = df1.memNormal().end(); it < et; ++it) {
//        idxNode = it.idxNode();
//        tmpDiff = ::fabs(df1(idxNode) - df2(idxNode));
//
//        if (tmpDiff > errLinf) {
//           errLinf = tmpDiff;
//        }
//    }

#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    private(ii, idxNode, isNormal) \
    shared(df1, df2, nNodesTotal, idxNodeFirstSub, idxNodeLastSub) \
    reduction(max: errLinf)
#endif
    for (ii = 0; ii < nNodesTotal; ++ii) {
        idxNode = df1.memoryPos2IdxNode(ii);
        isNormal = true;
        for (std::size_t jj = 0; jj < DIM; ++jj) {
            if ( (idxNodeFirstSub.elem(jj) > idxNode.elem(jj))
                || (idxNodeLastSub.elem(jj) < idxNode.elem(jj)) ) {
                isNormal = false;
            }
        }

        if (isNormal) {
            if (::fabs(df1(idxNode) - df2(idxNode)) > errLinf) {
                errLinf = ::fabs(df1(idxNode) - df2(idxNode));
            }
        }
    }
    return errLinf;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline double compErrLinfFree(DataField<Complex<double>, DIM> const& df1,
                              DataField<Complex<double>, DIM> const& df2)
{
    typename DataField<Complex<double>, DIM>::Index idxNode;
    double errLinf = 0.0;
    double tmpDiff = 0.0;

//     /*---------------------------------------------------------------------*/
//     typename GridSub<DIM>::iterator it, et;
//     int iSize = df1.memNormal().nNodesTotal();
//     std::valarray<const Ntuple<int,DIM>*> items(iSize);
//     int k = 0;
//     for (it = df1.memNormal().begin(),
//         et = df1.memNormal().end(); it < et; ++it) {
//         items[k] = &(it.idxNode());
//         ++k;
//     }
// #pragma omp parallel for default(shared)
//     for (int ii = 0; ii < iSize; ++ii) {
//         idxNode = *(items[ii]);
//         tmpDiff = ::fabs(df1(idxNode) - df2(idxNode));
//         if (tmpDiff > errLinf) {
//              errLinf = tmpDiff;
//         }
//     }
//
//     /*---------------------------------------------------------------------*/
//     typename GridSub<DIM>::iterator it, et;
// #pragma omp parallel default(shared) private(it, et)
// {
//     for (it = df1.memNormal().begin(),
//          et = df1.memNormal().end(); it < et; ++it) {
// #pragma omp single nowait
// {
//         idxNode = it.idxNode();
// }
//         tmpDiff = ::fabs(df1(idxNode) - df2(idxNode));
//         if (tmpDiff > errLinf) {
//             errLinf = tmpDiff;
//         }
//     }
// }
//

    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
            et = df1.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();
        tmpDiff = fabs(df1(idxNode) - df2(idxNode));

        if (tmpDiff > errLinf) {
            errLinf = tmpDiff;
        }
    }
    return errLinf;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline double compErrLinfFree(DataField<Ntuple<double, DIM>, DIM> const& df1,
                              DataField<Ntuple<double, DIM>, DIM> const& df2)
{
    typename DataField<Ntuple<double, DIM>, DIM>::Index idxNode;
    double errLinf = 0.0;
    double tmpDiff = 0.0;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
            et = df1.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();
        tmpDiff = fabs(df1(idxNode) - df2(idxNode));

        if (tmpDiff > errLinf) {
            errLinf = tmpDiff;
        }
    }
    return errLinf;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool haveSameValuesFree(DataField<double, DIM> const& df1,
                               DataField<double, DIM> const& df2)
{
    bool isSame = true;
    typename DataField<double, DIM>::Index idxNode;
    double tmpDiff = 0.0;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
            et = df1.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();
        tmpDiff = ::fabs(df1(idxNode) - df2(idxNode));

        if (tmpDiff > EPSILON) {
            isSame = false;
            break;
        }
    }

    return isSame;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool haveSameValuesFree(DataField<Complex<double>, DIM> const& df1,
                               DataField<Complex<double>, DIM> const& df2)
{
    bool isSame = true;
    typename DataField<Complex<double>, DIM>::Index idxNode;
    double tmpDiff = 0.0;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
            et = df1.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();
        tmpDiff = fabs(df1(idxNode) - df2(idxNode));

        if (tmpDiff > EPSILON) {
            isSame = false;
            break;
        }
    }

    return isSame;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool haveSameValuesFree(DataField<Ntuple<double, DIM>, DIM> const& df1,
                               DataField<Ntuple<double, DIM>, DIM> const& df2)
{
    bool isSame = true;
    typename DataField<Ntuple<double, DIM>, DIM>::Index idxNode;
    double tmpDiff = 0.0;

//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = df1.memNormal().begin(),
            et = df1.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();
        tmpDiff = fabs(df1(idxNode) - df2(idxNode));

        if (tmpDiff > EPSILON) {
            isSame = false;
            break;
        }
    }

    return isSame;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline const DataField<CT, DIM> operator+(DataField<CT, DIM> const& lhs,
                                          DataField<CT, DIM> const& rhs)
{
    DataField<CT, DIM> tmp(lhs); // Creates a copy of the left operand.
    tmp += rhs; // Implementation via the += operator.
    return tmp;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline const DataField<CT, DIM> operator-(DataField<CT, DIM> const& lhs,
                                          DataField<CT, DIM> const& rhs)
{
    DataField<CT, DIM> tmp(lhs); // Creates a copy of the left operand.
    tmp -= rhs; // Implementation via the -= operator.
    return tmp;
}

} // End of namespace. //


/*******************************************************************************
 ******************************************************************************/
// namespace boost
// {
// namespace serialization
// {
// template<typename CT, std::size_t DIM>
// struct version< ScaFES::DataField<CT, DIM> > {
//     BOOST_STATIC_CONSTANT(unsigned long int, value = 2);
// };
// } // namespace serialization
// } // namespace boost

#endif
