/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_DATAFILEAPI_HPP_
#define SCAFES_DATAFILEAPI_HPP_

#include "ScaFES_Config.hpp"

#ifdef SCAFES_HAVE_NETCDF_PAR
#include <netcdf_par.h>
#else
#include <netcdf.h>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

#include <iostream>
#include <vector>

#include <boost/type_traits.hpp>

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_GridSub.hpp"
#include "ScaFES_Complex.hpp"
#include "ScaFES_Complex_FreeFunctions.hpp"

namespace ScaFES
{

/** Version number of data file. */
const int DATA_FILE_VERSION = 3;

/*******************************************************************************
 ******************************************************************************/
/** The class \c DataFile is responsible for writing (per default: in parallel)
 * a data field to a given file in the NetCDf output format.
 *
 * \n \remarks The call to the NetCDF creation of the data file is not set
 * within the constructor.
 * Instead, it will be created if the user calls the method \c write() the
 * first time. Thus, in case that no data will be written to the output
 * (using the method \c write()), then no data file will be created, too.

 * The class \c DataFile provides the interfaces for the visualisation.
 * The library netCDF will be coupled to ScaFES.
 * The user can decide during the configure process of ScaFEs if a serial of
 * parallel version of NetCDF will be used.
 * The name of the file will be given via command line options.
 *
 * The user can add data fields and parameters to the file via the method
 * \c addDataField2OutputFile and  \c addParameter2OutputFile.
 * \n The command \code ncdump -h <netcdffile>\endcode provides some information
 * aobut the stored data fields and parameters.
 * \code login@host:~/> ncdump -h out.nc
   dimensions:
        Nx = 130 ;
        Ny = 400 ;
        Nz = 200 ;
        time = UNLIMITED ; // (52 currently)
   variables:
        int version ;
        double D_0(time, Nz, Ny, Nx) ;
        double D_1(time, Nz, Ny, Nx) ;
        double D_2(time, Nz, Ny, Nx) ;
        int writeSteps ;
                writeSteps:_FillValue = 10 ;
        int timesteps ;
                timesteps:_FillValue = 500 ;
        double tau ;
                tau:_FillValue = 1.e-07 ;
  \endcode
*/
template<typename TT, std::size_t DIM>
class DataFile
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor. */
        DataFile();

        /** Creates own constructor.
         *  \remarks This constructor only sets the names of the file and the
         *  variable which should be written, the file itself WILL NOT be
         *  created here. */
        DataFile(std::string const& nameDataFile,
                 std::vector<std::string> const& nameVariables,
                 boost::mpi::communicator const& myWorld,
                 Index const& nNodes);

        /** Creates own constructor:
         * The file itself WILL be created here, too. */
        DataFile(std::string const& nameDataFile,
                 std::vector<std::string> const& nameVariables,
                 boost::mpi::communicator const& myWorld,
                 Index const& nNodes,
                 GridSub<DIM> const& memNormal);

        /** Creates own copy constructor. */
        DataFile(DataFile<TT, DIM> const&);

        /** Creates own assignment operator. */
        DataFile& operator= (DataFile<TT, DIM> rhs);

        /** Creates own destructor: Closes the file and releases the memory. */
        ~DataFile();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the name of the data file. */
        std::string const& nameDataFile() const;

        /** Returns the name of the variable which should be written. */
        std::vector<std::string> const& nameVariables() const;

        /** Returns the identifier of the file. */
        int const& idFile() const;

        /** Returns the identifiers of the components of the variable. */
        std::vector< std::vector<int> > const& idVariable() const;

        /** Returns if the file is already open or not. */
        bool const& isOpen() const;

        /** Returns the number the data field was written to the file. */
        int const& nWritesCurr() const;

        /** Returns the memory grid. */
        GridSub<DIM> const& memNormal() const;

        /** Returns a pointer to the data which should be written. */
        std::vector<TT*> const& elemData() const;

        /** Returns the data for a given memory position. */
        TT const& elemData(int const& df, int const& idx) const;

        /** Returns the number of elements stored in the memory \c mNcMemory. */
        int const& nElemsNcMem() const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the data file and sets the file layout of the
         *  underlying data field (depending on the template type). */
        void create(GridSub<DIM> const& memNormal);

        /** Writes the elements in the memory to which the pointer points to
         *  the file at the given time step. */
        void write(std::vector<TT*> const& elemData, 
                   GridSub<DIM> const& memNormal,
                   int const& timestep);

        /** Reads the elements in the memory to which the pointer points to
         *  from the file at the given time step. */
        void read(std::vector<TT*> & elemData, 
                  GridSub<DIM> const& memNormal,
                  int const& timestep);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Swaps two objects of the class \c DataFile. */
        template<typename SS, std::size_t DD>
        friend void swap(DataFile<SS, DD>& first,
                         DataFile<SS, DD>& second);

        /** Reads the data field (type \c short int) from the data file. */
        template<std::size_t DD>
        friend void readFree(DataFile<short int, DD>& output);

        /** Reads the data field (type \c double) from the data file. */
        template<std::size_t DD>
        friend void readFree(DataFile<double, DD>& output);

        /** Writes the data field (type \c short) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<short int, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c double) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<double, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c Complex<double>) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<Complex<double>, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c Ntuple) to the data file.*/
        template<std::size_t DD>
        friend void writeFree(DataFile<Ntuple<double, DD>, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Name of NetCDF data file. */
        std::string mNameDataFile;

        /** Name of the data field which should be written to the file. */
        std::vector<std::string> mNameVariables;

        /** MPI communicator. */
        boost::mpi::communicator mMyWorld;

        /** Number of nodes in each dimension. */
        Index mNnodes;

        /** Identifier of a NetCDF dataset.
         * "A netCDF ID that can subsequently be used to refer to the netCDF
         * dataset in other netCDF function calls."
         * [http://www.unidata.ucar.edu/software/netcdf/docs/netcdf-c.html#nc_005fcreate]
         */
        int mIdFile;

        /** Every variable has a unique identifier. */
        std::vector< std::vector<int> > mIdVariable;

        /** Identifiers of space and time dimensions
         * = idNetCDF + nameDimension + lengthDimension. */
        int mIdDims[DIM + 1];

        /** Is file open? */
        bool mIsOpen;

        /** Number of writes. */
        int mNwritesCurr;

        /** Memory grid. */
        GridSub<DIM> mMemNormal;

        /** Pointer to the memory. */
        std::vector<TT*> mElemData;

        /** Number of elements which will be allocated. */
        int mNelemsNcMem;

        /** Linear continuous memory grid for the elements of a data field.
         * \remarks A data field can only store values of one type.
         * Moreover, the type remains constant during the life time
         * of the data field.
         */
        double* mNcMemory;

}; // End of class. //

/*******************************************************************************
 * DECLARATION OF FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c DataFile. */
template<typename TT, std::size_t DIM>
void swap(DataFile<TT, DIM>& first, DataFile<TT, DIM>& second);
/*----------------------------------------------------------------------------*/
/** Reads the data field (type \c short int) from the data file. */
template<std::size_t DIM>
void readFree(DataFile<short int, DIM>& output);
/*----------------------------------------------------------------------------*/
/** Reads the data field (type \c double) from the data file. */
template<std::size_t DIM>
void readFree(DataFile<double, DIM>& output);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c short) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<short int, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c double) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<double, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c Complex<double>) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<Complex<double>, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c Ntuple<double,DIM>) to the data file.*/
template<std::size_t DIM>
void writeFree(DataFile<Ntuple<double, DIM>, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Free function that handles all errors in conjunction with NetCDF.
 * \param status Return value of NetCDF method.
 * \param file Name of file where error occurs.
 * \param line Number of line where error occurs.
 */
void handle_netCDF_error(int const& status, std::string const& file,
                         int const& line);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile()
    : mNameDataFile("EMPTYFILENAME")
    , mNameVariables()
    , mMyWorld()
    , mNnodes(0)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal()
    , mElemData()
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(std::string const& nameDataFile,
                                   std::vector< std::string> const& nameVariables,
                                   boost::mpi::communicator const& myWorld,
                                   Index const& nNodes)
    : mNameDataFile()
    , mNameVariables(nameVariables)
    , mMyWorld(myWorld)
    , mNnodes(nNodes)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal()
    , mElemData()
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{
    // Create name of the data file and the name of the partition file
    // including the dimensions of the global grid.
    std::ostringstream tmpStringstream;
    tmpStringstream << nameDataFile
                    << "_" << (nNodes.elem(0));

    for (std::size_t ii = 1; ii < DIM; ++ii) {
        tmpStringstream << "x" <<  (nNodes.elem(ii)) ;
    }

    tmpStringstream << "_" << nameVariables[0] << ".nc";
    mNameDataFile = tmpStringstream.str();
    mIdVariable.resize(nameVariables.size());
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(std::string const& nameDataFile,
                                   std::vector<std::string> const& nameVariables,
                                   boost::mpi::communicator const& myWorld,
                                   Index const& nNodes,
                                   GridSub<DIM> const& memNormal)
    : mNameDataFile()
    , mNameVariables(nameVariables)
    , mMyWorld(myWorld)
    , mNnodes(nNodes)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal(memNormal)
    , mElemData(elemData)
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{
    // Create name of the data file and the name of the partition file
    // including the dimensions of the global grid.
    std::ostringstream tmpStringstream;
    tmpStringstream << nameDataFile
                    << "_" << (nNodes.elem(0));

    for (std::size_t ii = 1; ii < DIM; ++ii) {
        tmpStringstream << "x" <<  (nNodes.elem(ii)) ;
    }

    tmpStringstream << "_" << nameVariables[0] << ".nc";
    mNameDataFile = tmpStringstream.str();
    mIdVariable.resize(nameVariables.size());
    // Create the data file using the NetCDF methods.
    create(memNormal);
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(DataFile<TT, DIM> const& rhs)
    : mNameDataFile(rhs.nameDataFile())
    , mNameVariables(rhs.nameVariables())
    , mMyWorld(rhs.mMyWorld)
    , mNnodes(rhs.mNnodes)
    , mIdFile(rhs.idFile())
    , mIdVariable(rhs.idVariable())
    , mIsOpen(rhs.isOpen())
    , mNwritesCurr(rhs.nWritesCurr())
    , mMemNormal(rhs.memNormal())
    , mElemData(rhs.elemData())
{
    for (std::size_t ii = 0; ii < DIM + 1; ++ii) {
        mIdDims[ii] = rhs.mIdDims[ii];
    }

    mNcMemory = 0;
    mNelemsNcMem = 0;
    if (0 < rhs.mNelemsNcMem) {
        mNcMemory = new double[rhs.mNelemsNcMem];
    }
    for (int ii = 0; ii < rhs.mNelemsNcMem; ++ii) {
        mNcMemory[ii] = rhs.mNcMemory[ii];
    }
    // mNelemsNcMem has to be set AFTER the for loops.
    mNelemsNcMem = rhs.mNelemsNcMem;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::~DataFile()
{
    if (0 < mIdFile) {
        if (int status = nc_close(mIdFile)) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
    }

    delete[] mNcMemory;
    mNcMemory = NULL;
}

/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>& DataFile<TT, DIM>::operator=(DataFile<TT, DIM> rhs)
{
    swap(*this, rhs);
    return *this;
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline std::string const& DataFile<TT, DIM>::nameDataFile() const
{
    return mNameDataFile;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector<std::string> const& DataFile<TT, DIM>::nameVariables() const
{
    return mNameVariables;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::idFile() const
{
    return mIdFile;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector< std::vector<int> > const& DataFile<TT, DIM>::idVariable() const
{
    return mIdVariable;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline bool const& DataFile<TT, DIM>::isOpen() const
{
    return mIsOpen;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::nWritesCurr() const
{
    return mNwritesCurr;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline GridSub<DIM> const& DataFile<TT, DIM>::memNormal() const
{
    return mMemNormal;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector<TT*> const& DataFile<TT, DIM>::elemData() const
{
    return mElemData;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline TT const& DataFile<TT, DIM>::elemData(int const& jj, 
                                             int const& idx) const
{
    return mElemData[jj][idx];
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::nElemsNcMem() const
{
    return mNelemsNcMem;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::create(GridSub<DIM> const& memNormal)
{
#ifdef VTRACE
    VT_TRACER("DataFile::create()");
#endif
    /*------------------------------------------------------------------------*/
    int version = DATA_FILE_VERSION;
    int versionID;
    int status; // Return value of netCDF methods.
    // Create netCDF dataset.
#ifndef SCAFES_HAVE_NETCDF_PAR
    status = nc_create(nameDataFile().c_str(), NC_FORMAT_NETCDF4,
                       &mIdFile);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

#else
    // Flag NC_NETCDF4 specifies that HDF5 library is used for parallel I/O //
    MPI_Info info = MPI_INFO_NULL;
    status = nc_create_par(mNameDataFile.c_str(), NC_NETCDF4 | NC_MPIIO,
                           mMyWorld, info, &mIdFile);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

#endif
    this->mIsOpen = true;

    // Add the new variable "version" to an open netCDF dataset in define mode.
    // Variable is of type NC_INT and a scalar.
    status = nc_def_var(mIdFile, "version", NC_INT, 0, 0, &versionID);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

    /*------------------------------------------------------------------------*/
    // Create the local memory mNcMemory for writing the data to the file once:
    // Here: During the creation process.
    if (::boost::is_same<TT, Complex<double> >::value) {
        mNelemsNcMem = 2 * (memNormal.nNodesTotal());
    }
    else if (::boost::is_same<TT, Ntuple<double, DIM> >::value) {
        mNelemsNcMem = DIM * (memNormal.nNodesTotal());
    }
    else {
        mNelemsNcMem = memNormal.nNodesTotal();          
    }
    
    if (0 < mNelemsNcMem) {
        mNcMemory = new double[mNelemsNcMem];

        // Set data field at all nodes to zero in order to be sure that
        // there is no dump in the memory.
        for (int ii = 0; ii < mNelemsNcMem; ++ii) {
            mNcMemory[ii] = -3.897; // Default value.
        }
    }

    /*------------------------------------------------------------------------*/
    // Add a new dimension to an open netCDF dataset.
    std::string tmpName = "nNodes";

    for (std::size_t ii = 0; ii < DIM; ++ii) {
        std::ostringstream oss;
        oss << tmpName << "_" << ii;
        std::string nameOfDim(oss.str());
        status = nc_def_dim(mIdFile, nameOfDim.c_str(),
                            mNnodes.elem(ii),
                            &mIdDims[DIM - ii]);

        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
    }

    status = nc_def_dim(mIdFile, "time", NC_UNLIMITED, &mIdDims[0]);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

    int vID;

  for (std::size_t ii = 0; ii < mNameVariables.size(); ++ii) {
    if (::boost::is_same<TT, Complex<double> >::value) {
        std::string strReal = mNameVariables[ii] + "_re";
        std::string strImag = mNameVariables[ii] + "_im";
        status = nc_def_var(mIdFile, strReal.c_str(), NC_DOUBLE,
                            DIM + 1, mIdDims, &vID);

        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }

        mIdVariable[ii].push_back(vID);
        status = nc_def_var(mIdFile, strImag.c_str(), NC_DOUBLE,
                            DIM + 1, mIdDims, &vID);

        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }

        mIdVariable[ii].push_back(vID);
    }

    else if (::boost::is_same<TT, Ntuple<double, DIM> >::value) {
        std::string strElem = mNameVariables[ii];

        for (std::size_t jj = 0; jj < DIM; ++jj) {
            std::ostringstream oss;
            oss << strElem << "_" << jj;
            std::string tmp(oss.str());
            status = nc_def_var(mIdFile, tmp.c_str(), NC_DOUBLE,
                                DIM + 1, mIdDims, &vID);

            if (status != NC_NOERR) {
                handle_netCDF_error(status, __FILE__, __LINE__);
            }

            mIdVariable[ii].push_back(vID);
        }
    }
    else { 
        status = nc_def_var(mIdFile, (mNameVariables[ii]).c_str(), NC_DOUBLE,
                            DIM + 1, mIdDims, &vID);

        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
        mIdVariable[ii].push_back(vID);
    }

  }

    status = nc_enddef(mIdFile);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

    /*------------------------------------------------------------------------*/
    // Write the version number to file.
    status = nc_put_var_int(mIdFile, versionID, &version);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::write(std::vector<TT*> const& elemData,
                              GridSub<DIM> const& memNormal,
                              int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("DataFile::write()");
#endif

    mElemData = elemData;
    mMemNormal = memNormal;
    // Create the file if it was not done before.
    if (!(this->mIsOpen)) {
        this->create(memNormal);
    }

    const int nDimensions = 1 + DIM;
    Index idxNodeFirst = memNormal.idxNodeFirstSub();
    Index nNodes        = memNormal.nNodesSub();

    // Set dimensions and transform member variables to NC input format.
    std::size_t elemFirst[nDimensions];
    std::size_t nElems[nDimensions];
    elemFirst[0] = mNwritesCurr; // Has to start with 0.
    ++mNwritesCurr;

    for (int ii = 1; ii < nDimensions; ++ii) {
        elemFirst[ii] = (unsigned long int) idxNodeFirst.elem(DIM - ii);
    }

    nElems[0] = 1;

    for (int ii = 1; ii < nDimensions; ++ii) {
        nElems[ii] = (unsigned long int) nNodes.elem(DIM - ii);
    }

    for (typename GridSub<DIM>::iterator it = memNormal.begin(),
            et = memNormal.end(); it < et; ++it) {
    }
    writeFree(*this, elemFirst, nElems);
//     if (0 == mMyWorld.rank()) {
//         std::cout << "* ScaFES: Rank=" << mMyWorld.rank()
//                   << " TIME STEP=" << timestep
//                   << " DF=" << this->mNameVariables[0]
//                   << "...: Written snapshot " << mNwritesCurr
//                   << "." << std::endl;
//     }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::read(std::vector<TT*> & elemData,
                             GridSub<DIM> const& memNormal,
                             int const& timestep)
{
    int idVersion;
    int version = 0;
    int status; // Return value of netCDF methods.

    const int nDimensions = 1 + DIM;
    std::size_t nElems[nDimensions];

    for (std::size_t ii = 0; ii < elemData.size(); ++ii) {
        mElemData[ii] = elemData[ii];
    }
    mMemNormal = memNormal;

#ifdef VTRACE
    VT_TRACER("DataFile::read()");
#endif

    /*------------------------------------------------------------------------*/
    // Open the file with read-only access if it was not done before.
    if (!(this->mIsOpen)) {
        status = nc_open(mNameDataFile.c_str(), NC_NOWRITE, &mIdFile);
        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
        mIsOpen = true;
    }

    /*------------------------------------------------------------------------*/
    // Read in version number.
    status = nc_get_att_int(mIdFile, NC_GLOBAL, "version", &version);
    if (status != NC_NOERR) {
        status = nc_inq_varid(mIdFile, "version", &idVersion);
        if (status != NC_NOERR) {
            version = 0;
        } else {
            nc_get_var_int(mIdFile, idVersion, &version);
        }
    }

    /*------------------------------------------------------------------------*/
    // Get the varid of the data variable, based on its name.
  for (std::size_t ii = 0; ii < elemData.size(); ++ii) {
    status = nc_inq_varid(mIdFile, mNameVariables[ii].c_str(), 
                          &(mIdVariable[ii][0]));
    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }

    /*------------------------------------------------------------------------*/
    // Get the var id and the lengths of the grid and time dimensions.
    std::string tmpName = "nNodes";
    for (std::size_t ii = 0; ii < DIM; ++ii) {
        std::ostringstream oss;
        oss << tmpName << "_" << ii;
        std::string nameOfDim(oss.str());
        status = nc_inq_dimid(mIdFile, nameOfDim.c_str(), &mIdDims[DIM - ii]);
        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
        status = nc_inq_dimlen(mIdFile, mIdDims[DIM - ii], &(nElems[ii]));
        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
        if (memNormal.nNodes().elem(ii) != static_cast<int>(nElems[ii])) {
            std::cerr << " Grid dimensions does not fit to given grid. "
                      << std::endl;
            break;
        }
    }
    status = nc_inq_dimid(mIdFile, "time", &mIdDims[0]);
    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
    status = nc_inq_dimlen(mIdFile, mIdDims[0], &(nElems[nDimensions-1]));
    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

    /*------------------------------------------------------------------------*/
    /* Read data. */
    readFree(*this);
//     if (0 == mMyWorld.rank()) {
//         std::cout << "* ScaFES: Rank=" << mMyWorld.rank()
//                   << " TIME STEP=" << timestep
//                   << " DF=" << this->mNameVariables[0]
//                   << "...: Read data from file."
//                   << std::endl;
//     }
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline void swap(DataFile<TT, DIM>& first,
                 DataFile<TT, DIM>& second)
{
    std::swap(first.mNameDataFile, second.mNameDataFile);
    std::swap(first.mNameVariables, second.mNameVariables);
    std::swap(first.mMyWorld, second.mMyWorld);
    std::swap(first.mNnodes, second.mNnodes);
    std::swap(first.mIdFile, second.mIdFile);

    for (std::size_t ii = 0; ii < DIM + 1; ++ii) {
        int tmp3 = first.mIdDims[ii];
        first.mIdDims[ii] = second.mIdDims[ii];
        second.mIdDims[ii] = tmp3;
    }

    std::swap(first.mIdVariable, second.mIdVariable);
    std::swap(first.mNwritesCurr, second.mNwritesCurr);
    std::swap(first.mIsOpen, second.mIsOpen);
    ScaFES::swap(first.mMemNormal, second.mMemNormal);
    std::swap(first.mElemData, second.mElemData);
    std::swap(first.mNelemsNcMem, second.mNelemsNcMem);
    double* tmp2 = first.mNcMemory;
    first.mNcMemory = second.mNcMemory;
    second.mNcMemory = tmp2;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void readFree(DataFile<short int, DIM>& output)
{
    int status;
    double* ptrToMem;
    double value;
    // Get real number and store it in mNcMemory.
    double tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
    // mIdVariable[0] means real_part!
    status = nc_get_var_double(output.mIdFile, output.mIdVariable[ii][0],
                            output.mNcMemory);

    ptrToMem = output.mNcMemory;
//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        tmpVal = *ptrToMem;
        value  = tmpVal;
        output.mElemData[ii][it.idxScalarNode()] = static_cast<short int>(value);
        ++ptrToMem;
    }

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void readFree(DataFile<double, DIM>& output)
{
    int status;
    double* ptrToMem;
    double value;
    // Get real number and store it in mNcMemory.
    double tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
    // mIdVariable[0] means real_part!
    status = nc_get_var_double(output.mIdFile, output.mIdVariable[ii][0],
                               output.mNcMemory);

    ptrToMem = output.mNcMemory;
//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        tmpVal = *ptrToMem;
        value  = tmpVal;
        output.mElemData[ii][it.idxScalarNode()] = value; // Will be written to elemData.
        ++ptrToMem;
    }

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<short int, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
    int status;
    double* ptrToMem;
    short int value;
    // Get real number and store it in mNcMemory.
    short int tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
   ptrToMem = output.mNcMemory;
    int kk = 0;
//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        tmpVal = output.elemData(ii, it.idxScalarNode());
        value = tmpVal;
        ptrToMem[kk] = static_cast<double>(value);
        ++kk;
    } 
    // mIdVariable[0] means real_part!
    status = nc_put_vara_double(output.mIdFile, output.mIdVariable[ii][0],
                                elemFirst, nElems, output.mNcMemory);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<double, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
    int status;
    double* ptrToMem;
    double value;
    // Get real number and store it in mNcMemory.
    double tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
    ptrToMem = output.mNcMemory;
//#pragma omp parallel for
    int kk = 0;
    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        tmpVal = output.elemData(ii, it.idxScalarNode());
        value = tmpVal;
        ptrToMem[kk] = value; // Value will be written to mNcMemory.
        ++kk;
    }
    // mIdVariable[0] means real_part!
    status = nc_put_vara_double(output.mIdFile, output.mIdVariable[ii][0],
                                elemFirst, nElems, output.mNcMemory);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<Complex<double>, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
    int status;
    double* ptrToMem;
    double value;
    // Get real part of complex value and store it in mNcMemory.
    ptrToMem = output.mNcMemory;
    Complex<double> tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
//#pragma omp parallel for
    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        tmpVal = output.elemData(ii, it.idxScalarNode());
        value = tmpVal.real();
        *ptrToMem = value; // Value will be written to mNcMemory.
        ++ptrToMem;
    }

    // mIdVariable[0] means real_part!
    status = nc_put_vara_double(output.mIdFile,
                                output.mIdVariable[ii][0],
                                elemFirst, nElems, output.mNcMemory);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }

    // Get imaginary part of complex value and store it in mNcMemory.
    ptrToMem = output.mNcMemory; // Reset memory counter.

    for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
            et = output.memNormal().end(); it < et; ++it) {
        // Get real part of complex value and store it in mNcMemory.
        tmpVal = output.elemData(ii, it.idxScalarNode());
        value = tmpVal.imag();
        *ptrToMem = value; // Value will be written to mNcMemory.
        ++ptrToMem;
    }

    // mIdVariable[1] means imaginary part!
    status = nc_put_vara_double(output.mIdFile,
                                output.mIdVariable[ii][1],
                                elemFirst, nElems, output.mNcMemory);

    if (status != NC_NOERR) {
        handle_netCDF_error(status, __FILE__, __LINE__);
    }
  }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<Ntuple<double, DIM>, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
    int status;
    double* ptrToMem;
    double value;
    Ntuple<double, DIM> tmpVal;

 for (std::size_t ii = 0; ii < output.nameVariables().size(); ++ii) {
    for (std::size_t kk = 0; kk < DIM; ++kk) {
        // Get value of kk-th element of n-tuple and store it in mNcMemory.
        ptrToMem = output.mNcMemory;

//#pragma omp parallel for
        for (typename GridSub<DIM>::iterator it = output.memNormal().begin(),
                et = output.memNormal().end(); it < et; ++it) {
            tmpVal = output.elemData(ii, it.idxScalarNode());
            value = tmpVal.elem(kk);
            *ptrToMem = value;
            ++ptrToMem;
        }

        // mIdVariable[kk] means kk-th element!
        status = nc_put_vara_double(output.mIdFile,
                                    output.mIdVariable[ii][kk],
                                    elemFirst, nElems, output.mNcMemory);

        if (status != NC_NOERR) {
            handle_netCDF_error(status, __FILE__, __LINE__);
        }
    }
  }
}

/*----------------------------------------------------------------------------*/
inline void handle_netCDF_error(int const& status, std::string const& file,
                                int const& line)
{
    std::string error;

    switch (status) {
        case NC_NOERR :  // no error
            error = "no error";
            break;

        case NC_EBADID :  // Not a netcdf id
            error = "Not a netcdf id";
            break;

        case NC_ENFILE :  // Too many netcdfs open
            error = "Too many netcdfs open";
            break;

        case NC_EEXIST : // netcdf file exists && NC_NOCLOBBER
            error = "netcdf file exists && NC_NOCLOBBER";
            break;

        case NC_EINVAL : // Invalid Argument
            error = "Invalid Argument";
            break;

        case NC_EPERM : // Write to read only
            error = "Write to read only";
            break;

        case NC_ENOTINDEFINE : // Operation not allowed in data mode
            error = "Operation not allowed in data mode";
            break;

        case NC_EINDEFINE : // Operation not allowed in define mode
            error = "Operation not allowed in define mode";
            break;

        case NC_EINVALCOORDS : // Index exceeds dimension bound
            error = "Index exceeds dimension bound";
            break;

        case NC_EMAXDIMS : // NC_MAX_DIMS exceeded
            error = "NC_MAX_DIMS exceeded";
            break;

        case NC_ENAMEINUSE : // String match to name in use
            error = "String match to name in use";
            break;

        case NC_ENOTATT : // Attribute not found
            error = "Attribute not found";
            break;

        case NC_EMAXATTS : // NC_MAX_ATTRS exceeded
            error = "NC_MAX_ATTRS exceeded";
            break;

        case NC_EBADTYPE : // Not a netcdf data type
            error = "Not a netcdf data type";
            break;

        case NC_EBADDIM : // Invalid dimension id or name
            error = "Invalid dimension id or name";
            break;

        case NC_EUNLIMPOS : // NC_UNLIMITED in the wrong index
            error = "NC_UNLIMITED in the wrong index";
            break;

        case NC_EMAXVARS : // NC_MAX_VARS exceeded
            error = "NC_MAX_VARS exceeded";
            break;

        case NC_ENOTVAR : // Variable not found
            error = "Variable not found";
            break;

        case NC_EGLOBAL : // Action prohibited on NC_GLOBAL varid
            error = "Action prohibited on NC_GLOBAL varid";
            break;

        case NC_ENOTNC : // Not a netcdf file
            error = "Not a netcdf file";
            break;

        case NC_ESTS : // In Fortran, string too short
            error = "In Fortran, string too short";
            break;

        case NC_EMAXNAME : // NC_MAX_NAME exceeded
            error = "NC_MAX_NAME exceeded";
            break;

        case NC_EUNLIMIT : // NC_UNLIMITED size already in use
            error = "NC_UNLIMITED size already in use";
            break;

        case NC_ENORECVARS : // nc_rec op when there are no record vars
            error = "nc_rec op when there are no record vars";
            break;

        case NC_ECHAR : // Attempt to convert between text & numbers
            error = "Attempt to convert between text & numbers";
            break;

        case NC_EEDGE : // Edge+start exceeds dimension bound
            error = "Edge+start exceeds dimension bound";
            break;

        case NC_ESTRIDE : // Illegal stride
            error = "Illegal stride";
            break;

        case NC_EBADNAME : // Attribute or variable name
            // contains illegal characters
            error = "Attribute or variable name contains illegal characters";
            break;

            // N.B. following must match value in ncx.h
        case NC_ERANGE : // Math result not representable
            error = "Math result not representable";
            break;

        case NC_ENOMEM : // Memory allocation (malloc) failure
            error = "Memory allocation (malloc) failure";
            break;

        case NC_EVARSIZE : // One or more variable sizes violate
            // format constraints
            error = "One or more variable sizes violate format constraints";
            break;

        case NC_EDIMSIZE : // Invalid dimension size
            error = "Invalid dimension size";
            break;

        case NC_ETRUNC : // File likely truncated or possibly corrupted
            error = "File likely truncated or possibly corrupted";
            break;

        case NC_EAXISTYPE : // Unknown axis type.
            error = "Unknown axis type.";
            break;

            // HDF5 errors --------------------------
//         case NC_EHDFERR : // Error at HDF5 layer.
//             error = "Error at HDF5 layer";
//             break;
//         case NC_ECANTREAD : // Can't read.
//             error = "Can't read";
//             break;
//         case NC_ECANTWRITE : // Can't write.
//             error = "Can't write";
//             break;
//         case NC_ECANTCREATE : // Can't create.
//             error = "Can't create";
//             break;
//         case NC_EFILEMETA : // Problem with file metadata.
//             error = "Problem with file metadata.\n";
//             error += "Check where to write your "
//                     "file and if you have rights to write there!";
//             break;
//         case NC_EDIMMETA : // Problem with dimension metadata.
//             error = "Problem with dimension metadata.";
//             break;
//         case NC_EATTMETA : // Problem with attribute metadata.
//             error = "Problem with attribute metadata.";
//             break;
//         case NC_EVARMETA : //Problem with variable metadata.
//             error = "Problem with variable metadata.";
//             break;
//         case NC_ENOCOMPOUND : // Not a compound type.
//             error = "Not a compound type";
//             break;
//         case NC_EATTEXISTS : // Attribute already exists.
//             error = "Attribute already exists.";
//             break;
//         case NC_ENOTNC3 : // Attempting netcdf-3 operation on netcdf-4 file.
//             error = "Attempting netcdf-3 operation on netcdf-4 file.";
//             break;
//         case NC_ENOTNC4 : // Attempting netcdf-4 operation on netcdf-3 file.
//             error = "Attempting netcdf-4 operation on netcdf-3 file.";
//             break;
//         case NC_ESTRICTNC3 : // Attempting netcdf-4 operation on strict nc3 netcdf-4 file.
//             error = "Attempting netcdf-4 operation on strict nc3 netcdf-4 file.";
//             break;
//         case NC_ENOPAR : // Parallel operation on file opened for non-parallel access.
//             error = "Parallel operation on file opened for non-parallel access.";
//             break;
//         case NC_EPARINIT : // Error initializing for parallel access.
//             error = "Error initializing for parallel access.";
//             break;
//         case NC_EBADGRPID : // Bad group ID.
//             error = "Bad group ID";
//             break;
//         case NC_EBADTYPID : // Bad type ID.
//             error = "Bad type ID.";
//             break;
//         case NC_ETYPDEFINED: // Type has already been defined and may not be edited.
//             error = "Type has already been defined and may not be edited.";
//             break;
//         case NC_EBADFIELD : // Bad field ID.
//             error = "Bad field ID.";
//             break;
//         case NC_EBADCLASS : // Bad class.
//             error = "Bad class.";
//             break;
//         case NC_EMAPTYPE : // Mapped access for atomic types only.
//             error = "Mapped access for atomic types only";
//             break;
//         case NC_ELATEFILL : // Attempt to define fill value when data already exists.
//             error = "Attempt to define fill value when data already exists.";
//             break;
//         case NC_ELATEDEF : // Attempt to define var properties, like deflate, after enddef.
//             error = "Attempt to define var properties, like deflate, after enddef.";
//             break;
//         case NC_EDIMSCALE : // Probem with HDF5 dimscales.
//             error = "Probem with HDF5 dimscales.";
//             break;
//         case NC_ENOGRP : // No group found.
//             error = "No group found.";
//             break;
//         case NC_ESTORAGE : // Can't specify both contiguous and chunking.
//             error = "Can't specify both contiguous and chunking.";
//             break;
//         case NC_EBADCHUNK : // Bad chunksize.
//             error = "Bad chunksize.";
//             break;
//        case NC_ENOTBUILT : // Attempt to use feature that was not turned on when netCDF was built.
//            error = "Attempt to use feature that was not turned on when netCDF was built.";
//            break;
//        case NC_EUNKNAME :
//            error = "NC_EUNKNAME";
//            break;
        default:
            error = "UNKNOWN ERROR!";
            break;
    }

    std::cout << file << " " << line
              << "\t status = " << status << " Error: " << error << std::endl;

    if (status != NC_NOERR) {
        std::cout << "EXIT...\n";
        std::exit(81);
    }
}

} // End of namespace. //

#endif
