/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_DATAFILEDUMMY_HPP_
#define SCAFES_DATAFILEDUMMY_HPP_

#include "ScaFES_Config.hpp"

#ifdef VTRACE
#include "vt_user.h"
#endif

#include <iostream>
#include <vector>

#include <boost/type_traits.hpp>

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_GridSub.hpp"
#include "ScaFES_Complex.hpp"
#include "ScaFES_Complex_FreeFunctions.hpp"

namespace ScaFES
{

const int DATA_FILE_VERSION = 3;

/*******************************************************************************
 ******************************************************************************/
/** The class \c DataFile is responsible for writing (per default: in parallel)
 * a data field to a given file in the NetCDf output format.
 *
 * \n \remarks The call to the NetCDF creation of the data file is not set
 * within the constructor.
 * Instead, it will be created if the user calls the method \c write() the
 * first time. Thus, in case that no data will be written to the output
 * (using the method \c write()), then no data file will be created, too.
 */
template<typename TT, std::size_t DIM>
class DataFile
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor. */
        DataFile();

        /** Creates own constructor.
         *  \remarks This constructor only sets the names of the file and the
         *  variable which should be written, the file itself WILL NOT be
         *  created here. */
        DataFile(std::string const& nameDataFile,
                 std::vector<std::string> const& nameVariables,
                 boost::mpi::communicator const& myWorld,
                 Index const& nNodes);

        /** Creates own constructor:
         * The file itself WILL be created here, too. */
        DataFile(std::string const& nameDataFile,
                 std::vector<std::string> const& nameVariables,
                 boost::mpi::communicator const& myWorld,
                 Index const& nNodes,
                 GridSub<DIM> const& memNormal);

        /** Creates own copy constructor. */
        DataFile(DataFile<TT, DIM> const&);

        /** Creates own assignment operator. */
        DataFile& operator= (DataFile<TT, DIM> rhs);

        /** Creates own destructor: Closes the file and releases the memory. */
        ~DataFile();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the name of the data file. */
        std::string const& nameDataFile() const;

        /** Returns the name of the variable which should be written. */
        std::vector<std::string> const& nameVariables() const;

        /** Returns the identifier of the file. */
        int const& idFile() const;

        /** Returns the identifiers of the components of the variable. */
        std::vector< std::vector<int> > const& idVariable() const;

        /** Returns if the file is already open or not. */
        bool const& isOpen() const;

        /** Returns the number the data field was written to the file. */
        int const& nWritesCurr() const;

        /** Returns the memory grid. */
        GridSub<DIM> const& memNormal() const;

        /** Returns a pointer to the data which should be written. */
        std::vector<TT*> const& elemData() const;

        /** Returns the data for a given memory position. */
        TT const& elemData(int const& df, int const& idx) const;

        /** Returns the number of elements stored in the memory \c mNcMemory. */
        int const& nElemsNcMem() const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the data file and sets the file layout of the
         *  underlying data field (depending on the template type). */
        void create(GridSub<DIM> const& memNormal);

        /** Writes the elements in the memory to which the pointer points to
         *  the file at the given time step. */
        void write(std::vector<TT*> const& elemData, 
                   GridSub<DIM> const& memNormal,
                   int const& timestep);

        /** Reads the elements in the memory to which the pointer points to
         *  from the file at the given time step. */
        void read(std::vector<TT*> & elemData, 
                  GridSub<DIM> const& memNormal,
                  int const& timestep);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Swaps two objects of the class \c DataFile. */
        template<typename SS, std::size_t DD>
        friend void swap(DataFile<SS, DD>& first,
                         DataFile<SS, DD>& second);

        /** Reads the data field (type \c short int) from the data file. */
        template<std::size_t DD>
        friend void readFree(DataFile<short int, DD>& output);

        /** Reads the data field (type \c double) from the data file. */
        template<std::size_t DD>
        friend void readFree(DataFile<double, DD>& output);

        /** Writes the data field (type \c short) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<short int, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c double) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<double, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c Complex<double>) to the data file. */
        template<std::size_t DD>
        friend void writeFree(DataFile<Complex<double>, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

        /** Writes the data field (type \c Ntuple<double,3>) to the data file.*/
        template<std::size_t DD>
        friend void writeFree(DataFile<Ntuple<double, DD>, DD>& output,
                              std::size_t const elemFirst[DD + 1],
                              std::size_t const nElems[DD + 1]);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Name of NetCDF data file. */
        std::string mNameDataFile;

        /** Name of the data field which should be written to the file. */
        std::vector<std::string> mNameVariables;

        /** MPI communicator. */
        boost::mpi::communicator mMyWorld;

        /** Number of nodes in each dimension. */
        Index mNnodes;

        /** Identifier of a NetCDF dataset.
         * "A netCDF ID that can subsequently be used to refer to the netCDF
         * dataset in other netCDF function calls."
         * [http://www.unidata.ucar.edu/software/netcdf/docs/netcdf-c.html#nc_005fcreate]
         */
        int mIdFile;

        /** Every variable has a unique identifier. */
        std::vector< std::vector<int> > mIdVariable;

        /** Identifiers of space and time dimensions
         * = idNetCDF + nameDimension + lengthDimension. */
        int mIdDims[DIM + 1];

        /** Is file open? */
        bool mIsOpen;

        /** Number of writes. */
        int mNwritesCurr;

        /** Memory grid. */
        GridSub<DIM> mMemNormal;

        /** Pointer to the memory. */
        std::vector<TT*> mElemData;

        /** Number of elements which will be allocated. */
        int mNelemsNcMem;

        /** Linear continuous memory grid for the elements of a data field.
         * \remarks A data field can only store values of one type.
         * Moreover, the type remains constant during the life time
         * of the data field.
         */
        double* mNcMemory;

}; // End of class. //

/*******************************************************************************
 * DECLARATION OF FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c DataFile. */
template<typename TT, std::size_t DIM>
void swap(DataFile<TT, DIM>& first, DataFile<TT, DIM>& second);
/*----------------------------------------------------------------------------*/
/** Reads the data field (type \c short int) from the data file. */
template<std::size_t DIM>
void readFree(DataFile<short int, DIM>& output);
/*----------------------------------------------------------------------------*/
/** Reads the data field (type \c double) from the data file. */
template<std::size_t DIM>
void readFree(DataFile<double, DIM>& output);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c short) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<short int, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c double) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<double, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c Complex<double>) to the data file. */
template<std::size_t DIM>
void writeFree(DataFile<Complex<double>, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Writes the data field (type \c Ntuple<double,DIM>) to the data file.*/
template<std::size_t DIM>
void writeFree(DataFile<Ntuple<double, DIM>, DIM>& output,
               std::size_t const elemFirst[DIM + 1],
               std::size_t const nElems[DIM + 1]);
/*----------------------------------------------------------------------------*/
/** Free function that handles all errors in conjunction with NetCDF.
 * \param status return value of NetCDF-method
 * \param file name of file, where error occurs
 * \param line number of line, where error occurs
 */
void handle_netCDF_error(int const& status, std::string const& file,
                         int const& line);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile()
    : mNameDataFile("EMPTYFILENAME")
    , mNameVariables()
    , mMyWorld()
    , mNnodes(0)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal()
    , mElemData()
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{ }
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(std::string const& nameDataFile,
                                   std::vector< std::string> const& nameVariables,
                                   boost::mpi::communicator const& myWorld,
                                   Index const& nNodes)
    : mNameDataFile()
    , mNameVariables(nameVariables)
    , mMyWorld(myWorld)
    , mNnodes(nNodes)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal()
    , mElemData()
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{
    // Create name of the data file and the name of the partition file
    // including the dimensions of the global grid.
    std::ostringstream tmpStringstream;
    tmpStringstream << nameDataFile
                    << "_" << (nNodes.elem(0));

    for (std::size_t ii = 1; ii < DIM; ++ii) {
        tmpStringstream << "x" <<  (nNodes.elem(ii)) ;
    }

    tmpStringstream << "_" << nameVariables[0] << ".nc";
    mNameDataFile = tmpStringstream.str();
    mIdVariable.resize(nameVariables.size());
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(std::string const& nameDataFile,
                                   std::vector<std::string> const& nameVariables,
                                   boost::mpi::communicator const& myWorld,
                                   Index const& nNodes,
                                   GridSub<DIM> const& memNormal)
    : mNameDataFile()
    , mNameVariables(nameVariables)
    , mMyWorld(myWorld)
    , mNnodes(nNodes)
    , mIdFile(0)
    , mIdVariable()
    , mIdDims()
    , mIsOpen(false)
    , mNwritesCurr(0)
    , mMemNormal(memNormal)
    , mElemData(elemData)
    , mNelemsNcMem(0)
    , mNcMemory(NULL)
{
    // Create name of the data file and the name of the partition file
    // including the dimensions of the global grid.
    std::ostringstream tmpStringstream;
    tmpStringstream << nameDataFile
                    << "_" << (nNodes.elem(0));

    for (std::size_t ii = 1; ii < DIM; ++ii) {
        tmpStringstream << "x" <<  (nNodes.elem(ii)) ;
    }

    tmpStringstream << "_" << nameVariables[0] << ".nc";
    mNameDataFile = tmpStringstream.str();
    mIdVariable.resize(nameVariables.size());
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::DataFile(DataFile<TT, DIM> const& rhs)
    : mNameDataFile(rhs.nameDataFile())
    , mNameVariables(rhs.nameVariables())
    , mMyWorld(rhs.mMyWorld)
    , mNnodes(rhs.mNnodes)
    , mIdFile(rhs.idFile())
    , mIdVariable(rhs.idVariable())
    , mIsOpen(rhs.isOpen())
    , mNwritesCurr(rhs.nWritesCurr())
    , mMemNormal(rhs.memNormal())
    , mElemData(rhs.elemData())
{
    for (std::size_t ii = 0; ii < DIM + 1; ++ii) {
        mIdDims[ii] = rhs.mIdDims[ii];
    }

    mNcMemory = 0;
    mNelemsNcMem = 0;
    if (0 < rhs.mNelemsNcMem) {
        mNcMemory = new double[rhs.mNelemsNcMem];
    }
    for (int ii = 0; ii < rhs.mNelemsNcMem; ++ii) {
        mNcMemory[ii] = rhs.mNcMemory[ii];
    }
    // mNelemsNcMem has to be set AFTER the for loops.
    mNelemsNcMem = rhs.mNelemsNcMem;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>::~DataFile()
{
    delete[] mNcMemory;
    mNcMemory = NULL;
}

/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline DataFile<TT, DIM>& DataFile<TT, DIM>::operator=(DataFile<TT, DIM> rhs)
{
    swap(*this, rhs);
    return *this;
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline std::string const& DataFile<TT, DIM>::nameDataFile() const
{
    return mNameDataFile;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector<std::string> const& DataFile<TT, DIM>::nameVariables() const
{
    return mNameVariables;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::idFile() const
{
    return mIdFile;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector< std::vector<int> > const& DataFile<TT, DIM>::idVariable() const
{
    return mIdVariable;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline bool const& DataFile<TT, DIM>::isOpen() const
{
    return mIsOpen;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::nWritesCurr() const
{
    return mNwritesCurr;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline GridSub<DIM> const& DataFile<TT, DIM>::memNormal() const
{
    return mMemNormal;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline std::vector<TT*> const& DataFile<TT, DIM>::elemData() const
{
    return mElemData;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline TT const& DataFile<TT, DIM>::elemData(int const& jj, 
                                             int const& idx) const
{
    return mElemData[jj][idx];
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
inline int const& DataFile<TT, DIM>::nElemsNcMem() const
{
    return mNelemsNcMem;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::create(GridSub<DIM> const& memNormal)
{
#ifdef VTRACE
    VT_TRACER("DataFile::create()");
#endif
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::write(std::vector<TT*> const& elemData,
                              GridSub<DIM> const& memNormal,
                              int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("DataFile::write()");
#endif

    mElemData = elemData;
    mMemNormal = memNormal;
    // Create the file if it was not done before.
    if (!(this->mIsOpen)) {
        this->create(memNormal);
    }

    const int nDimensions = 1 + DIM;
    Index idxNodeFirst = memNormal.idxNodeFirstSub();
    Index nNodes        = memNormal.nNodesSub();

    // Set dimensions and transform member variables to NC input format.
    std::size_t elemFirst[nDimensions];
    std::size_t nElems[nDimensions];
    elemFirst[0] = mNwritesCurr; // Has to start with 0.
    ++mNwritesCurr;

    for (int ii = 1; ii < nDimensions; ++ii) {
        elemFirst[ii] = (unsigned long int) idxNodeFirst.elem(DIM - ii);
    }

    nElems[0] = 1;

    for (int ii = 1; ii < nDimensions; ++ii) {
        nElems[ii] = (unsigned long int) nNodes.elem(DIM - ii);
    }

    for (typename GridSub<DIM>::iterator it = memNormal.begin(),
            et = memNormal.end(); it < et; ++it) {
    }
    writeFree(*this, elemFirst, nElems);
//     if (0 == mMyWorld.rank()) {
//         std::cout << "* ScaFES: Rank=" << mMyWorld.rank()
//                   << " TIME STEP=" << timestep
//                   << " DF=" << this->mNameVariables[0]
//                   << "...: Written snapshot " << mNwritesCurr
//                   << "." << std::endl;
//     }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t DIM>
void DataFile<TT, DIM>::read(std::vector<TT*> & elemData,
                             GridSub<DIM> const& memNormal,
                             int const& timestep)
{
#ifdef VTRACE
    VT_TRACER("DataFile::read()");
#endif

    /* Read data. */
    readFree(*this);
//     if (0 == mMyWorld.rank()) {
//         std::cout << "* ScaFES: Rank=" << mMyWorld.rank()
//                   << " TIME STEP=" << timestep
//                   << " DF=" << this->mNameVariables[0]
//                   << "...: Read data from file."
//                   << std::endl;
//     }
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
template<typename TT, std::size_t DIM>
inline void swap(DataFile<TT, DIM>& first,
                 DataFile<TT, DIM>& second)
{
    std::swap(first.mNameDataFile, second.mNameDataFile);
    std::swap(first.mNameVariables, second.mNameVariables);
    std::swap(first.mMyWorld, second.mMyWorld);
    std::swap(first.mNnodes, second.mNnodes);
    std::swap(first.mIdFile, second.mIdFile);

    for (std::size_t ii = 0; ii < DIM + 1; ++ii) {
        int tmp3 = first.mIdDims[ii];
        first.mIdDims[ii] = second.mIdDims[ii];
        second.mIdDims[ii] = tmp3;
    }

    std::swap(first.mIdVariable, second.mIdVariable);
    std::swap(first.mNwritesCurr, second.mNwritesCurr);
    std::swap(first.mIsOpen, second.mIsOpen);
    ScaFES::swap(first.mMemNormal, second.mMemNormal);
    std::swap(first.mElemData, second.mElemData);
    std::swap(first.mNelemsNcMem, second.mNelemsNcMem);
    double* tmp2 = first.mNcMemory;
    first.mNcMemory = second.mNcMemory;
    second.mNcMemory = tmp2;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void readFree(DataFile<short int, DIM>& output)
{
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void readFree(DataFile<double, DIM>& output)
{
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<short int, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<double, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<Complex<double>, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void writeFree(DataFile<Ntuple<double, DIM>, DIM>& output,
                      std::size_t const elemFirst[DIM + 1],
                      std::size_t const nElems[DIM + 1])
{
}

/*----------------------------------------------------------------------------*/
inline void handle_netCDF_error(int const& status, std::string const& file,
                                int const& line)
{
}

} // End of namespace. //

#endif
