/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_GRIDGLOBAL_HPP_
#define SCAFES_GRIDGLOBAL_HPP_

#include "ScaFES_Config.hpp"

//#include <iomanip>
#include <stdexcept>
#include <vector>


#include <boost/serialization/serialization.hpp>
// #ifdef SCAFES_HAVE_BOOST_SERIALIZATION
// namespace boost
// {
// namespace serialization
// {
// class access;
// }
// }
// #include <boost/serialization/version.hpp>
// #include <boost/serialization/pfto.hpp>
// #include <boost/serialization/vector.hpp>
// #include <boost/serialization/base_object.hpp>
// #endif

#include <boost/mpi.hpp>

#include "ScaFES_Parameters.hpp"
#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Grid.hpp"
#include "ScaFES_GridSub.hpp"

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class \c GridGlobal represents a global grid.
 *
 * \n Each grid partition gets a unique identifier. This identifier will be
 * equal to the MPI process identifier.
 * \n The global grid will be divided into a given number of grid partitions
 * using a RCB algorithm. This algorithm delivers the minimal number of
 * cutting faces.
 *
 * \n The global grid is a structured grid. Thus, it is derived from the class
 * \c Grid and the class possesses the usual grid properties.
 *
 * \n The order of the direct neighbours within in the connectivity matrix
 * is as follows:
 * LEFT   0
 * RIGHT  1
 * BACK   2
 * FRONT  3
 * BOTTOM 4
 * TOP    5
 *
 * \n The global grid contains information about the global computational
 * domain and the number of partitions, into which the global grid is divided.
 * The computational domain will be divided into sub domains. This process
 * will be usually called domain decomposition.
 * \n The partitioning can be controlled via the command line option
 * \code divideGrid(a,b,c) \endcode with a,b,c in {0,1}.
 * If a component of \c divideGrid is set to 1, cutting the computational domain
 * into this direction is allowed. Analogously, if a component of \c divideGrid
 * is set to 0, then cutting into this direction is forbidden.
 */
template<std::size_t DIM = 3>
class GridGlobal : public Grid<DIM>
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

    public:
        /*----------------------------------------------------------------------
        | ENUMERATIONS.
        ----------------------------------------------------------------------*/
        /** Possible types of a grid node. */
        enum scafes_gridnode_type {
            SCAFES_GLOBAL_ALL_NODE       =    1,
            SCAFES_GLOBAL_NORMAL_NODE    =    2,
            SCAFES_GLOBAL_INNER_NODE     =    4,
            SCAFES_GLOBAL_BORDER_NODE    =    8,
            SCAFES_GLOBAL_GHOST_NODE     =   16,
            SCAFES_GLOBAL_COMM_NODE      =   32,
            SCAFES_PARTITITON_ALL_NODE   =   64,
            SCAFES_PARTITION_INNER_NODE  =  128,
            SCAFES_PARTITION_BORDER_NODE =  256,
            SCAFES_PARTITION_GHOST_NODE  =  512,
            SCAFES_PARTITION_COMM_NODE   = 1024
        };

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own default constructor. */
        GridGlobal();

        /** Creates own constructor.
         * \param params Parameters of ScaFES.
         */
        GridGlobal(Parameters const& params);

        /** Creates another constructor.
         * \param nPartitions Number of grid partitions.
         * \param nNodes Number of grid nodes in each dimension.
         * \param coordNodeFirst Coordinates of the first node.
         * \param coordNodeLast Coordinates of the last node.
         * \param divideGridIntoDir Responsible for cutting the grid.
         */
        GridGlobal(int const& nPartitions,
                   Index const& nNodes,
                   Ntuple<double, DIM> const& coordNodeFirst,
                   Ntuple<double, DIM> const& coordNodeLast,
                   Ntuple<bool, DIM> const& divideGridIntoDir);

        /** Forbids the compiler provided copy constructor. */
        GridGlobal(GridGlobal<DIM> const& rhs);

        /** Forbids the compiler provided assignment operator. */
        GridGlobal& operator= (GridGlobal<DIM> from);

        /** Creates own destructor. */
        ~GridGlobal();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the number of partitions. */
        int const& nPartitions() const;

        /** Returns the decision in which direction the grid
         * should be divided.*/
        Ntuple<bool, DIM> const& divideGrid() const;

        /** Returns the vector of all partitions. */
        std::vector< GridSub<DIM> > const& partition() const;

        /** Returns the partition \c idx. */
        GridSub<DIM> const& partition(int const& ii) const;

        /** Returns the identifiers of all neighbours of all partitions. */
        std::vector< std::vector<int> > const& neighbourId() const;

        /** Returns the identifiers of all neighbours of partition \c ii. */
        std::vector<int> const& neighbourId(int const& ii) const;

        /** Returns the identifier of neighbour \c idx of partition \c ii. */
        int const& neighbourId(int const& ii, int const& idx) const;

        /** Returns the directions of all neighbours of all partitions. */
        std::vector< std::vector<int> > const& neighbourDir() const;

        /** Returns the directions of all neighbours of partition \c ii. */
        std::vector<int> const& neighbourDir(int const& ii) const;

        /** Returns the direction of neighbour \c idx of partition \c ii. */
        int const& neighbourDir(int const& ii, int const& idx) const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Divides the computational domain into partitions.
         * \param nPartitions Number of partitions.
         * \param idxNodeFirst Node number of first grid point.
         * \param nNodes Number of nodes of the grid.
         * \param divideGridIntoDir If a component is set to \c false,
         * the domain will not be partitioned into the
         * corresponding direction.
         * \remarks The algorithm works recursively.
         * It stops if the number of partitions is equal to one.
         */
        void applyRCB(int const& nPartitions,
                      Index const& idxNodeFirst,
                      Index const& nNodes,
                      Ntuple<bool, DIM> const& divideGridIntoDir
                      = Ntuple<bool, DIM>(true));

        /** Searches all neighbours in a given grid dimension.
         * \param currDim Grid dimension.
         */
        void searchNeighbours(int const& currDim);

        /** Searches neighbours in all grid dimensions. */
        void searchNeighbours();

    public:
        /*----------------------------------------------------------------------
        | FREE FRIENDS METHODS.
        ----------------------------------------------------------------------*/
        /** Swaps two objects of this class. */
        template<std::size_t DD>
        friend void swap(GridGlobal<DD>& first, GridGlobal<DD>& second);

        /** Overloads the output operator \c operator<<.
         * Prints information about all grid partitions.
         * This method should be called by only one process! */
        template<std::size_t DD>
        friend std::ostream& operator<<(std::ostream& output,
                                        GridGlobal<DD> const& gg);

    private:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the sum over all CPU loads.
         * \param direction Index of direction.
         * \param idxNodeFirst First node number.
         * \param nNodes Number of grid nodes in each dimension.
         * \return Sum of all CPU loads within a grid along a coordinate.
         */
        float* sum(int const& direction, Index const& idxNodeFirst,
                   Index const& nNodes) const;

        /** Initialises dependent member variables. */
        void initDependentMembers();

    private:
        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        /** Be friend for serializing this class. */
        friend class boost::serialization::access;

        /** Serializes the class. */
        template<class Archive>
        void serialize(Archive& ar, const unsigned int version);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Number of partitions. */
        int mNpartitions;

        /** Divide grid in which directions? */
        Ntuple<bool, DIM> mDivideGridIntoDir;

        /** Vector of all partitions. */
        std::vector< GridSub<DIM> > mPartition;

        /** Identifiers of all neighbours of all partitions. */
        std::vector< std::vector<int> > mvNeighbourId;

        /** Directions of all neighbours of all partitions. */
        std::vector< std::vector<int> > mvNeighbourDir;
}; // End of class. //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c GridGlobal. */
template<std::size_t DIM>
void swap(GridGlobal<DIM>& first, GridGlobal<DIM>& second);
/*----------------------------------------------------------------------------*/
/** Preprares writing an object of the class \c GridGlobal output. */
template<std::size_t DIM>
std::ostream& operator<<(std::ostream& output,
                         GridGlobal<DIM> const& gg);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline GridGlobal<DIM>::GridGlobal()
    : Grid<DIM>()
    , mNpartitions(1)
    , mDivideGridIntoDir(Ntuple<bool, DIM>(false))
    , mPartition()
    , mvNeighbourId()
    , mvNeighbourDir()
{
    initDependentMembers();
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridGlobal<DIM>::GridGlobal(Parameters const& params)
    : Grid<DIM>(0,
                Ntuple<int, DIM>(params.nNodes()) - Index(1),
                Ntuple<double, DIM>(params.coordNodeFirst()),
                Ntuple<double, DIM>(params.coordNodeLast()))
    , mNpartitions(params.nPartitions())
    , mDivideGridIntoDir(Ntuple<bool, DIM>(params.divideGrid()))
    , mPartition()
    , mvNeighbourId()
    , mvNeighbourDir()
{
    if (0 == params.myWorld().rank() && (0 < params.indentDepth())) {
        std::cout << getPrefix(params.indentDepth(), params.myWorld().rank())
                  << " * Partition the global grid..."
                  << std::endl;
    }
    boost::mpi::timer timerTotal;
    timerTotal.restart();
    initDependentMembers();
    double tTotal = timerTotal.elapsed();
    if (0 == params.myWorld().rank() && (0 < params.indentDepth())) {
        std::cout << std::scientific << std::fixed
                  << getPrefix(params.indentDepth(), params.myWorld().rank())
                  << "      * CPU time(gridGlobal):"
                  << std::endl
                  << getPrefix(params.indentDepth(), params.myWorld().rank())
                  << "      ----> t(gridGlobal) = "
                  << std::setw(8) << std::setprecision(5) << tTotal
                  << " s. <-------"
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl
                  << getPrefix(params.indentDepth(), params.myWorld().rank())
                  << "   Partitioned."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridGlobal<DIM>::GridGlobal(int const& nPartitions,
                                   Index const& nNodes,
                                   Ntuple<double, DIM> const& coordNodeFirst,
                                   Ntuple<double, DIM> const& coordNodeLast,
                                   Ntuple<bool, DIM> const& divideGridIntoDir)
    : Grid<DIM>(Index(0),
                nNodes - Index(1),
                coordNodeFirst,
                coordNodeLast)
    , mNpartitions(nPartitions)
    , mDivideGridIntoDir(divideGridIntoDir)
    , mPartition()
    , mvNeighbourId()
    , mvNeighbourDir()
{
    initDependentMembers();
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridGlobal<DIM>::GridGlobal(GridGlobal<DIM> const& from)
    : Grid<DIM>(from)
    , mNpartitions(from.nPartitions())
    , mDivideGridIntoDir(from.divideGrid())
    , mPartition(from.partition())
    , mvNeighbourId(from.neighbourId())
    , mvNeighbourDir(from.neighbourDir())
{ }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridGlobal<DIM>& GridGlobal<DIM>::operator=(GridGlobal<DIM> rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridGlobal<DIM>::~GridGlobal()
{ }

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline int const& GridGlobal<DIM>::nPartitions() const
{
    return mNpartitions;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline Ntuple<bool, DIM> const& GridGlobal<DIM>::divideGrid() const
{
    return mDivideGridIntoDir;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::vector< GridSub<DIM> > const& GridGlobal<DIM>::partition() const
{
    return mPartition;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM> const& GridGlobal<DIM>::partition(int const& ii) const
{
    return mPartition.at(ii);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridGlobal<DIM>::neighbourId(int const& ii,
                                               int const& idx) const
{
    return mvNeighbourId.at(ii).at(idx);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::vector<int> const& GridGlobal<DIM>::neighbourId(int const& ii) const
{
    return mvNeighbourId.at(ii);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::vector< std::vector<int> > const& GridGlobal<DIM>::neighbourId() const
{
    return mvNeighbourId;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridGlobal<DIM>::neighbourDir(int const& ii,
                                                int const& idx) const
{
    return mvNeighbourDir.at(ii).at(idx);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::vector<int> const& GridGlobal<DIM>::neighbourDir(int const& ii) const
{
    return mvNeighbourDir.at(ii);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::vector< std::vector<int> > const& GridGlobal<DIM>::neighbourDir() const
{
    return mvNeighbourDir;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline void GridGlobal<DIM>::applyRCB(const int& nPartitions,
                                      const Index& idxNodeFirst,
                                      const Index& nNodes,
                                      const Ntuple<bool, DIM>& divideGrid)
{
    // Recursion tail: One partition.
    if (1 == nPartitions) {
        Index idxNodeLast = idxNodeFirst + nNodes - Index(1);
        assert(idxNodeFirst < (idxNodeFirst + nNodes));

        if (idxNodeLast <= idxNodeFirst) {
            std::cerr << "ERROR(applyRCB()): idxNodeLast <= idxNodeFirst : "
                      << idxNodeFirst << "\t" << idxNodeLast << std::endl;
        }

        mPartition.push_back( GridSub<DIM>(idxNodeFirst,
                                           idxNodeLast,
                                           this->idxNodeFirst(),
                                           this->idxNodeLast(),
                                           this->coordNodeFirst(),
                                           this->coordNodeLast()) );
        return;
    }

    // Problem is decomposed into more than one partition.
    Index enDir(0);

    for (std::size_t kk = 0; kk < DIM; ++kk) {
        enDir.elem(kk) = (int) (divideGrid.elem(kk));
    }

    int maxnNodes = (enDir * nNodes).idxMaxElem();
    //     int maxnNodes = (divideGrid * nNodes).idxMaxElem();
    int max = nNodes.elem(maxnNodes);
    int p2 = nPartitions / 2;
    int c;
    float* s = sum(maxnNodes, idxNodeFirst, nNodes);
    float border = p2 * s[max - 1] / nPartitions;

    for (c = 0; (c < max) && (s[c] < border); ++c);

    {
        ++c;
    }
    Index s2 = idxNodeFirst;
    s2.elem(maxnNodes) += c;

    Index d1 = nNodes;
    Index d2 = nNodes;
    d1.elem(maxnNodes) += c - max;
    d2.elem(maxnNodes) -= c;

    delete[] s;

    // Recursive call.
    applyRCB(p2, idxNodeFirst, d1, divideGrid);
    applyRCB(nPartitions - p2, s2, d2, divideGrid);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void GridGlobal<DIM>::searchNeighbours(int const& currDim)
{
    // Direction (0|1|...|DIM-1) for grid dimensions.
    Ntuple < int, DIM - 1 > dirs;

    const int DIM_MINUS_ONE = DIM - 1;
    for (int kk = 0; kk < DIM_MINUS_ONE; ++kk) {
        dirs[kk] = (currDim + 1) % DIM;
    }

    bool addGrid = false;
    Index tempSt;
    int idJJ;

    // Loop over all partitions.
    for (std::size_t ii = 0; ii < partition().size(); ++ii) {
        // Loop over all other partitions.
        for (std::size_t jj = 0; jj < partition().size(); ++jj) {
            idJJ = jj;

            // Ignore reference cuboid.
            if (jj != ii) {
                Index startII = partition(ii).idxNodeFirstSub();
                Index endII   = partition(ii).idxNodeLastSub();
                Index startJJ = partition(jj).idxNodeFirstSub();
                Index endJJ   = partition(jj).idxNodeLastSub();

                // Do the cuboids have a common face?
                // in der gleichen Ebene?
                if (startII.elem(currDim) == (endJJ.elem(currDim) + 1)
                        || (endII.elem(currDim) + 1) == startJJ.elem(currDim)) {

                    // Case DIM=1: There DOES exist a neighboured partition.
                    // Overlap the cuboids with this layer?
                    // Quick hack: Partitions are in x-y-layer.
                    // TODO: How does it look in general??
                    addGrid = true;

                    for (int ll = 0; ll < DIM_MINUS_ONE; ++ll) {
                        if (std::min(endII.elem(dirs[ll]),
                                     endJJ.elem(dirs[ll]))
                                < std::max(startII.elem(dirs[ll]),
                                           startJJ.elem(dirs[ll])) ) {
                            addGrid = false;
                            break;
                        }
                    }

                    if (addGrid) {
                        tempSt = Index(0);
                        tempSt[currDim] = std::max(startII.elem(currDim),
                                                   startJJ.elem(currDim));

                        for (int kk = 0; kk < DIM_MINUS_ONE; ++kk) {
                            tempSt[dirs[kk]] = std::max(startII.elem(dirs[kk]),
                                                        startJJ.elem(dirs[kk]));
                        }

                        // TODO: Why?
                        int dn = 2 * currDim;

                        if (!(tempSt < (endII + Index(1)))) {
                            ++dn;
                        }

                        mvNeighbourId.at(ii).push_back(idJJ);
                        mvNeighbourDir.at(ii).push_back(dn);
                    }
                }
            }
        }
    }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void GridGlobal<DIM>::searchNeighbours()
{
    for (std::size_t ii = 0; ii < DIM; ++ii) {
        searchNeighbours(ii);
    }
}

/*******************************************************************************
 * PRIVATE WORK METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline float* GridGlobal<DIM>::sum(int const& currDim,
                                   Index const& idxNodeFirst,
                                   Index const& nNodes) const
{
    assert(0 <= currDim && currDim <= (int) DIM);

    float* s = new float[nNodes.elem(currDim)];
    int d1 = (currDim + 1) % DIM;
    int d2 = (currDim + 2) % DIM;

    Index index;
    double su = 0.0;

    for (int i = 0; i < nNodes.elem(currDim); ++i) {
        index[currDim] = i + idxNodeFirst.elem(currDim);

        for (int j = idxNodeFirst.elem(d1);
                j < idxNodeFirst.elem(d1) + nNodes.elem(d1); ++j) {
            index[d1] = j;

            for (int k = idxNodeFirst.elem(d2);
                    k < idxNodeFirst.elem(d2) + nNodes.elem(d2); ++k) {
                index[d2] = k;
                //su += (double) loadComp(index);
                su += 1.0;
            }
        }

        s[i] = (float) su;
    }

    return s;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline void GridGlobal<DIM>::initDependentMembers()
{
    if (0 < nPartitions()) {
        //mPartition.resize(nPartitions());
        mvNeighbourId.resize(nPartitions());
        mvNeighbourDir.resize(nPartitions());
    }

    assert(0 < this->nNodes().size());
    if (Index(0) < this->nNodes()) {
//      for (std::size_t ii = 0; ii < DIM; ++ii) {
//          if (divideGrid().elem(ii)) {
                applyRCB(mNpartitions, Index(0), this->nNodes(), divideGrid());
                searchNeighbours();
//              break;
//          }
//      }
    }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
template<class Archive>
inline void GridGlobal<DIM>::serialize(Archive& ar, const unsigned int version)
{
    if (1 <= version) {
        ar& boost::serialization::base_object< Grid<DIM> >(*this);
        ar & (mNpartitions);
        ar & (mDivideGridIntoDir);
        ar & (mPartition);
        ar & (mvNeighbourId);
        ar & (mvNeighbourDir);
    }
}

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline void swap(GridGlobal<DIM>& first, GridGlobal<DIM>& second)
{
    ScaFES::swap(static_cast<Grid<DIM>&>(first),
                 static_cast<Grid<DIM>&>(second));
    std::swap(first.mNpartitions, second.mNpartitions);
    ScaFES::swap(first.mDivideGridIntoDir, second.mDivideGridIntoDir);
    std::swap(first.mPartition, second.mPartition);
    std::swap(first.mvNeighbourId, second.mvNeighbourId);
    std::swap(first.mvNeighbourDir, second.mvNeighbourDir);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::ostream& operator<<(std::ostream& output,
                                GridGlobal<DIM> const& gg)
{
    for (int ii = 0; ii < gg.nPartitions(); ++ii) {
        output << std::fixed;
        output << "-----------------------------------------------" << std::endl
               << "* Grid partition id   : " << ii
               << std::endl
               << "  Grid partition range: " << std::endl
               << gg.idxNodeFirst() << " --- "
               << gg.idxNodeLast()
               << std::endl;

        for (std::size_t jj = 0; jj < gg.neighbourId(ii).size(); ++jj) {
            output << "* Neighbour grid id   : " << gg.neighbourId(ii, jj)
                   << std::endl
                   << "  Neighbour direction : " << gg.neighbourDir(ii, jj)
                   << std::endl
                   << "  Neighbour grid range: " << std::endl
                   << gg.partition(jj) << std::endl;
        }
    }

    return output << std::endl << std::endl;
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
namespace boost
{
namespace serialization
{
/** Designed to set the boost serialization version of a class template. */
template<std::size_t DIM>
struct version< ScaFES::GridGlobal<DIM> > {
    /** Sets the version number for serialization. */
    BOOST_STATIC_CONSTANT(unsigned long int, value = 2);
};
} // namespace serialization
} // namespace boost

#endif
