/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_GRIDSUB_HPP_
#define SCAFES_GRIDSUB_HPP_

#include "ScaFES_Config.hpp"

#include <iterator>
#include <vector>

#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
// Forward declaration.
namespace boost
{
namespace serialization
{
class access;
}
}
#include <boost/serialization/version.hpp>
#include <boost/serialization/pfto.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/base_object.hpp>
#endif

#include <boost/config/suffix.hpp>

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Ntuple_FreeFunctions.hpp"
#include "ScaFES_Grid.hpp"

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class template \c GridSub<DIM> represents a structured (sub-) grid
 * of dimension \c DIM. A grid is a \c DIM - dimensional hypercuboid.
 *
 * \n The nodes of the grid are lexicographically numbered. Therefore,
 * only the first and the last node number of the grid together
 * with its corresponding coordinates have to be stored.
 *
 * \n Furthermore, the class contains an iterator class. The iterator can be
 * used like an iterator of the STL.
 *
 * \n It will be assumed that the sub grid and the base grid have the same
 * grid size.
 */
template<std::size_t DIM = 3>
class GridSub : public Grid<DIM>
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

    public:
        /*----------------------------------------------------------------------
        | INTERNAL CLASSES.
        ----------------------------------------------------------------------*/
        /** Class for iterating through the grid.*/
        class Iterator
            : public std::iterator<std::random_access_iterator_tag , Index>
        {

                /*--------------------------------------------------------------
                | FRIEND CLASSES.
                --------------------------------------------------------------*/
                friend class GridSub<DIM>;

            public:
                /*--------------------------------------------------------------
                | LIFE CYCLE METHODS.
                --------------------------------------------------------------*/
                /** Creates own constructor. */
                Iterator(const GridSub<DIM>* gd, Index const& idxNode);

                Iterator();

//                 /** Creates own copy constructor. */
//                 Iterator(Iterator const&); // = default;
// 
//                 /** Creates own assignment operator. */
//                 Iterator& operator=(Iterator const&); // = default;

            public:
                /*--------------------------------------------------------------
                | GETTER METHODS.
                --------------------------------------------------------------*/
                /** Returns the current global node number. */
                Index const& idxNode() const;

                /** Returns the current local node number. */
                unsigned long int const& idxScalarNode() const;

            public:
                /*--------------------------------------------------------------
                | COMPARISON METHODS.
                --------------------------------------------------------------*/
                /** Checks if the local node number of this iterator is unequal
                 * to the one of the given iterator. */
                bool operator!=(const Iterator& other) const;

                /** Checks if the local node number of this iterator is smaller
                 * than the one of the given iterator. */
                bool operator<(const Iterator& other) const;

            public:
                /*--------------------------------------------------------------
                | WORK METHODS.
                --------------------------------------------------------------*/
                /** Returns the current global node number. */
                Index const& operator*();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global
                 * node number.
                 * \remarks Prefix variant.
                 */
                Iterator& operator++();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global node
                 * number.
                 * \remarks Postfix variant.
                 */
                Iterator operator++(int incr);

            private:
                /*--------------------------------------------------------------
                | MEMBER VARIABLES.
                --------------------------------------------------------------*/
                /** Grid over which should be iterated. */
                const GridSub<DIM>* mGridIter;

                /** Current global node number. */
                Index mIdxNode;

                /** Current local node number. */
                unsigned long int mIdxScalarNode;

                /** Helper variable in order to improve the performance.
                 * Jumps in numbering while iterating rhs one grid dimension
                 * to the next one. These jumps occur due to the fact that
                 * the base grid is normally not the sub grid. */
                Index mJumpInNumbering;

            private:
                /*--------------------------------------------------------------
                | INTERNAL MEMBER METHODS.
                --------------------------------------------------------------*/
                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global
                 * node number iterating "columnwise" through the grid.
                 */
                Iterator& nextStyleC();

                /** Updates the member variables of the iterator class:
                 * Increases the current local and the current global
                 * node number iterating "rowwise" through the grid.
                 */
                Iterator& nextStyleFortran();
        };

    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for iterator class analogously to STL class. */
        typedef Iterator iterator;

        /** Type definition for iterator class analogously to STL class. */
        typedef const Iterator const_iterator;

        /** Type definition for iterator class analogously to STL class. */
        typedef Index& reference_type;

        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor. */
        GridSub();

        /** Creates a new sub grid with given first global node number and
         * given last global node related to a given base grid.
         * \param idxNodeFirstSub First node number of the sub grid.
         * \param idxNodeLastSub Last node number of the sub grid.
         * \param idxNodeFirst First node number of the BASE GRID.
         * \param idxNodeLast Last node number of the BASE GRID.
         * \param coordNodeFirst Coordinates of first node number.
         * \param coordNodeLast Coordinates of last node number.
         */
        GridSub(Index const& idxNodeFirstSub,
                Index const& idxNodeLastSub,
                Index const& idxNodeFirst,
                Index const& idxNodeLast,
                Ntuple<double, DIM> coordNodeFirst,
                Ntuple<double, DIM> coordNodeLast);

        /** Creates own constructor.
         * The first and the last node number are explicitly set.
         * \param idxNodeFirstSub First node number of the sub grid.
         * \param idxNodeLastSub Last node number of the sub grid.
         * \param ggBase BASE GRID.
         */
        GridSub(Index const& idxNodeFirstSub,
                Index const& idxNodeLastSub,
                Grid<DIM> const& ggBase);

        /** Creates own copy constructor. */
        GridSub(GridSub<DIM> const& rhs);

        /** Creates own assignment operator using the copy-and-swap idiom.
         */
        GridSub& operator=(GridSub<DIM> rhs);

        /** Creates own destructor.  */
        ~GridSub();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the first node number. */
        Index const& idxNodeFirstSub() const;

        /** Returns the component \c idx of the first node number. */
        int const& idxNodeFirstSub(int const& idx) const;

        /** Returns the last node number. */
        Index const& idxNodeLastSub() const;

        /** Returns the component \c idx of the last node number. */
        int const& idxNodeLastSub(int const& idx) const;

        /** Returns the first local node number in the grid. */
        unsigned long int const& idxScalarNodeFirst() const;

        /** Returns the last local node number in the grid. */
        unsigned long int const& idxScalarNodeLast() const;

        /** Returns the number of all nodes of the grid. */
        int const& nNodesTotal() const;

        /** Returns the number of nodes in each direction. */
        Index const& nNodesSub() const;

        /** Returns the number of nodes in one given dimension. */
        int const& nNodesSub(int const& idx) const;

    public:
        /*----------------------------------------------------------------------
        | COMPARISON METHODS.
        ----------------------------------------------------------------------*/
        /** Tests if two grids are equal. */
        bool operator==(GridSub<DIM> const& rhs) const;

        /** Tests if two grids are not equal. */
        bool operator!=(GridSub<DIM> const& rhs) const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the iterator corresponding to the underlying grid. */
        iterator begin() const;

        /** Returns the default iterator. */
        iterator end() const;

//         /** Computes the corresponding local node number of a given global
//          * node number of the grid.
//          */
//         unsigned long int idxNodeTuple2Scalar(Index const& idxNode) const;

        /** Returns true if the first global node number is equal or smaller the
         * the last global node number, i.e. if the grid consists of at
         * least one node. */
        bool valid() const;

        /** Returns true if a given global node number is inside the grid.
         */
        bool inside(Index const& idxNode) const;

        /** Extends the grid by a given border width into a given direction.
         * \param dir Direction to expand.
         * \param nLayers Number of layers to expand.
         */
        GridSub<DIM> extend(int const& dir, int const& nLayers);

        /** Returns the union of this grid and another given grid.
         * Assumption: Both grids must have the same base grid.
         * \return Grid gg(min(mIdxNodeFirst, rhs.idxNodeFirstSub),
         *                 max(mIdxNodeLast, rhs.idxNodeLastSub))
         */
        GridSub<DIM> unionWith(GridSub<DIM> const& rhs) const;

        /** Returns the intersection of this grid and another given grid.
         * Assumption: Both grids must have the same base grid.
         * \return Grid gg(max(mIdxNodeFirst, rhs.idxNodeFirstSub),
         *                 min(mIdxNodeLast, rhs.idxNodeLastSub))
         */
        GridSub<DIM> intersectWith(GridSub<DIM> const& rhs) const;

        /** Serializes the class. */
        template<class Archive>
        void serialize(Archive& ar, const unsigned int version);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS.
        ----------------------------------------------------------------------*/
        /** Swaps the members of two grids. */
        template<std::size_t SS>
        friend void swap(GridSub<SS>& first, GridSub<SS>& second);

        /** Overloads the output operator '<<'. */
        template<std::size_t SS>
        friend std::ostream& operator<<(std::ostream& output,
                                        GridSub<SS> const& rhs);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** First global node number of the grid. */
        Index mIdxNodeFirstSub;

        /** Last global node number of the grid.
         * \remarks The number of nodes in each direction is given by the
         difference of the last and first node number + 1. */
        Index mIdxNodeLastSub;

        /** Number of nodes of the grid in all dimensions. */
        Index mNnodesSub;

        /** Total number of grid nodes. */
        int mNnodesTotal;

        /** First MPI local node number of the grid. */
        unsigned long int mIdxScalarNodeFirst;

        /** Last MPI local node number of the grid. */
        unsigned long int mIdxScalarNodeLast;

    private:
        /*----------------------------------------------------------------------
        | INTERNAL METHODS.
        ----------------------------------------------------------------------*/
        /** Computes the corresponding local node number of a given global
         * node number of the grid using the C style.
         * \param idxNode Given global node number
         */
        unsigned long int idxNodeTuple2ScalarStyleC(Index const& idxNode) const;

        /** Computes the corresponding local node numbervof a given global
         * node number of the grid using the FORTRAN style.
         * \param idxNode Given global node number
         */
        unsigned long int idxNodeTuple2ScalarStyleFortran(Index const& idxNode) const;
}; // End of class. //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c GridSub. */
template<std::size_t DIM>
void swap(GridSub<DIM>& first, GridSub<DIM>& second);
/*----------------------------------------------------------------------------*/
/** Preprares writing an object of the class \c GridSub to output. */
template<std::size_t DIM>
std::ostream& operator<<(std::ostream& output, GridSub<DIM> const& gg);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline GridSub<DIM>::GridSub()
    : Grid<DIM>()
    , mIdxNodeFirstSub(Index(0))
    , mIdxNodeLastSub(Index(-1))
    , mNnodesSub(Index(0))
    , mNnodesTotal(0)
    , mIdxScalarNodeFirst(0)
    , mIdxScalarNodeLast(0)
{ }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>::GridSub(Index const& idxNodeFirstSub,
                             Index const& idxNodeLastSub,
                             Index const& idxNodeFirst,
                             Index const& idxNodeLast,
                             Ntuple<double, DIM> coordNodeFirst,
                             Ntuple<double, DIM> coordNodeLast)
    : Grid<DIM>(idxNodeFirst, idxNodeLast,
                coordNodeFirst, coordNodeLast)
    , mIdxNodeFirstSub(idxNodeFirstSub)
    , mIdxNodeLastSub(idxNodeLastSub)
    , mNnodesSub(mIdxNodeLastSub - mIdxNodeFirstSub + Index(1))
    , mNnodesTotal(mNnodesSub.size())
    , mIdxScalarNodeFirst(this->idxNodeTuple2Scalar(mIdxNodeFirstSub))
    , mIdxScalarNodeLast(this->idxNodeTuple2Scalar(mIdxNodeLastSub))
{
    // Check if the given grid is a sub grid of the base grid.
//     if (!(this->inside(idxNodeFirstSub))
//          || !(this->inside(idxNodeLastSub)) ) {
//         std::cout << " WARNING: Given grid is NOT a sub grid of the base grid."
//                   << std::endl;
//     }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>::GridSub(Index const& idxNodeFirstSub,
                             Index const& idxNodeLastSub,
                             Grid<DIM> const& ggBase)
    : Grid<DIM>(ggBase)
    , mIdxNodeFirstSub(idxNodeFirstSub)
    , mIdxNodeLastSub(idxNodeLastSub)
    , mNnodesSub(mIdxNodeLastSub - mIdxNodeFirstSub + Index(1))
    , mNnodesTotal(mNnodesSub.size())
    , mIdxScalarNodeFirst(this->idxNodeTuple2Scalar(mIdxNodeFirstSub))
    , mIdxScalarNodeLast(this->idxNodeTuple2Scalar(mIdxNodeLastSub))
{ }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>::GridSub(GridSub<DIM> const& rhs)
    : Grid<DIM>(rhs)
    , mIdxNodeFirstSub(rhs.idxNodeFirstSub())
    , mIdxNodeLastSub(rhs.idxNodeLastSub())
    , mNnodesSub(rhs.nNodesSub())
    , mNnodesTotal(rhs.nNodesTotal())
    , mIdxScalarNodeFirst(rhs.idxScalarNodeFirst())
    , mIdxScalarNodeLast(rhs.idxScalarNodeLast())
{  }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>& GridSub<DIM>::operator=(GridSub<DIM> rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>::~GridSub()
{
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline typename GridSub<DIM>::Index const& GridSub<DIM>::idxNodeFirstSub() const
{
    return mIdxNodeFirstSub;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridSub<DIM>::idxNodeFirstSub(int const& idx) const
{
    return mIdxNodeFirstSub.elem(idx);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Index const& GridSub<DIM>::idxNodeLastSub() const
{
    return mIdxNodeLastSub;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridSub<DIM>::idxNodeLastSub(int const& idx) const
{
    return mIdxNodeLastSub.elem(idx);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline unsigned long int const& GridSub<DIM>::idxScalarNodeFirst() const
{
    return mIdxScalarNodeFirst;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline unsigned long int const& GridSub<DIM>::idxScalarNodeLast() const
{
    return mIdxScalarNodeLast;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Index const& GridSub<DIM>::nNodesSub() const
{
    return mNnodesSub;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridSub<DIM>::nNodesSub(int const& idx) const
{
    return mNnodesSub.elem(idx);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline int const& GridSub<DIM>::nNodesTotal() const
{
    return mNnodesTotal;
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline bool GridSub<DIM>::operator==(GridSub<DIM> const& rhs) const
{
    return (idxNodeFirstSub() == rhs.idxNodeFirstSub()
            && idxNodeLastSub() == rhs.idxNodeLastSub());
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool GridSub<DIM>::operator!=(GridSub<DIM> const& rhs) const
{
    return !(*this == rhs);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline typename GridSub<DIM>::iterator GridSub<DIM>::begin() const
{
    return iterator(this, idxNodeFirstSub());
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::iterator GridSub<DIM>::end() const
{
    return iterator(this, idxNodeLastSub());
}
// /*----------------------------------------------------------------------------*/
// template<std::size_t DIM>
// inline unsigned long int GridSub<DIM>::idxNodeTuple2Scalar(Index const& idxNode) const
// {
//     if (this->mIsNumberedStyleC) {
//         return (this->idxNodeTuple2ScalarStyleC(idxNode));
//     }
// 
//     return (this->idxNodeTuple2ScalarStyleFortran(idxNode));
// }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool GridSub<DIM>::valid() const
{
    return (idxNodeFirstSub() <= idxNodeLastSub());
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool GridSub<DIM>::inside(Index const& idxNode) const
{
    return ( (idxNodeFirstSub() <= idxNode)
             && (idxNode <= idxNodeLastSub()));
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM> GridSub<DIM>::extend(int const& dir, int const& nLayers)
{
    Index idxNodeFirstSub(this->mIdxNodeFirstSub);
    Index idxNodeLastSub(this->mIdxNodeLastSub);
    int idx = dir / 2;

    if (0 == (dir % 2)) {
        idxNodeLastSub[idx] += nLayers;
    } else {
        idxNodeFirstSub[idx] -= nLayers;
    }

    return GridSub<DIM>(idxNodeFirstSub,
                        idxNodeLastSub,
                        static_cast< Grid<DIM> >(*this));
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM> GridSub<DIM>::unionWith(GridSub<DIM> const& rhs) const
{
    return GridSub<DIM>(min(idxNodeFirstSub(), rhs.idxNodeFirstSub()),
                        max(idxNodeLastSub(), rhs.idxNodeLastSub()),
                        this->idxNodeFirst(),
                        this->idxNodeLast(),
                        this->coordNodeFirst(),
                        this->coordNodeLast());
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM> GridSub<DIM>::intersectWith(GridSub<DIM> const& rhs) const
{
    return GridSub<DIM>(max(idxNodeFirstSub(), rhs.idxNodeFirstSub()),
                        min(idxNodeLastSub(), rhs.idxNodeLastSub()),
                        this->idxNodeFirst(),
                        this->idxNodeLast(),
                        this->coordNodeFirst(),
                        this->coordNodeLast());
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
template<class Archive>
inline void GridSub<DIM>::serialize(Archive& ar, const unsigned int version)
{
    if (1 <= version) {
        ar& boost::serialization::base_object< Grid<DIM> >(*this);
        ar & (mIdxNodeFirstSub);
        ar & (mIdxNodeLastSub);
        ar & (mNnodesSub);
        ar & (mNnodesTotal);
        ar & (mIdxScalarNodeFirst);
        ar & (mIdxScalarNodeLast);
    }
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline unsigned long int GridSub<DIM>::idxNodeTuple2ScalarStyleC(Index const& idxNode) const
{
    unsigned long int idxScalarNode
        = idxNode.elem(0) - idxNodeFirstSub(0);

    for (std::size_t ii = 1; ii < DIM; ++ii) {
        idxScalarNode = idxScalarNode * this->nNodesSub(ii)
            + idxNode.elem(ii) - this->idxNodeFirst(ii);
    }

    return idxScalarNode;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline unsigned long int GridSub<DIM>::idxNodeTuple2ScalarStyleFortran(Index const& idxNode) const
{
    unsigned long int idxScalarNode
        = idxNode.elem(DIM - 1) - idxNodeFirstSub(DIM - 1);
    const int dimLocal = DIM;

    for (int ii = dimLocal - 2; ii >= 0; --ii) {
        idxScalarNode = idxScalarNode * this->nNodesSub(ii)
            + idxNode.elem(ii) - this->idxNodeFirst(ii);
    }

    return idxScalarNode;
}

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline void swap(GridSub<DIM>& first, GridSub<DIM>& second)
{
    swap(static_cast< Grid<DIM>&>(first),
         static_cast< Grid<DIM>&>(second));
    ScaFES::swap(first.mIdxNodeFirstSub, second.mIdxNodeFirstSub);
    ScaFES::swap(first.mIdxNodeLastSub, second.mIdxNodeLastSub);
    ScaFES::swap(first.mNnodesSub, second.mNnodesSub);
    int tmp = first.mNnodesTotal;
    first.mNnodesTotal = second.mNnodesTotal;
    second.mNnodesTotal = tmp;
    std::swap(first.mIdxScalarNodeFirst, second.mIdxScalarNodeFirst);
    std::swap(first.mIdxScalarNodeLast, second.mIdxScalarNodeLast);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline std::ostream& operator<<(std::ostream& output, GridSub<DIM> const& gg)
{
    output << "Sub Grid   <-" << gg.idxNodeFirstSub()
           << "-" << gg.idxNodeLastSub() << "->\n"
           << "Base Grid: <-" << gg.idxNodeFirst()
           << "-" << gg.idxNodeLast() << "->";
    return output;
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: LIFE CYCLE METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline GridSub<DIM>::Iterator::Iterator()
    : mGridIter(NULL)
    , mIdxNode(Index(0))
    , mIdxScalarNode(0L)
{ }
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline GridSub<DIM>::Iterator::Iterator(const GridSub<DIM>* gd,
                                        Index const& currIdxNode)
    : mGridIter(gd)
    , mIdxNode(currIdxNode)
    , mIdxScalarNode(gd->idxNodeTuple2Scalar(currIdxNode))
{
    // TODO: Check if \c currIdxNode is inside underlying grid.
    if (gd->isNumberedStyleC()) {
        mJumpInNumbering[DIM - 1] = 1;
        mJumpInNumbering[DIM - 2] = (this->mGridIter->idxNodeLast(DIM - 1)
                                     - this->mGridIter->idxNodeLastSub(DIM - 1))
                                    - (this->mGridIter->idxNodeFirst(DIM - 1)
                                       - this->mGridIter->idxNodeFirstSub(DIM - 1));
        const int dimLocal = DIM;
        for (int ii = dimLocal - 2; ii >= 1; --ii) {
            mJumpInNumbering[ii - 1] = mJumpInNumbering.elem(ii)
                                       + this->mGridIter->nNodes(ii + 1)
                                       * ( (this->mGridIter->idxNodeLast(ii)
                                            - this->mGridIter->idxNodeLastSub(ii))
                                           - (this->mGridIter->idxNodeFirst(ii)
                                              - this->mGridIter->idxNodeFirstSub(ii)) );
        }
    } else {
        mJumpInNumbering[0] = 1;
        mJumpInNumbering[1] = (this->mGridIter->idxNodeLast(0)
                               - this->mGridIter->idxNodeLastSub(0))
                              - (this->mGridIter->idxNodeFirst(0)
                                 - this->mGridIter->idxNodeFirstSub(0));

       const int DIM_MINUS_ONE = DIM - 1;
       for (int ii = 1; ii < DIM_MINUS_ONE; ++ii) {
            mJumpInNumbering[ii + 1] = mJumpInNumbering.elem(ii)
                                       + this->mGridIter->nNodes(ii - 1)
                                       * ( (this->mGridIter->idxNodeLast(ii)
                                            - this->mGridIter->idxNodeLastSub(ii))
                                           - (this->mGridIter->idxNodeFirst(ii)
                                              - this->mGridIter->idxNodeFirstSub(ii)) );
        }
    }
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: GETTER METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline typename GridSub<DIM>::Index const& GridSub<DIM>::Iterator::idxNode() const
{
    return mIdxNode;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline unsigned long int const& GridSub<DIM>::Iterator::idxScalarNode() const
{
    return mIdxScalarNode;
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: COMPARISON METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline bool GridSub<DIM>::Iterator::operator!=(const Iterator& other) const
{
    return (*this < other) || (other < *this);
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline bool GridSub<DIM>::Iterator::operator<(const Iterator& it) const
{
    return (idxScalarNode() <= it.idxScalarNode());
}

/*******************************************************************************
 * INTERNAL CLASS ITERATOR: WORK METHODS.
 ******************************************************************************/
template<std::size_t DIM>
inline typename GridSub<DIM>::Index const& GridSub<DIM>::Iterator::operator*()
{
    return idxNode();
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Iterator& GridSub<DIM>::Iterator::operator++()
{
    if (mGridIter->isNumberedStyleC()) {
        return this->nextStyleC();
    }

    return this->nextStyleFortran();
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Iterator GridSub<DIM>::Iterator::operator++(int)
{
    Iterator tmp = *this;
    ++(*this);
    return tmp;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Iterator& GridSub<DIM>::Iterator::nextStyleC()
{
    ++(mIdxScalarNode);
    // Case: Next node in first dimension?
    ++(mIdxNode[DIM - 1]);

    if (mIdxNode.elem(DIM - 1) > this->mGridIter->idxNodeLastSub(DIM - 1)) {
        int dimLocal = DIM;

        // Case: Next node in (ii-1)-th dimension?
        for (int ii = dimLocal - 2; ii >= 0; --ii) {
            mIdxNode[ii + 1] = this->mGridIter->idxNodeFirstSub(ii + 1);
            ++(mIdxNode[ii]);

            if (mIdxNode.elem(ii) <= this->mGridIter->idxNodeLastSub(ii)) {
                mIdxScalarNode += mJumpInNumbering.elem(ii);
                break;
            }
        }
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<std::size_t DIM>
inline typename GridSub<DIM>::Iterator& GridSub<DIM>::Iterator::nextStyleFortran()
{
    ++(mIdxScalarNode);
    // Case: Next node in first dimension?
    ++(mIdxNode[0]);

    if (mIdxNode.elem(0) > this->mGridIter->idxNodeLastSub(0)) {
        // Case: Next node in (ii+1)-th dimension?
        for (std::size_t ii = 1; ii < DIM; ++ii) {
            mIdxNode[ii - 1] = this->mGridIter->idxNodeFirstSub(ii - 1);
            ++(mIdxNode[ii]);

            if (mIdxNode.elem(ii) <= this->mGridIter->idxNodeLastSub(ii)) {
                mIdxScalarNode += mJumpInNumbering.elem(ii);
                break;
            }
        }
    }

    return *this;
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
namespace boost
{
namespace serialization
{
/** Designed to set the boost serialization version of a class template. */
template<std::size_t DIM>
struct version< ScaFES::GridSub<DIM> > {
    /** Sets the version number for serialization. */
    BOOST_STATIC_CONSTANT(unsigned long int, value = 2);
};
} // namespace serialization
} // namespace boost

#endif
