/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_NTUPLE_HPP_
#define SCAFES_NTUPLE_HPP_

#include "ScaFES_Config.hpp"

//#include <cstdint>
#include <iomanip>
#include <iostream>
#include <ios>
#include <vector>
#include <cmath>

#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
namespace boost
{
namespace serialization
{
class access;
}
}
#include <boost/serialization/version.hpp>
#include <boost/serialization/pfto.hpp>
#endif

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class template \code Ntuple<TT,NN> \endcode provides methods for
 * handling \c NN dimensional vectors of a given type \c TT.
 *
 * \n All important unary and binary operators are implemented resp. overloaded
 * and behave as expected. Especially, the most operators work componentwise.
 *
 * \n Usually, the memory for the vector elements must be allocated dynamically,
 * because the number of elements is known at run time. This is related
 * with additional effort for creating and deleting the elements on the
 * heap.
 * Due to the nontype template parameter \c NN the number of vector elements
 * is known at compile time. Thus, it is possible to allocate the memory
 * for the vector elements statically.
 * Optimizations
 * Automatic compiler SIMD vectorizations can be better applied.
 * ==> Using the class template \c Ntuple instead of a class \c Triple
 * (which contains exactly three elements) does not affect the performance,
 * but will give more flexibility in solving other than three-dimensional
 * problems.
 *
 * \n \remarks The defined type \c Index is a specialization of the class
 * \c Ntuple with three elements of type \c int:
 * \code typedef Ntuple<int,3> Index \endcode.
 *
 *\n \code
 * // Constructors.
 * ScaFES::Ntuple<double,3> a(1.1, 2.2, 3.3);  // a = (1.1, 2.2, 3.3)
 * ScaFES::Ntuple<double,3> b(6.0, 6.0, 4.0);  // b = (3.0, 3.0, 2.0)
 * ScaFES::Ntuple<double,3> c = a + 2 * b;     // c = (7.1, 8.2, 7.3)
 *
 * // Arithmetic operators.
 * a += 2.0;                                   // a = (3.1, 4.2, 5.3)
 * c = b * b;                                  // c = (9.0, 9.0, 4.0)
 *
 * // Relational operators
 * a < b;                                      // false: 3.3 > 2.0
 * b *= 2.0;                                   // b = (6.0, 6.0, 4.0)
 * b > a;                                      // true
 * \endcode
 */
template<typename TT, std::size_t NN = 3>
class Ntuple
{

    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Re-export typename TT STL-like. */
        typedef TT value_type;

        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates the default constructor.
         *  All elements are initialized by zero.*/
        Ntuple<TT, NN>();

        /** Creates a special constructor for twodimensional vectors:
         *  The elements are initialized by given values. */
        Ntuple<TT, NN>(TT const& a, TT const& b);

        /** Creates a special constructor for threedimensional vectors:
         *  The elements are initialized by given values. */
        Ntuple<TT, NN>(TT const& a, TT const& b, TT const& c);

        /** Creates own constructor:
         *  All elements are initalized by the same value. */
        Ntuple<TT, NN>(TT const& scal);

        /** Creates own constructor: Conversion method.
         *  All elements are set by a given STL vector. */
        Ntuple<TT, NN>(std::vector<TT> const& vec);

        /** Creates copy constructor.
         * \remarks Use copy-and-swap idiom.
         */
        Ntuple<TT, NN>(Ntuple<TT, NN> const& rhs);

        /** Creates copy assignment operator.
         * All elements of the rhs. vector will be assigned. */
        Ntuple<TT, NN>& operator=(Ntuple<TT, NN> rhs);

        /** Creates the destructor. */
        ~Ntuple() { };

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the dimension (= number of elements) of the n-tuple. */
        size_t dim() const;

        /** Returns a reference to the element at idx \c ii in the n-tuple.
         * \remarks With range check!
         */
        const TT& elem(std::size_t const& ii) const;

        /** Returns the element at idx \c ii in the n-tuple.
         * \remarks Without range check! */
        const TT& operator[](std::size_t const& ii) const;

    public:
        /*----------------------------------------------------------------------
        | SETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns a reference to the element at idx ii in the Ntuple.
         * \remarks With range check!
         */
        TT& elem(std::size_t const& ii);

        /** Returns a reference to the element at idx \c ii in the n-tuple.
         * \remarks Without range check! */
        TT& operator[](std::size_t const& ii);

    public:
        /*----------------------------------------------------------------------
        | ARITHMETIC METHODS.
        ----------------------------------------------------------------------*/
        /** Creates an assignment operator.
         * All elements will be assigned to a given scalar.*/
        Ntuple<TT, NN>& operator=(TT const& sca);

        /** Multiplies this n-tuple by a given scalar (elementwise). */
        template<typename CT>
        Ntuple<TT, NN>& operator*=(CT const& sca);

        /** Divides this n-tuple by a given scalar (elementwise). */
        template<typename CT>
        Ntuple<TT, NN>& operator/=(CT const& sca);

        /** Adds a given rhs n-tuple to this n-tuple. */
        Ntuple<TT, NN>& operator+=(Ntuple<TT, NN> const& rhs);

        /** Subtracts a given rhs n-tuple from this n-tuple. */
        Ntuple<TT, NN>& operator-=(Ntuple<TT, NN> const& rhs);

        /** Multiplies this n-tuple by a given rhs n-tuple. */
        Ntuple<TT, NN>& operator*=(Ntuple<TT, NN> const& rhs);

        /** Divides this n-tuple by a given rhs n-tuple. */
        Ntuple<TT, NN>& operator/=(Ntuple<TT, NN> const& rhs);

    public:
        /*----------------------------------------------------------------------
        | COMPARISON METHODS.
        ----------------------------------------------------------------------*/
        /** Compares elementwise if this n-tuple is smaller than a a given rhs
         *  n-tuple. */
        bool operator<(Ntuple<TT, NN> const& rhs) const;

        /** Compares elementwise if this n-tuple is greater or equal
         *  than a a given rhs n-tuple. */
        bool operator>=(Ntuple<TT, NN> const& rhs) const;

        /** Compares elementwise if this n-tuple is smaller or equal
         *  than a a given rhs n-tuple. */
        bool operator<=(Ntuple<TT, NN> const& rhs) const;

        /** Compares elementwise if this n-tuple is greater than a a given rhs
         *  n-tuple. */
        bool operator>(Ntuple<TT, NN> const& rhs) const;

        /** Compares elementwise if this n-tuple is equal to a given rhs
         *  n-tuple. */
        bool operator==(Ntuple<TT, NN> const& rhs) const;

        /** Compares elementwise if this n-tuple is unequal to a given rhs
         *  n-tuple. */
        bool operator!=(Ntuple<TT, NN> const& rhs) const;

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the index of the element with maximal value. */
        std::size_t idxMaxElem() const;

        /** Computes the dimension of the n-tuple,
         * i.e., all elements will be multiplied. */
        TT size() const;

        /** Computes the 1-norm. */
        double norm1() const;

        /** Serializes this class. */
        template<class Archive>
        void serialize(Archive& ar, unsigned int const version);

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Overloads the output operator \c operator<<.
         * An n-tuple is printed this way: '(a_0, a_1,..., a_{n-1})' */
        template<typename ST, std::size_t MM>
        friend std::ostream& operator<< (std::ostream& output,
                                         Ntuple <ST, MM> const& t);

        /** Method to swap members of two Ntuples. */
        template<typename ST, std::size_t MM>
        friend void swap(Ntuple<ST, MM>& first, Ntuple<ST, MM>& second);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Static array for storing the elements of an n-tuple. */
        TT mElem[NN];

}; // End of class //

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
/** Swaps two objects of the class \c Ntuple. */
template<typename TT, std::size_t NN>
void swap(Ntuple<TT, NN>& first, Ntuple<TT, NN>& second);
/*----------------------------------------------------------------------------*/
/** Preprares writing an object of the class \c Ntuple to output. */
template<typename TT, std::size_t NN>
std::ostream& operator<<(std::ostream& output, Ntuple <TT, NN> const& t);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple()
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] = 0;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple(TT const& scal)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] = scal;
    }
}
/*----------------------------------------------------------------------------*/
// TODO: Only valid for NN=2. --> Partial specialization.
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple(TT const& a, TT const& b)
{
    mElem[0] = a;
    mElem[1] = b;
}
/*----------------------------------------------------------------------------*/
// TODO: Only valid for NN=3. --> Partial specialization.
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple(TT const& a, TT const& b, TT const& c)
{
    mElem[0] = a;
    mElem[1] = b;
    mElem[2] = c;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple(std::vector<TT> const& vec)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] = vec.at(iii);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>::Ntuple(Ntuple<TT, NN> const& rhs)
{
    for (std::size_t iii = 0; iii < rhs.dim(); ++iii) {
        mElem[iii] = rhs.elem(iii);
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator=(Ntuple<TT, NN> rhs)
{
    swap(*this, rhs);
    return *this;
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline std::size_t Ntuple<TT, NN>::dim() const
{
    return NN;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline const TT& Ntuple<TT, NN>::elem(std::size_t const& idx) const
{
    return mElem[idx];
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline const TT& Ntuple<TT, NN>::operator[](std::size_t const& idx) const
{
    return mElem[idx];
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline TT& Ntuple<TT, NN>::elem(std::size_t const& idx)
{
    return mElem[idx];
    //const_cast<TT&>(static_cast<const Ntuple<TT,NN>&>(*this).elem(idx));
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline TT& Ntuple<TT, NN>::operator[] (std::size_t const& idx)
{
    return mElem[idx];
    //const_cast<TT&>(static_cast<const Ntuple<TT,NN>&>(*this)[idx]);
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator<(Ntuple<TT, NN> const& rhs) const
{
    const int DIM = NN;
    bool res = true;

    for (int iii = 0; iii < DIM; ++iii) {
        if (!(mElem[iii] < rhs.elem(iii))) {
            res = false;
            break;
        }
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator<=(Ntuple<TT, NN> const& rhs) const
{
    const int DIM = NN;
    bool res = true;

    for (int iii = 0; iii < DIM; ++iii) {
        if (mElem[iii] > rhs.elem(iii)) {
            res = false;
            break;
        }
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator>=(Ntuple<TT, NN> const& rhs) const
{
    const int DIM = NN;
    bool res = true;

    for (int iii = 0; iii < DIM; ++iii) {
        if ((mElem[iii] < rhs.elem(iii))) {
            res = false;
            break;
        }
    }

    return res;
    //return !(*this < rhs); Does not work because we are working elementwise!
}

/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator>(Ntuple<TT, NN> const& rhs) const
{
    const int DIM = NN;
    bool res = true;

    for (int iii = 0; iii < DIM; ++iii) {
        if (!(mElem[iii] > rhs.elem(iii))) {
            res = false;
            break;
        }
    }

    return res;
    //return !(*this <= rhs); Does not work because we are working elementwise!
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator==(Ntuple<TT, NN> const& rhs) const
{
    const int DIM = NN;
    bool res = true;

    for (int iii = 0; iii < DIM; ++iii) {
        if (!(::fabs(mElem[iii] - rhs.elem(iii)) < 2.2e-15)) {
            res = false;
            break;
        }
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline bool Ntuple<TT, NN>::operator!=(Ntuple<TT, NN> const& rhs) const
{
    return !(*this == rhs);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline std::size_t Ntuple<TT, NN>::idxMaxElem() const
{
    std::size_t idx = 0;
    TT maxElem = mElem[0];

    for (std::size_t iii = 1; iii < NN; ++iii) {
        if (maxElem < mElem[iii]) {
            maxElem = mElem[iii];
            idx = iii;
        }
    }

    return idx;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline TT Ntuple<TT, NN>::size() const
{
    const int DIM = NN;
    std::size_t prod = mElem[0];

    for (int iii = 1; iii < DIM; ++iii) {
        prod *= mElem[iii];
    }

    return prod;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline double Ntuple<TT, NN>::norm1() const
{
    const int DIM = NN;
    double res = ::fabs(mElem[0]);

    for (int iii = 1; iii < DIM; ++iii) {
        res += ::fabs(mElem[iii]);
    }

    return res;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
template<class Archive>
inline void Ntuple<TT, NN>::serialize(Archive& ar, unsigned int const version)
{
    if (1 <= version) {
        ar & (mElem);
    }
}

/*******************************************************************************
 * ARITHMETIC METHODS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator= (TT const& sca)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] = static_cast<TT>(sca);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
template<typename CT>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator*= (CT const& sca)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] *= static_cast<TT>(sca);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
// TODO: Throw an exception if the given scalar is near to zero.
template<typename TT, std::size_t NN>
template<typename CT>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator/= (CT const& sca)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] /= static_cast<TT>(sca);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator+=(Ntuple<TT, NN> const& rhs)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] += rhs.elem(iii);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator-=(Ntuple<TT, NN> const& rhs)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] -= rhs.elem(iii);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator*=(Ntuple<TT, NN> const& rhs)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] *= rhs.elem(iii);
    }

    return *this;
}
/*----------------------------------------------------------------------------*/
// TODO: Throw exception if 'sca' is near to zero.
template<typename TT, std::size_t NN>
inline Ntuple<TT, NN>& Ntuple<TT, NN>::operator/=(Ntuple<TT, NN> const& rhs)
{
    const int DIM = NN;
    for (int iii = 0; iii < DIM; ++iii) {
        mElem[iii] /= rhs.elem(iii);
    }

    return *this;
}

/*******************************************************************************
* FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
template<typename TT, std::size_t NN>
inline void swap(Ntuple<TT, NN>& first, Ntuple<TT, NN>& second)
{
    TT tmp;

    for (std::size_t iii = 0; iii < second.dim(); ++iii) {
        tmp = second.mElem[iii];
        second.mElem[iii] = first.mElem[iii];
        first.mElem[iii] = tmp;
    }
}
/*----------------------------------------------------------------------------*/
template<typename TT, std::size_t NN>
inline std::ostream& operator<<(std::ostream& output, Ntuple <TT, NN> const& t)
{
    const int DIM = NN;
    output << "[";
    for (int iii = 0; iii < DIM; ++iii) {
        output << ::std::setw(4) << ::std::right << t.elem(iii);

        if (DIM - 1 > iii) {
            output << "; ";
        }
    }

    output <<  "]";
    return output;
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
namespace boost
{
namespace serialization
{
/** Designed to set the boost serialization version of a class template. */
template<typename TT, std::size_t NN>
struct version< ScaFES::Ntuple<TT, NN> > {
    /** Sets the version number for serialization. */
    BOOST_STATIC_CONSTANT(unsigned long int, value = 2);
};
}
}

#endif
