/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_PARAMETERS_HPP_
#define SCAFES_PARAMETERS_HPP_

#include "ScaFES_Config.hpp"

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <vector>
#include <string>

#include <boost/serialization/serialization.hpp>

// #ifdef SCAFES_HAVE_BOOST_SERIALIZATION
// namespace boost
// {
// namespace serialization
// {
// class access;
// }
// }
// #include <boost/serialization/vector.hpp>
// #include <boost/serialization/version.hpp>
// #endif

// #ifdef SCAFES_HAVE_BOOST_MPI
#include <boost/mpi.hpp>
/*#include <boost/mpi/communicator.hpp>
#include <boost/mpi/environment.hpp>
#include <boost/mpi/collectives.hpp>*/
// #endif

#ifdef SCAFES_HAVE_BOOST_PROGRAM_OPTIONS
#include <boost/program_options.hpp>
#endif

#ifdef SCAFES_HAVE_BOOST
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#endif

#ifdef SCAFES_HAVE_BOOST_REGEX
#include <boost/regex.hpp>
#endif

#include "ScaFES_Ntuple.hpp"

namespace ScaFES
{

#ifdef SCAFES_HAVE_BOOST_PROGRAM_OPTIONS
namespace po = boost::program_options;
#endif
#ifdef SCAFES_HAVE_BOOST_MPI
namespace boost_mpi = boost::mpi;
#endif

/*******************************************************************************
 * GLOBAL VARIABLES.
 ******************************************************************************/
/** Maximal indentation depth level of console output. */
int gOutputLevelMax = 2;

/** Current indentation depth level of console output. */
int gOutputLevelCurr = gOutputLevelMax;

/*******************************************************************************
 ******************************************************************************/
/** The class \c Parameters stores the parameters given via the
 * command line of a ScaFES application.
 *
 * \n If a command line parameter is invalid, a help message will be displayed.
 *
 * \n The MPI environment lives as long as an instance of this class is living.
 * In particular, if the destructor of this class is called,
 * the MPI environment will be finalized. Thus, if the descructor is called
 * too early, the instance of the calls goes out of scope. So, please be careful
 * with calls to the destructor of this class.
 */
class Parameters
{
    public:
        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own default constructor. */
        Parameters();

        /** Creates own constructor from command line parameters.
         * Parses the command line of a program call and provides a help
         * message.
         */
        Parameters(int argc, char* argv[]);

        /** Creates own copy constructor. */
        Parameters(Parameters const&);

        /** Creates own assignment operator. */
        Parameters& operator=(Parameters rhs);

        /** Creates own destructor. */
        ~Parameters();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
         ----------------------------------------------------------------------*/
        /** Returns the space dimension. */
        int const& dim() const;

        /** Returns the number of grid nodes in each dimension. */
        std::vector<int> const& nNodes() const;

        /** Returns the coordinates of the first grid point. */
        std::vector<double> const& coordNodeFirst() const;

        /** Returns the coordinates of the last grid point. */
        std::vector<double> const& coordNodeLast() const;

        /** Returns the number of partitions. */
        int const& nPartitions() const;

        /** Returns the directions in which the grid should be partitioned. */
        std::vector<bool> const& divideGrid() const;

        /** Returns the decision if the kind field should be read in. */
        bool const& readKindFile() const;

        /** Returns the name of the kind File. */
        std::string const& kindFile() const;

        /** Returns the name of the partition File for writing. */
        std::string const& nameWritePartitionFile() const;

        /** Returns the decision if a partition file should be written. */
        bool const& writePartitionFile() const;

        /** Returns the name of the partition file for reading. */
        std::string const& nameReadPartitionFile() const;

        /** Returns the decision if a partition file should be written. */
        bool const& readPartitionFile() const;

        /** Returns the start of the time interval. */
        double const& timeIntervalStart() const;

        /** Returns the end of the time interval. */
        double const& timeIntervalEnd() const;

        /** Returns time step width. */
        double const& tau() const;

        /** Returns the number of time steps. */
        int const& nTimesteps() const;

        /** Returns the number of snapshots. */
        int const& nSnapshots() const;

        /** Returns the decision if an output file should be written. */
        bool const& writeDataFile() const;

        /** Returns the name of the output file. */
        std::string const& nameDataFile() const;

        /** Returns the name of the config file. */
        std::string const& nameConfigFile() const;

        /** Returns the decision if ADOL-C is enabled. */
        bool const& enabledAdolc() const;

#ifdef SCAFES_HAVE_BOOST_MPI
        /** Returns the MPI communicator. */
        boost_mpi::communicator const& myWorld() const;
#else
        /** Returns the MPI communicator. */
        MPI_Comm const& myWorld() const;
#endif

        /** Returns the maximum level of console output. */
        int const& indentDepth() const;

    public:
        /*----------------------------------------------------------------------
        | SETTER METHODS.
         ----------------------------------------------------------------------*/
        /** Increase output level by one. */
        void increaseLevel();

        /** Decrease output level by one. */
        void decreaseLevel();

    public:
        /*----------------------------------------------------------------------
        | COMPARISON METHODS.
        ----------------------------------------------------------------------*/
        /** Tests if two grids are equal. */
        bool operator==(Parameters const& rhs) const;

        /** Tests if two grids are not equal. */
        bool operator!=(Parameters const& rhs) const;

    public:
        /*----------------------------------------------------------------------
        | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
        ----------------------------------------------------------------------*/
        /** Swaps to elements of this class. */
        friend void swap(Parameters& first, Parameters& second);

        /** Overloads the output operator \c operator<<. */
        friend std::ostream& operator<< (std::ostream& output,
                                         Parameters const& pp);

    private:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Checks a parameter given by its name. */
        template<typename TT>
        void checkParameter(po::variables_map const& vm,
                            std::string const& nameParam,
                            std::vector<TT>& param,
                            bool isNecessary);

        /** Serializes class. */
        template<class Archive>
        void serialize(Archive& ar, const unsigned int version);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Space dimension. */
        int mDim;

        /** Number of grid nodes in each dimension. */
        std::vector<int> mNnodes;

        /** Coordinate of first grid point (left lower point of grid). */
        std::vector<double> mCoordNodeFirst;

        /** Coordinate of last grid point (right upper point of grid). */
        std::vector<double> mCoordNodeLast;

        /** Number of partitions. */
        int mNpartitions;

        /** Divide grid into partitions: Which directions are enabled? */
        std::vector<bool> mDivideGrid;

        /** Read kind field from a given kind file? */
        bool mReadKindFile;

        /** Name of the kind file. */
        std::string mKindFile;

        /** Write partition table to a partition file? */
        bool mWritePartitionFile;

        /** Name of partition file which should be written. */
        std::string mNameWritePartitionFile;

        /** Read partition table from a partition file? */
        bool mReadPartitionFile;

        /** Name of the partition file which should be read in. */
        std::string mNameReadPartitionFile;

        /** Start of time interval [tstart;tend]. */
        double mTimeIntervalStart;

        /** End of time interval [tstart;t_end]. */
        double mTimeIntervalEnd;

        /** Time step width. */
        double mTau;

        /** Number of timesteps. */
        int mNtimesteps;

        /** Number of snapshots: Write data each m timesteps. */
        int mNsnapshots;

        /** Write output file? */
        bool mWriteDataFile;

        /** Name of the output file. */
        std::string mNameOfDataFile;

        /** Name of the configuration file. */
        std::string mNameOfConfigFile;

        /** Is ADOL-C enabled? */
        bool mEnabledAdolc;

#ifdef SCAFES_HAVE_BOOST_MPI
        /** Boost MPI environment. */
        boost_mpi::environment mMyEnv;

        /** Boost MPI communicator. */
        boost_mpi::communicator mMyWorld;
#else
        /** MPI communicator. */
        MPI_Comm mMyWorld;
#endif
}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
inline Parameters::Parameters()
    : mDim(3)
    , mNnodes(3)
    , mCoordNodeFirst(3)
    , mCoordNodeLast(3)
    , mNpartitions(1)
    , mDivideGrid(3)
    , mReadKindFile(false)
    , mKindFile("EMPTY")
    , mWritePartitionFile(false)
    , mNameWritePartitionFile("EMPTY")
    , mReadPartitionFile(false)
    , mNameReadPartitionFile("EMPTY")
    , mTimeIntervalStart(0.0)
    , mTimeIntervalEnd(1.0)
    , mTau(1.0)
    , mNtimesteps(1)
    , mNsnapshots(1)
    , mWriteDataFile(false)
    , mNameOfDataFile("EMPTYDATAFILE")
    , mNameOfConfigFile("EMPTYCONFIGFILE")
    , mEnabledAdolc(true)
#ifdef SCAFES_HAVE_BOOST_MPI    
    , mMyEnv()
#endif
    , mMyWorld()
{
    for (int ii = 0; ii < dim(); ++ii) {
        mNnodes.at(ii) = 8;
        mCoordNodeFirst.at(ii) = 0.0;
        mCoordNodeLast.at(ii) = 1.0;
        mDivideGrid.at(ii) = false;
    }
}
/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(int argc, char* argv[])
    : mDim(3)
    , mNnodes()
    , mCoordNodeFirst()
    , mCoordNodeLast()
    , mNpartitions(1)
    , mDivideGrid()
    , mReadKindFile(false)
    , mKindFile("EMPTY")
    , mWritePartitionFile(false)
    , mNameWritePartitionFile("EMPTY")
    , mReadPartitionFile(false)
    , mNameReadPartitionFile("EMPTY")
    , mTimeIntervalStart(0.0)
    , mTimeIntervalEnd(1.0)
    , mTau(0.0)
    , mNtimesteps(1)
    , mNsnapshots(1)
    , mWriteDataFile(false)
    , mNameOfDataFile("EMPTYDATAFILE")
    , mNameOfConfigFile("EMPTYCONFIGFILE")
    , mEnabledAdolc(true)
#ifdef SCAFES_HAVE_BOOST_MPI
    , mMyEnv(argc, argv)
#endif
    , mMyWorld()
{
    if (0 == myWorld().rank()) {

        po::variables_map vm;

        /*----------------------------------------------------------------------
        | Describe the options of the command line.
        ----------------------------------------------------------------------*/
        po::options_description desc("Allowed options:");
        desc.add_options()
        ("help", "Displays this help message.")
        ("dim", po::value<int>(),
         "Sets the space dimension.")
        ("nNodes", po::value<std::string>(),
         "Sets the number of grid nodes in each dimension (N1x..xNd).")
        ("coordNodeFirst", po::value<std::string>(),
         "Sets the coordinates of the first grid node (c1x..xcd).")
        ("coordNodeLast", po::value<std::string>(),
         "Sets the coordinates of the last grid node (c1x..xcd).")
        ("nPartitions", po::value<int>(),
         "Sets the number of partitions.")
        ("divideGrid", po::value<std::string>(),
         "Divides the grid in which dimension? (g1x..xgd).")
        ("readKindFile", po::value<std::string>(),
         "Reads the kind field from a given kind file.")
        ("readPartitionfile", po::value<std::string>(),
         "Reads the partition table from a given partition file.")
        ("writePartitionfile", po::value<std::string>(),
         "Writes the partition table to a given partition file.")
        ("starttime", po::value<double>(),
         "Sets the start time.")
        ("endtime", po::value<double>(),
         "Sets the end time.")
        ("tau", po::value<double>(),
         "Sets the time step width.")
        ("nTimesteps", po::value<int>(),
         "Sets the number of time steps.")
        ("nSnapshots", po::value<int>(),
         "Sets how many times the data fields are written to output.")
        ("outputfile", po::value<std::string>(),
         "Sets the name of the output data file.")
        ("configfile", po::value<std::string>(),
         "Sets the name of the config file.")
        ("outputlevel", po::value<int>(),
         "Controls the amount of console output.")
        ("enableAdolc", po::value<int>(),
         "Enables(=1) or disables(=0) ADOL-C.");

        /*----------------------------------------------------------------------
        | Try to parse the command line.
        ----------------------------------------------------------------------*/
        try {
            po::store(po::parse_command_line(argc, argv, desc), vm);

            if (vm.count("help")) {
                std::cout << "Command Line: \n" << desc << "\n";
            }

            po::notify(vm);
        } catch (std::exception& e) {
            std::cerr << "\nERROR: Wrong command line options.\n"
                      << e.what() << "\n";
        }

        /*----------------------------------------------------------------------
        | Actions of command line options.
        ----------------------------------------------------------------------*/
        if (vm.count("help")) {
            std::cout << desc << "\n";
        }

        /*----------------------------------------------------------------------
        | Necessary parameters.
        ----------------------------------------------------------------------*/
        if (vm.count("dim")) {
            mDim = vm["dim"].as<int> ();

            if (0 >= dim()) {
                std::cerr << "\nERROR: Space dimension is less or equal"
                          " to zero. Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --dim=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        mNnodes.resize(dim());
        checkParameter(vm, "nNodes", mNnodes, true);

        /*--------------------------------------------------------------------*/
        if (vm.count("nTimesteps")) {
            mNtimesteps = vm["nTimesteps"].as<int> ();

            if (0 >= nTimesteps()) {
                std::cerr << "\nERROR: Number of time steps is less or equal"
                          " to zero. Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --nTimesteps=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        mCoordNodeLast.resize(dim());
        checkParameter(vm, "coordNodeLast", mCoordNodeLast, true);

        /*----------------------------------------------------------------------
        | Optional parameters.
        ----------------------------------------------------------------------*/
        mCoordNodeFirst.resize(dim());
        checkParameter(vm, "coordNodeFirst", mCoordNodeFirst, false);

        if (vm.count("nPartitions")) {
            mNpartitions = vm["nPartitions"].as<int> ();

            if (0 >= mNpartitions) {
                std::cerr << "\nERROR: Number of partitions is less or equal"
                          " to zero. It has to be greater or equal to one."
                          " Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --nPartitions=<integral value>.\n\n";
            MPI_Abort(myWorld(), 0);
        }

        /*--------------------------------------------------------------------*/
        mDivideGrid.resize(dim());
        checkParameter(vm, "divideGrid", mDivideGrid, false);

        /*--------------------------------------------------------------------*/
        if (vm.count("readKindFile")) {
            mKindFile = vm["readKindFile"].as<std::string >();
            mReadKindFile = true;
        } else {
            std::cerr << "\nUse --readKindFile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("readPartitionFile")) {
            mNameReadPartitionFile = vm["readPartitionfile"].as<std::string >();
            mReadPartitionFile = true;
        } else {
            std::cerr << "\nUse --readPartitionfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("writePartitionfile")) {
            mNameWritePartitionFile = vm["writePartitionfile"].as<std::string >();
            mWritePartitionFile = true;
        } else {
            std::cerr << "\nUse --writePartitionfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("starttime")) {
            mTimeIntervalStart = vm["starttime"].as<double> ();

            if (0.0 > timeIntervalStart()) {
                std::cerr << "\nERROR: Start time < 0.0."
                          " Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --starttime=<real value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("endtime")) {
            mTimeIntervalEnd = vm["endtime"].as<double> ();

            if (0.0 >= timeIntervalEnd()) {
                std::cerr << "\nERROR: End time <= 0.0."
                          " Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --endtime=<real value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("tau")) {
            mTau = vm["tau"].as<double> ();

            if (0.0 >= tau()) {
                std::cerr << "\nERROR: Time step width tau <= 0.0.   "
                          " Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else if (vm.count("endtime") && vm.count("nTimesteps")) {
            mTau = (timeIntervalEnd() - timeIntervalStart()) / nTimesteps();
        } else {
            mTau = 1.0;
            std::cout << "Remark: No time step width TAU was specified. It is set"
                      " to the default value 1.0. Proceed... \n\n";
            std::cerr << "\nUse --tau=<real value>.\n\n";
        }

        if (vm.count("tau") && vm.count("endTime") && vm.count("mNtimesteps")) {
            if (mTau != (timeIntervalEnd() - timeIntervalStart()) / nTimesteps()) {
                std::cerr << "\nERROR:  Inconsistency between time step width, "
                          "number of time steps and time interval. Exit...\n\n";
                exit(-1);
            }
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("nSnapshots")) {
            mNsnapshots = vm["nSnapshots"].as<int> ();

            if (0 >= nSnapshots()) {
                std::cerr << "\nERROR: Number of snap shots is less or equal"
                          " to zero. It has to be greater or equal to one."
                          " Exit...\n\n";
                MPI_Abort(myWorld(), 0);
            }
        } else {
            std::cerr << "\nUse --nSnapshots=<integral value>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("outputfile")) {
            mNameOfDataFile = vm["outputfile"].as<std::string > ();
            mWriteDataFile = true;
        } else {
            std::cerr << "\nUse --outputfile=<name>.\n\n";
        }

        /*--------------------------------------------------------------------*/
        if (vm.count("configfile")) {
            mNameOfConfigFile = vm["configfile"].as<std::string > ();
        } else {
            std::cerr << "\nUse --configfile=<name>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("outputlevel")) {
            gOutputLevelMax = vm["outputlevel"].as<int> ();
            gOutputLevelCurr = gOutputLevelMax;
        } else {
            std::cerr << "\nUse --outputlevel=<integral value>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("enableAdolc")) {
            int tmpAdolc = vm["enableAdolc"].as< int >();
            if (tmpAdolc) {
                mEnabledAdolc = true;
            }
            else {
                mEnabledAdolc = false;
            }
        } else {
            // std::cerr << "\nUse --enableAdolc=<0 / 1>.\n\n";
        }

    }

    // Communicate result.
    boost_mpi::broadcast(myWorld(), *this, 0);
}

/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(Parameters const& rhs)
    : mDim(rhs.dim())
    , mNnodes(rhs.nNodes())
    , mCoordNodeFirst(rhs.coordNodeFirst())
    , mCoordNodeLast(rhs.coordNodeLast())
    , mNpartitions(rhs.nPartitions())
    , mDivideGrid(rhs.divideGrid())
    , mReadKindFile(rhs.readKindFile())
    , mKindFile(rhs.kindFile())
    , mWritePartitionFile(rhs.writePartitionFile())
    , mNameWritePartitionFile(rhs.nameWritePartitionFile())
    , mReadPartitionFile(rhs.readPartitionFile())
    , mNameReadPartitionFile(rhs.nameReadPartitionFile())
    , mTimeIntervalStart(rhs.timeIntervalStart())
    , mTimeIntervalEnd(rhs.timeIntervalEnd())
    , mTau(rhs.tau())
    , mNtimesteps(rhs.nTimesteps())
    , mNsnapshots(rhs.nSnapshots())
    , mWriteDataFile(rhs.writeDataFile())
    , mNameOfDataFile(rhs.nameDataFile())
    , mNameOfConfigFile(rhs.nameConfigFile())
    , mEnabledAdolc(rhs.enabledAdolc())
    , mMyWorld(rhs.myWorld())
{  }
/*----------------------------------------------------------------------------*/
inline Parameters::~Parameters()
{ }
/*----------------------------------------------------------------------------*/
inline Parameters& Parameters::operator=(Parameters rhs)
{
    swap(*this, rhs);
    return *this;
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
inline int const& Parameters::dim() const
{
    return mDim;
}
/*----------------------------------------------------------------------------*/
inline std::vector<int> const& Parameters::nNodes() const
{
    return mNnodes;
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeFirst() const
{
    return mCoordNodeFirst;
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeLast() const
{
    return mCoordNodeLast;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nPartitions() const
{
    return mNpartitions;
}
/*----------------------------------------------------------------------------*/
inline std::vector<bool> const& Parameters::divideGrid() const
{
    return mDivideGrid;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readKindFile() const
{
    return mReadKindFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::kindFile() const
{
    return mKindFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writePartitionFile() const
{
    return mWritePartitionFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameWritePartitionFile() const
{
    return mNameWritePartitionFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readPartitionFile() const
{
    return mReadPartitionFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameReadPartitionFile() const
{
    return mNameReadPartitionFile;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalStart() const
{
    return mTimeIntervalStart;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalEnd() const
{
    return mTimeIntervalEnd;
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::tau() const
{
    return mTau;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nTimesteps() const
{
    return mNtimesteps;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nSnapshots() const
{
    return mNsnapshots;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writeDataFile() const
{
    return mWriteDataFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameDataFile() const
{
    return mNameOfDataFile;
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameConfigFile() const
{
    return mNameOfConfigFile;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::enabledAdolc() const
{
    return mEnabledAdolc;
}
/*----------------------------------------------------------------------------*/
#ifdef SCAFES_HAVE_BOOST_MPI
inline boost_mpi::communicator const& Parameters::myWorld() const
{
    return mMyWorld;
}
#else
inline MPI_Comm const& Parameters::myWorld() const
{
    return mMyWorld;
}
#endif
/*----------------------------------------------------------------------------*/
inline int const& Parameters::indentDepth() const
{
    return gOutputLevelCurr;
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
inline void Parameters::increaseLevel()
{
    ++gOutputLevelCurr;
}
/*----------------------------------------------------------------------------*/
inline void Parameters::decreaseLevel()
{
    --gOutputLevelCurr;
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
inline bool Parameters::operator==(Parameters const& rhs) const
{
    bool isEqual = true;

    if (dim() != rhs.dim()) {
         isEqual = false;
    }
    if (nNodes() != rhs.nNodes()) {
         isEqual = false;
    }
    if (coordNodeFirst() != rhs.coordNodeFirst()) {
         isEqual = false;
    }
    if (coordNodeLast() != rhs.coordNodeLast()) {
         isEqual = false;
    }
    if (nPartitions() != rhs.nPartitions()) {
         isEqual = false;
    }
    if (divideGrid() != rhs.divideGrid()) {
         isEqual = false;
    }
    if (readKindFile() != rhs.readKindFile()) {
         isEqual = false;
    }
    if (kindFile() != rhs.kindFile()) {
         isEqual = false;
    }
    if (writePartitionFile() != rhs.writePartitionFile()) {
         isEqual = false;
    }
    if (nameWritePartitionFile() != rhs.nameWritePartitionFile()) {
         isEqual = false;
    }
    if (nameReadPartitionFile() != rhs.nameReadPartitionFile()) {
         isEqual = false;
    }
    if (timeIntervalStart() != rhs.timeIntervalStart()) {
         isEqual = false;
    }
    if (timeIntervalEnd() != rhs.timeIntervalEnd()) {
         isEqual = false;
    }
    if (tau() != rhs.tau()) {
         isEqual = false;
    }
    if (nTimesteps() != rhs.nTimesteps()) {
         isEqual = false;
    }
    if (nSnapshots() != rhs.nSnapshots()) {
         isEqual = false;
    }
    if (writeDataFile() != rhs.writeDataFile()) {
         isEqual = false;
    }
    if (nameDataFile() != rhs.nameDataFile()) {
         isEqual = false;
    }
    if (nameConfigFile() != rhs.nameConfigFile()) {
         isEqual = false;
    }
    if (enabledAdolc() != rhs.enabledAdolc()) {
         isEqual = false;
    }
    if (myWorld() != rhs.myWorld()) {
         isEqual = false;
    }

    return isEqual;
}
/*----------------------------------------------------------------------------*/
inline bool Parameters::operator!=(Parameters const& rhs) const
{
    return !(*this == rhs);
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<typename TT>
inline void Parameters::checkParameter(po::variables_map const& vm,
                                       std::string const& nameParam,
                                       std::vector<TT>& param,
                                       bool isNecessary)
{
    if (vm.count(nameParam)) {
        std::ostringstream ostrRegex;

        for (int ii = 0; ii < dim(); ++ii) {
            ostrRegex << "(-?\\d+)";

            if ((dim() - 1) > ii) {
                ostrRegex << "x";
            }
        }

        std::string strRegex = ostrRegex.str();
        boost::regex re(strRegex);
        std::string strParam;
        boost::cmatch matches;

        strParam = vm[nameParam].as<std::string > ();
        if (boost::regex_match(strParam.c_str(), matches, re)) {
            for (std::size_t i = 1; i < matches.size(); ++i) {
                std::string match(matches[i].first, matches[i].second);
                param.at(i - 1) = atoi(match.c_str());
            }
        } else {
            std::ostringstream ostrErr;

            for (int ii = 0; ii < dim(); ++ii) {
                ostrErr << "n" << ii;

                if ((dim() - 1) > ii) {
                    ostrErr << "x";
                }
            }

            std::string strErr = ostrErr.str();
            std::cerr << "\nERROR: The parameter --" << nameParam << "="
                      << strParam
                      << " is not of a format like"
                      " --" << nameParam
                      << "=" << strErr << ". Exit... \n\n";
            MPI_Abort(myWorld(), 0);
        }
    } else {
        std::cerr << "\nUse --" << nameParam << "=<a1xa2xa3>\n\n";

        if (isNecessary) {
            MPI_Abort(myWorld(), 0);
        }
    }
}
/*----------------------------------------------------------------------------*/
template<class Archive>
void Parameters::serialize(Archive& ar, const unsigned int version)
{
    if (1 <= version) {
        ar & (mDim);
        ar & (mNnodes);
        ar & (mCoordNodeFirst);
        ar & (mCoordNodeLast);
        ar & (mNpartitions);
        ar & (mDivideGrid);
        ar & (mReadKindFile);
        ar & (mKindFile);
        ar & (mWritePartitionFile);
        ar & (mNameWritePartitionFile);
        ar & (mReadPartitionFile);
        ar & (mNameReadPartitionFile);
        ar & (mTimeIntervalStart);
        ar & (mTimeIntervalEnd);
        ar & (mNtimesteps);
        ar & (mNsnapshots);
        ar & (mWriteDataFile);
        ar & (mNameOfDataFile);
        ar & (mNameOfConfigFile);
        ar & (mEnabledAdolc);
        ar & (mTau);
    }
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
inline void swap(Parameters& first, Parameters& second)
{
    std::swap(first.mDim, second.mDim);
    std::swap(first.mNnodes, second.mNnodes);
    std::swap(first.mCoordNodeFirst, second.mCoordNodeFirst);
    std::swap(first.mCoordNodeLast, second.mCoordNodeLast);
    std::swap(first.mNpartitions, second.mNpartitions);
    std::swap(first.mDivideGrid, second.mDivideGrid);
    std::swap(first.mReadKindFile, second.mReadKindFile);
    std::swap(first.mKindFile, second.mKindFile);
    std::swap(first.mWritePartitionFile, second.mWritePartitionFile);
    std::swap(first.mNameWritePartitionFile, second.mNameWritePartitionFile);
    std::swap(first.mReadPartitionFile, second.mReadPartitionFile);
    std::swap(first.mNameReadPartitionFile, second.mNameReadPartitionFile);
    std::swap(first.mTimeIntervalStart, second.mTimeIntervalStart);
    std::swap(first.mTimeIntervalEnd, second.mTimeIntervalEnd);
    std::swap(first.mTau, second.mTau);
    std::swap(first.mNtimesteps, second.mNtimesteps);
    std::swap(first.mNsnapshots, second.mNsnapshots);
    std::swap(first.mWriteDataFile, second.mWriteDataFile);
    std::swap(first.mNameOfDataFile, second.mNameOfDataFile);
    std::swap(first.mNameOfConfigFile, second.mNameOfConfigFile);
    std::swap(first.mEnabledAdolc, second.mEnabledAdolc);
    std::swap(first.mMyWorld, second.mMyWorld);
}
/*----------------------------------------------------------------------------*/
inline std::ostream& operator<< (std::ostream& output, Parameters const& pm)
{
    if (0 == pm.myWorld().rank()) {
        output << "* Print parameters... "
               << std::endl;
        output << "   o Coordinate(left lower node)     = (";

        for (int ii = 0; ii < pm.dim(); ++ii) {
            output << boost::format("%4.2f") % pm.coordNodeFirst().at(ii);

            if ((pm.dim() - 1) > ii) {
                output << ";";
            }
        }

        output << ")" << std::endl;
        output << "   o Coordinate(right upper node)    = (";

        for (int ii = 0; ii < pm.dim(); ++ii) {
            output << boost::format("%4.2f") % pm.coordNodeLast().at(ii);

            if ((pm.dim() - 1) > ii) {
                output << ";";
            }
        }

        output << ")" << std::endl;
        output << "   o End time                        = "
               << std::setw(8) << std::scientific << pm.timeIntervalEnd()
               << std::endl;
        output << "   o #{Time steps}                   = "
               << std::setw(8) << pm.nTimesteps() <<  std::right << std::endl;
        output << "   o #{Snapshots}                    = "
               << std::setw(8) << pm.nSnapshots()
               << std::right << std::endl;
        output << "   o Write output file?              = ";

        if (pm.writeDataFile()) {
            output << std::setw(8) << "YES" << std::right << std::endl;
        } else {
            output << std::setw(8) << "NO" << std::right << std::endl;
        }

        output << "   o Name of output file =  "
               << pm.nameDataFile() << std::endl;
        output << "  Printed." << std::endl;
    }

    return output;
}

/*******************************************************************************
 * FREE FUNCTIONS.
 ******************************************************************************/
/** Free function which sets a prefix to the output of process with
 * identifier \c rank. */
inline std::string getPrefix(int const& indentationDepth,
                             int const& rank)
{
    std::string blanks("     ");
    std::ostringstream blankStream;
    blankStream << "ScaFES: Rank=" << rank << ": ";
    int nIndents = gOutputLevelMax - indentationDepth;
    for (int jj = 0; jj < nIndents; ++jj) {
        blankStream << blanks;
    }
    return blankStream.str();
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
BOOST_CLASS_VERSION(ScaFES::Parameters, 2);

#endif
