/** ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_PROBLEM_HPP_
#define SCAFES_PROBLEM_HPP_

#include "ScaFES_Config.hpp"

#include <iostream>
#include <ios>
#include <fstream>
#include <string>
#include <cassert>
#include <unordered_map>

//#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
//namespace boost
//{
//namespace serialization
//{
//class access;
//}
//}
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/serialization/serialization.hpp>

//     #include <boost/archive/text_oarchive.hpp>
//     #include <boost/archive/text_iarchive.hpp>
//#include <boost/serialization/version.hpp>
//#include <boost/serialization/pfto.hpp>
//#include <boost/serialization/vector.hpp>
//#endif

//#ifdef SCAFES_HAVE_BOOST
#include <boost/mpi.hpp>
//#endif

#ifdef SCAFES_HAVE_BOOST
#include <boost/type_traits.hpp>
#include <boost/assign/list_of.hpp>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

extern "C" {
#ifdef _OPENMP
#include <omp.h>
#endif
}

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Parameters.hpp"
#include "ScaFES_GridSub.hpp"
#include "ScaFES_GridGlobal.hpp"
#include "ScaFES_DataField.hpp"
#ifdef SCAFES_HAVE_CUDA
#include "ScaFES_ProblemCuda.hpp"
#endif

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class \c Problem is designed to describe a problem (and also solve)
 * on a given MPI partition. A problem usually consists of several data fields
 * (some of them are unknown, some of them are given).
 *
 * \n The decomposition of the computational domain and of the physical data
 * fields is not obliged to happen in the same way, i.e. the domain and the
 * data fields within this domain can have different grids!
 *
 * \n In order to solve a particular problem, the user has to implement an own
 * class that is inherited by the class \c Problem and has to overwrite all
 * necessary methods (those which are marked as virtual).
 * Especially, the users should pay attention to the pure virtual method
 * \c computeTimestep().
 * Since this is a framework for explicit solvers, ScaFES provides the
 * method \c iterate() to handle the iteration over the time steps
 * (forward computation). The method \c computeTimestep() is declared
 * as virtual, and has no definition in ScaFES.
 * The definition depends on the concrete problem and has to be done
 * by the user.
 *
 * \todo
 * <ul>
 * <li> Communicators are used for synchronizing data with the neighbours. This
 * is necessary if there is one process that should do the communication and
 * a bunch of workers. </li>
 * </ul>
 *
 * \n Getter methods (related to the given parameter set).
 * Idea: All parameters which are set via the command line (and stored
 * in an object of class \c Parameter) should be available in the
 * user class (\c Problem_HeatEqnFDM, e.g.), i.e. in all subclasses of
 * the class \c Problem.
 * Thus, the corresponding getter methods (here:
 * the getter methods of the class \c Parameters will be wrapped) will be
 * provided.
 * 
 * \n The kind field is of type \c scafes_gridnode_type.
 * This type consists of 16 bit.
 * The mimic for kind field flags is \c CGXYZxyzttt...
 * \c ttt are reserved for problem specific flags that can be used in user code.
 * \n The six flags \c XYZxyz are intended to provide information concerning the
 * orientation of the outer normal vector for cells at the global border.
 * If the outer vector has a -1 in the first, second or third dimension,
 * then the bit X, Y or Z is set.
 * If the outer normal vector has a +1 in one dimension, then the bit x, y or z
 * is set.
 * Flag \c G stands for \c SCAFES_GLOBAL_BORDER_NODE...
 * DIRxM represents a -1 in dimension x of outer vector.
 * DIRxP represents a +1 in dimension x of outer vector.
 *
 *   1000 0000 0000 0000    32768  --> 0x8000
 *   0100 0000 0000 0000    16384  --> 0x4000
 *   0010 0000 0000 0000     8192  --> 0x2000
 *   0001 0000 0000 0000     4096  --> 0x1000
 *   0000 1000 0000 0000     2048  --> 0x800
 *   0000 0100 0000 0000     1024  --> 0x400
 *   0000 0010 0000 0000      512  --> 0x200
 *   0000 0001 0000 0000      256  --> 0x100
 *   0000 0000 1000 0000      128  --> 0x80
 *   0000 0000 0100 0000       64  --> 0x40
 *   0000 0000 0010 0000       32  --> 0x20
 *   0000 0000 0001 0000       16  --> 0x10
 *   0000 0000 0000 1000        8  --> 0x08
 *   0000 0000 0000 0100        4  --> 0x04
 *   0000 0000 0000 0010        2  --> 0x02
 *   0000 0000 0000 0001        1  --> 0x01
 *
 * \n Bitwise operators:
 *      AND (a & b): Bits that are set in both a and b are set.
 *      OR: (a | b): Bits that are set in either a or b are set.
 *      XOR (a ^ b): Bits that are set in a or b but not both are set.
 *      NOT (~a)   : Bits in a are complemented (inversed)
 *
 *  \n //TODO: Parallelize the partitioning of the domain:
 *   \code for (int ii = 0; ii < nPartitions; ++ii) {
 *              gg->addPartitions(slice[ii]);
 *         } \endcode
 *
 * Case: If there would exist more than 1 update() method:
 * The information which tape id and which independent field
 * should be connected to the dependent data field.
 * Total tape id for initialisation step.
 * \remarks The tape id must be different
 * for all data fields and all tracing parts
 * (initialisation, computation, optimization).
 * => Count all tapes which are created within one problem
 * and distribute the id to the different tracing parts of the added
 * data fields.
 *
 * \n The method \c iterate will process the following steps: 
 * Compute the new values of all data fields at all grid nodes, synchronise
 * values with neighboured partitions, redefine new situation (new data field
 * will be the old onein the next time step). It is sufficient to swap the
 * pointers to the memories.
 * 
 * \n The kind field will be initialised using the methods
 * \c setGridGlobalFlags and \c setGridPartitionFlags. For each grid node
 * of the global grid the type of grid node will be determined ant stored in
 * the data field \c mKind.
 */
template<class OWNPRBLM, typename CT = double, std::size_t DIM = 3>
class Problem
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for integer d-dimensional tuple. */
        typedef Ntuple<int, DIM> Index;

        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own default constructor. */
        Problem();

        /** Creates own constructor.
         * First, creates a partition file.
         * Second, runs the program based on the created partition file.
         * Third, generates the global grid from command line data.
         * Sets MPI-related stuff (communicator, MPI size,  MPI rank)
         * and adopts the global grid from the command line parameters.
         * Then, the master process broadcasts the parameters.
         * \param cl Command line parameters.
         * \param gg Global grid.
         */
        Problem(Parameters const& cl, GridGlobal<DIM> const& gg);

        /** Creates compiler provided copy constructor. */
        Problem(Problem<OWNPRBLM, CT, DIM> const& ppp);

        /** Creates compiler provided assignment operator. */
        Problem& operator=(Problem<OWNPRBLM, CT, DIM> const&) = default;

        /** Creates own destructor. */
        ~Problem();

    public:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the parameters. */
        Parameters const& params() const;

        /** Returns the grid size in a given direction. */
        double const& gridsize(int const& idx) const;

        /** Returns the time step size. */
        double const& tau() const;

        /** Returns the rank of certain process in MPI-communicator. */
        int myRank() const;

        /** Returns the number of data fields for the communicator. */
        int nDfComm() const;

        /** Returns the number of time steps. */
        int const& nTimesteps() const;

        /** Returns the pointer to global grid. */
        GridGlobal<DIM> const& globalGrid() const;

        /** Returns the node type of a given global node. */
        enum GridGlobal<DIM>::scafes_gridnode_type const& kind(Index const& idxNode) const;

        /** Returns the number of nodes in each direction. */
        Index const& nNodes() const;

        /** Returns the number of nodes in a given direction. */
        int const& nNodes(int const& dim) const;

        /** Returns the coordinates of a given global node. */
        Ntuple<double,DIM> coordinates(Index const& idxNode) const;

        /** Returns the global local node of a given global node which
         * resides in a given direction. */
        Index connect(Index const& idxNode, int const& dir);

        /** Returns the new vector of all data fields. */
        std::vector<DataField< CT, DIM > > const& vectNew() const;

        /** Returns the old vector of all data fields. */
        std::vector<DataField< CT, DIM > > const& vectOld() const;

        /** Returns the value of the known data field given by \c idx
         * at the grid node with node number \c idxNode.
         */
        CT const& knownDf(const int& idx, const Index& idxNode) const;

    public:
        /*----------------------------------------------------------------------
         | SETTER METHODS.
         ----------------------------------------------------------------------*/
        /** Increase output level by one. */
        void increaseLevel();

        /** Decrease output level by one. */
        void decreaseLevel();

    public:
        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Calls barrier for MPI-communicator.
         * Returns, if all processes have entered it. */
        void worldBarrier();

        /** Sets up the kind field w.r.t. the global grid
         * (global inner node, global border node). */
        void setGridGlobalFlags();

        /** Sets up the kind field w.r.t. the underlying grid partition:
         * (local inner node, local border node). */
        void setGridPartitionFlags(int const& maxStencilWidth);

        /** Iterates over time. */
        void iterate();

        /** Iterates over time given the geometrical parameters.
         * Data fields will be initialised depending on a given set of
         * geometrical parameters. */
        void iterate(std::vector<CT> const& pars);

        /** Iterates over time given initial data fields.
         * Data fields will be initialised depending on a given set of
         * data fields. */
        void iterate(std::vector<DataField<CT, DIM> > const& vect);

        /** Iterates over time. */
        void iterateInternal();

        /** Initialises the data fields using ADOL-C tapes. */
        void evalInit(std::vector< DataField<CT,DIM> >& dfDep,
                      std::vector<CT> const& dfIndep);

        /** Initialises all data fields depending on the function
         * implemented in the sub class.
         */
        void init(std::vector< DataField<CT,DIM> >& dfDep,
                  std::vector<CT> const& dfIndep);

        /** Initialises all data fields depending on the function
         * implemented in the sub class.
         */
        void init(std::vector< DataField<CT,DIM> >& dfDep,
                  std::vector< DataField<CT,DIM> > const& dfIndep);

        /** Updates the data fields using the methods \c updateGlobalInner
         * and \c updateGlobalBorder at a given time step. */
        void update(std::vector< DataField<CT,DIM> >& dfDep,
                    std::vector< DataField<CT,DIM> > const& dfIndep,
                    int const& timestep);

        /** Updates the data fields at all global inner grid nodes
         *  evaluating the method \c updateInner at a given time step. */
        void updateGlobalInner(std::vector< DataField<CT,DIM> >& dfDep,
                         std::vector< DataField<CT,DIM> > const& dfIndep,
                         int const& timestep);

        /** Updates the data fields at all global inner grid nodes
         *  evaluating the method \c updateBorder at a given time step. */
        void updateGlobalBorder(std::vector< DataField<CT,DIM> >& dfDep,
                         std::vector< DataField<CT,DIM> > const& dfIndep,
                         int const& timestep);

        /** INTERNAL: Executes Cuda kernel. */
//         void kernelUpdateInnerCuda(DataField<short int,DIM> const& mKind,
//                                std::vector< DataField<CT,DIM> >& dfDep,
//                                std::vector< DataField<CT,DIM> > const& dfIndep,
//                                int const& timestep,
//                                int const nodeType,
//                                int const nNodesTotalPartition);

       /** INTERNAL: Calls Cuda kernel. */
//         void updateGlobalInnerUsingCuda(DataField<short int,DIM> const& mKind,
//                                std::vector< DataField<CT,DIM> >& dfDep,
//                                std::vector< DataField<CT,DIM> > const& dfIndep,
//                                int const& timestep,
//                                int const nodeType,
//                                int const nNodesTotalPartition);
        
        /** Updates the data fields at all local inner grid nodes
         *  evaluating the method \c updateInner at a given time step. */
        void updateLocalInner(std::vector< DataField<CT,DIM> >& dfDep,
                              std::vector< DataField<CT,DIM> > const& dfIndep,
                              int const& timestep);

        /** Updates the data fields at all local inner grid nodes
         *  evaluating the method \c updateBorder at a given time step. */
        void updateLocalBorder(std::vector< DataField<CT,DIM> >& dfDep,
                               std::vector< DataField<CT,DIM> > const& dfIndep,
                               int const& timestep);

        /** Adds a new variable to the problem.
         * Two data fields will be created: One for the last time step,
         * a second one for the current time step. */
        void addVariable(std::string const& nameVar,
                         int const& stencilWidth);

        /** Adds a new variable to the problem.
         * Two data fields will be created: One for the last time step,
         * a second one for the current time step.
         * The new data field will be defined on on the boundary of the
         * global grid, only. */
        void addVariable(std::string const& nameVar,
                         int const& stencilWidth,
                         int const& nLayers);

        /** Adds a new variable to the problem.
         * Two data fields will be created: One for the last time step,
         * a second one for the current time step. */
        template<typename PT>
        void addVariable(std::string const& nameVar,
                         PT& prblm,
                         int const& stencilWidth,
                         void (PT::*fptrSol)(CT&,
                               Ntuple<CT, DIM> const&,
                               CT const&) = NULL);

        /** Adds a new data field to the problem.
         * Two data fields will be created: One for the last time step,
         * a second one for the current time step. */
        void addDataField(std::string const& nameDf,
                          int const& stencilWidth,
                          void (*fptrInit)(CT&,
                                Ntuple<CT, DIM> const&,
                                CT const&) = NULL,
                          bool const& isKnown = false);
        
        /** Adds a new data field to the problem.
        * Two data fields will be created: One for the last time step,
        * a second one for the current time step. */
        template<typename PT>
        void addDataField(std::string const& nameDf,
                          int const& stencilWidth,
                          PT& prblm,
                          void (PT::*fptrInit)(CT&,
                                Ntuple<CT, DIM> const&,
                                CT const&) = NULL,
                          bool const& isKnown = false);

//         /** Serializes this class. */
//         template<class Archive>
//         void serialize(Archive& ar, const unsigned int version);

    private:
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Parameters of the problem. */
        Parameters mParams;

        /** List of interfaces to the neighbours. */
        /** Pointer to related global grid to access partitions and more. */
        GridGlobal<DIM> mGG;

        /** Determines the type of each grid node. */
        DataField<short int, DIM> mKind;

        /** List of interfaces to the neighbours. */
        // std::vector<Communicator> vCommunicator;

        /** Unordered map of registered data fields. */
        //std::unordered_map<std::string, DataField<CT,DIM>&> mMapsOfFields;

        /** Vector of registered data fields. */
        std::vector<std::string> mRegisteredDf;

        /** Global output file. */
        DataFile<CT, DIM> mGlobalOutput;

        /** Numerical solution at last time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectOld;

        /** Numerical solution at current time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectNew;

        /** Exact solution at current time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectSol;

        /** Vector of known data fields. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectKnownDf;

        /** Is exact solution known? */
        bool mIsSolutionKnown;

        /** Function pointer to function of exact solution. */
        void (OWNPRBLM::*mFuncSol)(CT&, Ntuple<CT, DIM> const&, CT const&);

        /** Function pointer to a FREE function of the exact solution. */
        void (*mFuncFreeSol)(CT&, Ntuple<CT, DIM> const&, CT const&);
        
        /** Total time for complete problem. */
        boost::mpi::timer mTimerTotal;
}; // End of class. //

/*******************************************************************************
 * DECLARATIONS OF FREE METHODS.
 ******************************************************************************/
/** Sets the vector of data fields by a given array. */
template<typename CT, std::size_t DIM>
void setVectorDf2ArrayMemAll(CT*& array,
                       std::vector<DataField<CT, DIM> > const& vect,
                       std::vector<bool> const& transfComp);
/*----------------------------------------------------------------------------*/
/** Sets a array by a given vector of data fields. */
template<typename CT, std::size_t DIM>
void setArray2VectorDfMemAll(std::vector<DataField<CT, DIM> >& vect,
                       CT* const& array,
                       std::vector<bool> const& transfComp);
/*----------------------------------------------------------------------------*/
/** Sets the vector of data fields by a given array. */
template<typename CT, std::size_t DIM>
void setVectorDf2ArrayMemNormal(CT*& array,
                       std::vector<DataField<CT, DIM> > const& vect,
                       std::vector<bool> const& transfComp);
/*----------------------------------------------------------------------------*/
/** Sets a array by a given vector of data fields. */
template<typename CT, std::size_t DIM>
void setArray2VectorDfMemNormal(std::vector<DataField<CT, DIM> >& vect,
                       CT* const& array,
                       std::vector<bool> const& transfComp);

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem()
    : mParams()
    , mGG()
    , mKind()
    , mRegisteredDf()
    , mGlobalOutput()
    , mVectOld()
    , mVectNew()
    , mVectSol()
    , mVectKnownDf()
    , mIsSolutionKnown(false)
    , mFuncSol()
    , mFuncFreeSol()
    , mTimerTotal()
{  }
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem(Parameters const& cl,
                                           GridGlobal<DIM> const& gg)
    : mParams(cl)
    , mGG(gg)
    , mKind("kind", cl, gg, 0)
    , mRegisteredDf()
    , mGlobalOutput()
    , mVectOld()
    , mVectNew()
    , mVectSol()
    , mVectKnownDf()
    , mIsSolutionKnown(false)
    , mFuncSol()
    , mFuncFreeSol()
    , mTimerTotal()
{
#ifdef VTRACE
    VT_TRACER("ScaFES::Problem(cl,gg)");
#endif
    
    mTimerTotal.restart();    

    // Read in partition file. //
    if (cl.readPartitionFile()) {
        try {
            std::ifstream ifs(cl.nameReadPartitionFile().c_str());
            boost::archive::binary_iarchive ia(ifs);
            ia >> (mGG);

            // The partition file is okay.
            // Check if number of partitions == number of MPI procs.
            if (cl.myWorld().size() != mGG.nPartitions()) {
                std::cerr << "\nERROR: Number of partitions in the "
                          "partition file (" << mGG.nPartitions()
                          << ") != mpiSize(). Exit...\n";
            }
        } catch (std::exception& e) {
            std::cerr << "\nERROR: Could not open partition file "
                      << cl.nameReadPartitionFile() << ". Exit... \n";
        }
    }

    // Serialize the global grid into an archive and write file.
    if (cl.writePartitionFile()) {
        std::ofstream ofs(cl.nameWritePartitionFile().c_str());
        boost::archive::binary_oarchive oa(ofs);

        try {
            oa << (mGG);
        } catch (...) {
            std::cerr << "Something wrong in serializing the global grid.\n";
        }

        ofs.close();
    }

    // Create the name of the partition file including the dimensions
    // of the global grid.
    std::ostringstream tmpStringstream;

    for (std::size_t ii = 0; ii < DIM; ++ii) {
        tmpStringstream << "_" + mGG.nNodes().elem(ii);
    }

    std::string strNnodes = tmpStringstream.str();
    std::string namePartitionfile = cl.nameWritePartitionFile()
                                    + strNnodes + ".partfile";

    // Read in a given kind file?
    if (cl.readKindFile()) {
        mKind.read(0);
    } else {
        setGridGlobalFlags();
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem(Problem<OWNPRBLM, CT, DIM> const& from)
    : mParams(from.params())
    , mGG(from.globalGrid())
    , mKind(from.mKind)
    , mRegisteredDf(from.mRegisteredDf)
    , mGlobalOutput(from.mGlobalOutput)
    , mVectOld(from.mVectOld)
    , mVectNew(from.mVectNew)
    , mVectSol(from.mVectSol)
    , mVectKnownDf(from.mVectKnownDf)
    , mIsSolutionKnown(from.mIsSolutionKnown)
    , mFuncSol(from.mFuncSol)
    , mFuncFreeSol(from.mFuncFreeSol)    
    , mTimerTotal(from.mTimerTotal)
{ }
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::~Problem()
{ }

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Parameters const& Problem<OWNPRBLM, CT, DIM>::params() const
{
    return mParams;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::myRank() const
{
    return mParams.myWorld().rank();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nTimesteps() const
{
    return mParams.nTimesteps();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline double const& Problem<OWNPRBLM, CT, DIM>::gridsize(int const& idx) const
{
    return (mGG.gridsize(idx));
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline double const& Problem<OWNPRBLM, CT, DIM>::tau() const
{
    return mParams.tau();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline GridGlobal<DIM> const& Problem<OWNPRBLM, CT, DIM>::globalGrid() const
{
    return mGG;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline enum GridGlobal<DIM>::scafes_gridnode_type const& Problem<OWNPRBLM, CT, DIM>::kind(Index const& idxNode) const
{
    return mKind[idxNode];
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline typename Problem<OWNPRBLM, CT, DIM>::Index const& Problem<OWNPRBLM, CT, DIM>::nNodes() const
{
    return mGG.nNodes();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nNodes(int const& dim) const
{
    return mGG.nNodes().elem(dim);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Ntuple<double,DIM> Problem<OWNPRBLM, CT, DIM>::coordinates(Index const& idxNode) const
{
    return mGG.coordinates(idxNode);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline typename Problem<OWNPRBLM, CT, DIM>::Index Problem<OWNPRBLM, CT, DIM>::connect(Index const& idxNode,
        int const& dir)
{
    return mGG.neigh(idxNode, dir);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT, DIM> > const& Problem<OWNPRBLM, CT, DIM>::vectNew() const
{
    return mVectNew;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT, DIM> > const& Problem<OWNPRBLM, CT, DIM>::vectOld() const
{
    return mVectOld;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline CT const& Problem<OWNPRBLM, CT, DIM>::knownDf(const int& idx,
                                                     const Index& idxNode) const
{
    return mVectKnownDf[idx](idxNode);
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::increaseLevel()
{
    mParams.increaseLevel();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::decreaseLevel()
{
    mParams.decreaseLevel();
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::worldBarrier()
{
    mParams.myWorld().barrier();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setGridGlobalFlags()
{
    // Default: Set all nodes as MPI global inner nodes.
    for (typename GridSub<DIM>::iterator it = mKind.memNormal().begin(),
        et = mKind.memNormal().end(); it < et; ++it) {
        mKind(it.idxNode()) = globalGrid().SCAFES_GLOBAL_INNER_NODE;
    }
    Index idxNode;
    Index idxNodeFirstGlobal = globalGrid().idxNodeFirst();
    Index idxNodeLastGlobal = globalGrid().idxNodeLast();
    Index idxNodeFirstPart
         = globalGrid().partition(params().myWorld().rank()).idxNodeFirstSub();
    Index idxNodeLastPart
         = globalGrid().partition(params().myWorld().rank()).idxNodeLastSub();
    for (typename GridSub<DIM>::iterator it = mKind.memNormal().begin(),
          et = mKind.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();

        for (std::size_t ii = 0; ii < DIM; ++ii) {
            if ( (idxNodeFirstGlobal.elem(ii) == idxNode.elem(ii))
                || (idxNodeLastGlobal.elem(ii) == idxNode.elem(ii)) ) {
                mKind(idxNode) = globalGrid().SCAFES_GLOBAL_BORDER_NODE;
                break;
            }
        }
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setGridPartitionFlags(int const& maxStencilWidth)
{
    // Default: Set all nodes as MPI local inner nodes.
    for (typename GridSub<DIM>::iterator it = mKind.memNormal().begin(),
        et = mKind.memNormal().end(); it < et; ++it) {
        mKind(it.idxNode()) += globalGrid().SCAFES_PARTITION_INNER_NODE;
    }
    Index idxNode;
    Index idxNodeFirstGlobal = globalGrid().idxNodeFirst();
    Index idxNodeLastGlobal = globalGrid().idxNodeLast();
    Index idxNodeFirstPart
          = globalGrid().partition(params().myWorld().rank()).idxNodeFirstSub();
    Index idxNodeLastPart
          = globalGrid().partition(params().myWorld().rank()).idxNodeLastSub();
    for (typename GridSub<DIM>::iterator it = mKind.memNormal().begin(),
        et = mKind.memNormal().end(); it < et; ++it) {
        idxNode = it.idxNode();

        for (std::size_t ii = 0; ii < DIM; ++ii) {
            if ( ((idxNodeFirstPart.elem(ii) + maxStencilWidth) > idxNode.elem(ii))
               || ((idxNodeLastPart.elem(ii) - maxStencilWidth) < idxNode.elem(ii)) ) {
                mKind(idxNode) += globalGrid().SCAFES_PARTITION_BORDER_NODE;
                mKind(idxNode) -= globalGrid().SCAFES_PARTITION_INNER_NODE;
                break;
            }
        }
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterate()
{
    std::vector<CT> pars;
    iterate(pars);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterate(std::vector<CT> const& pars)
{
    init(mVectOld, pars);
    init(mVectNew, pars);
    for (std::size_t ii = 0; ii < mVectOld.size(); ++ii) {
        mVectOld[ii].sync(0);
    }
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        mVectNew[ii].sync(0);
    }

    iterateInternal();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterate(std::vector<DataField<CT,DIM> > const& vectInit)
{
    init(mVectOld, vectInit);
    init(mVectNew, vectInit);

    for (std::size_t ii = 0; ii < mVectOld.size(); ++ii) {
        mVectOld[ii].sync(0);
    }
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        mVectNew[ii].sync(0);
    }

    iterateInternal();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterateInternal()
{
    double tIterate = 0.0;
    double tInit = 0.0;
    double tUpdate = 0.0;
    double tSync = 0.0;
    double tWrite = 0.0;
    double tSolution = 0.0;
    boost::mpi::timer timerIterate;
    boost::mpi::timer timerInit;
    boost::mpi::timer timerUpdate;
    boost::mpi::timer timerSync;
    boost::mpi::timer timerWrite;
    boost::mpi::timer timerSolution;
    int currTimestep;

    timerIterate.restart();
#ifdef VTRACE
    unsigned int mid;
    mid = VT_MARKER_DEF("Prblm_iterate", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Problem: Start Loop time steps");
#endif

    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Iterate over time..." << std::endl;
    }
    mParams.decreaseLevel();
    int maxStencilWidth = 0;
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        if (maxStencilWidth < mVectNew[ii].stencilWidth()) {
            maxStencilWidth = mVectNew[ii].stencilWidth();
        }
    }
    setGridPartitionFlags(maxStencilWidth);

    timerWrite.restart();
    std::vector<std::string> tmpNames;
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        tmpNames.push_back(mVectNew[ii].name());
    }

    std::ostringstream tmpStringstream;
    tmpStringstream << params().nameDataFile();
    tmpStringstream << "_ALL";

    mGlobalOutput = DataFile<CT, DIM>(tmpStringstream.str(), tmpNames,
                                          params().myWorld(), 
                                          globalGrid().nNodes());
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        mVectNew[ii].write(0);
    }
    mKind.write(0);
    std::vector<CT*> tmpElemData;
    for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
        tmpElemData.push_back(mVectNew[ii].elemData());
    }
    mGlobalOutput.write(tmpElemData,
                        mVectNew[0].memNormal(),
                        0);
    tWrite += timerWrite.elapsed();

    for (currTimestep = 1; currTimestep <= nTimesteps(); ++currTimestep) {
        if ((0 == myRank()) && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), myRank())
                      << " * Process TIME STEP=" << currTimestep
                      << "..." << std::endl;
        }
        mParams.decreaseLevel();

        timerUpdate.restart();
        updateLocalBorder(mVectNew, mVectOld, currTimestep);
        tUpdate += timerUpdate.elapsed();
        timerSync.restart();
        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].copyValuesFromMemCommToSendBuffer(currTimestep);
        }
        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].exchangeValuesInBuffers(currTimestep);
        }
        tSync += timerSync.elapsed();
        timerUpdate.restart();
        updateLocalInner(mVectNew, mVectOld, currTimestep);
        tUpdate += timerUpdate.elapsed();
        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].waitAll();
        }
        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].copyValuesFromReceiveBufferToMemGhost(currTimestep);
        }
   
        int intervalOfWrites = params().nTimesteps();
        if (0 < params().nSnapshots()) {
            if (params().nSnapshots() <= params().nTimesteps()) {
                intervalOfWrites = params().nTimesteps()
                                   / params().nSnapshots();
            }
        }

        timerWrite.restart();
        if (0 == currTimestep % intervalOfWrites) {
            for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
                tmpElemData.push_back(mVectNew[ii].elemData());
            }
            mGlobalOutput.write(tmpElemData,
                                mVectNew[0].memNormal(),
                                currTimestep);
        }
        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].write(currTimestep);
        }
        mKind.write(currTimestep);
        tWrite += timerWrite.elapsed();

        timerSolution.restart();
        if (mIsSolutionKnown) {
            for (std::size_t ii = 0; ii < mVectSol.size(); ++ii) {
                if (mFuncSol) {
                    mVectSol[ii].template evalAtGrid<OWNPRBLM, CT>(
                                        static_cast<OWNPRBLM&>(*this), mFuncSol,
                                        currTimestep);
                }
                else if (mFuncFreeSol) {
                    mVectSol[ii].template evalAtGrid<CT>(
                                        mFuncFreeSol, currTimestep);
                }
            }
        }

        if (mIsSolutionKnown) {
            for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
                mVectNew[ii].compErrLinf(mVectSol[ii]);
            }
        }
        tSolution += timerSolution.elapsed();

        for (std::size_t ii = 0; ii < mVectNew.size(); ++ii) {
            mVectNew[ii].swapWith(mVectOld[ii]);
        }

        mParams.increaseLevel();
        if ((0 == myRank()) && (0 < params().indentDepth())) {
            std::cout << getPrefix(params().indentDepth(), myRank())
                      << "   Processed." << std::endl;
        }
    }
    tIterate = timerIterate.elapsed();

    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << std::scientific << std::fixed
                  << " * Summary of CPU times(iterate):"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << "       t(init)     = "
                  << std::setw(8) << std::setprecision(5)
                  << tInit
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << tInit / tIterate * 100.0 << " %%)"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << "       t(update)   = "
                  << std::setw(8) << std::setprecision(5)
                  << tUpdate
                  << " s (" 
                  << std::setw(5) << std::setprecision(2)
                  << tUpdate / tIterate * 100.0 << " %%)"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << "       t(sync)     = "
                  << std::setw(8) << std::setprecision(5)
                  << tSync
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << tSync / tIterate * 100.0 << " %%)"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << "       t(write)    = "
                  << std::setw(8) << std::setprecision(5)
                  << tWrite
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << tWrite / tIterate * 100.0 << " %%)"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << "       t(solution) = "
                  << std::setw(8) << std::setprecision(5)
                  << tSolution
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << tSolution / tIterate * 100.0 << " %%)"
                  << std::endl
                  << getPrefix(params().indentDepth(), myRank())
                  << " ----> t(iterate)  = "
                  << std::setw(8) << std::setprecision(5) << tIterate
                  << " s. <-------"
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl;
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Iterated."
                  << std::endl;
    }
#ifdef VTRACE
    mid = VT_MARKER_DEF("EndLoop", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Problem: End Loop time steps");
#endif
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::init(std::vector< DataField<CT,DIM> >& dfDep,
                                             std::vector<CT> const& dfIndep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Initialise all data fields using geom. params..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    Ntuple<int, DIM> idxNode;
    int nNodesTotalPartition = mKind.memNormal().nNodesSub().size();
    int ii;
    int timestep = 0;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        idxNode = mKind.memoryPos2IdxNode(ii);
        static_cast<OWNPRBLM*>(this)->template initialise<CT>(dfDep, dfIndep,
                                                              idxNode, timestep);
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::init(std::vector< DataField<CT,DIM> >& dfDep,
                                             std::vector< DataField<CT,DIM> > const& dfIndep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Initialise all data fields using given data fields..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    Ntuple<int, DIM> idxNode;
    int nNodesTotalPartition = mKind.memNormal().nNodesSub().size();
    int ii;
    int timestep = 0;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        idxNode = mKind.memoryPos2IdxNode(ii);
        static_cast<OWNPRBLM*>(this)->template initialise<CT>(dfDep, dfIndep,
                                                              idxNode, timestep);
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::update(std::vector< DataField<CT,DIM> >& dfDep,
                                              std::vector< DataField<CT,DIM> > const& dfIndep,
                                              int const& timestep)

{
    updateGlobalInner(dfDep, dfIndep, timestep);
    updateGlobalBorder(dfDep, dfIndep, timestep);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateGlobalInner(std::vector< DataField<CT,DIM> >& dfDep,
                                                   std::vector< DataField<CT,DIM> > const& dfIndep,
                                                   int const& timestep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Update at global inner nodes: " << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    mParams.decreaseLevel();
    Ntuple<int, DIM> idxNode;
    int nNodesTotalPartition = mKind.memNormal().nNodesSub().size();
    int ii;

#ifdef SCAFES_HAVE_CUDA
    std::vector<CT*> vectDfDepElemData;
    std::vector<CT*> vectDfIndepElemData;
    for (int ii = 0; ii < dfDep.size(); ++ii) {
        vectDfDepElemData.push_back( dfDep[ii].elemData() );
    }
    for (int ii = 0; ii < dfIndep.size(); ++ii) {
        vectDfIndepElemData.push_back( dfIndep[ii].elemData() );
    }
    updateGlobalInnerUsingCudaVect(mKind.elemData(), vectDfDepElemData,
                                   vectDfIndepElemData, timestep,
                                   globalGrid().SCAFES_GLOBAL_INNER_NODE,
                                   nNodesTotalPartition);
#else
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        idxNode = mKind.memoryPos2IdxNode(ii);
        if (mKind(ii) & globalGrid().SCAFES_GLOBAL_INNER_NODE) {
            static_cast<OWNPRBLM*>(this)->template updateInner<CT>(dfDep, dfIndep,
                                                                 idxNode, timestep);
        }
    }
#endif
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateLocalInner(std::vector< DataField<CT,DIM> >& dfDep,
                                                        std::vector< DataField<CT,DIM> > const& dfIndep,
                                                        int const& timestep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Update at local inner nodes: " << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    mParams.decreaseLevel();
    unsigned long int nNodesTotalPartition=mKind.memNormal().nNodesSub().size();
    unsigned long int ii;

#ifdef SCAFES_HAVE_CUDA
    std::vector<CT*> vectDfDepElemData;
    std::vector<CT*> vectDfIndepElemData;
    for (int ii = 0; ii < dfDep.size(); ++ii) {
        vectDfDepElemData.push_back( dfDep[ii].elemData() );
    }
    for (int ii = 0; ii < dfIndep.size(); ++ii) {
        vectDfIndepElemData.push_back( dfIndep[ii].elemData() );
    }
    updateGlobalInnerUsingCudaVect(mKind.elemData(), vectDfDepElemData,
                                   vectDfIndepElemData, timestep,
                                   globalGrid().SCAFES_GLOBAL_INNER_NODE,
                                   nNodesTotalPartition);
#else
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
    // * timestep READ access --> shared
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        Ntuple<int, DIM> idxNode(mKind.memoryPos2IdxNode(ii));
        if (mKind(ii) & globalGrid().SCAFES_PARTITION_INNER_NODE) {
            if (mKind(ii) & globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                static_cast<OWNPRBLM*>(this)->template updateInner<CT>(dfDep, dfIndep,
                                                                       idxNode, timestep);
            } else if (mKind(ii) & globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                static_cast<OWNPRBLM*>(this)->template updateBorder<CT>(dfDep, dfIndep,
                                                                       idxNode, timestep);
            }
        }
    }
#endif

    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateLocalBorder(std::vector< DataField<CT,DIM> >& dfDep,
                                                         std::vector< DataField<CT,DIM> > const& dfIndep,
                                                         int const& timestep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Update at local border nodes: " << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    mParams.decreaseLevel();
    unsigned long int nNodesTotalPartition=mKind.memNormal().nNodesSub().size();
    unsigned long int ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
    // * timestep READ access --> shared
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        Ntuple<int, DIM> idxNode(mKind.memoryPos2IdxNode(ii));
        if (mKind(ii) & globalGrid().SCAFES_PARTITION_BORDER_NODE) {
            if (mKind(ii) & globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                Ntuple<int, DIM> idxNode(mKind.memoryPos2IdxNode(ii));
                static_cast<OWNPRBLM*>(this)->template updateInner<CT>(dfDep, dfIndep,
                                                                       idxNode, timestep);
            } else if (mKind(ii) & globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                static_cast<OWNPRBLM*>(this)->template updateBorder<CT>(dfDep, dfIndep,
                                                                       idxNode, timestep);
            }
        }
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateGlobalBorder(std::vector< DataField<CT,DIM> >& dfDep,
                                                    std::vector< DataField<CT,DIM> > const& dfIndep,
                                                    int const& timestep)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Update at global border nodes: " << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    mParams.decreaseLevel();
    Ntuple<int, DIM> idxNode;
    int nNodesTotalPartition = mKind.memNormal().nNodesSub().size();
    int ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
    // * timestep READ access --> shared
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
#endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        idxNode = mKind.memoryPos2IdxNode(ii);
        if (mKind(ii) & globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
           static_cast<OWNPRBLM*>(this)->template updateBorder<CT>(dfDep, dfIndep,
                                                                  idxNode, timestep);
        }
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::addVariable(std::string const& nameVar,
                                                   int const& stencilWidth)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Add variable " << nameVar << " to problem..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    std::string nameDatafieldOld = nameVar + "Old";
    std::string nameDatafieldNew = nameVar + "New";
    mVectOld.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldOld, params(),
                                                   globalGrid(), stencilWidth) );
    mVectNew.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldNew, params(),
                                                   globalGrid(), stencilWidth) );
    mIsSolutionKnown = false;
    mFuncSol = NULL;
    mFuncFreeSol = NULL;
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Added." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::addVariable(std::string const& nameVar,
                                                   int const& stencilWidth,
                                                   int const& nLayers)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Add variable " << nameVar << " to problem..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    std::string nameDatafieldOld = nameVar + "Old";
    std::string nameDatafieldNew = nameVar + "New";
    mVectOld.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldOld, params(),
                                                   globalGrid(), stencilWidth,
                                                   nLayers) );
    mVectNew.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldNew, params(),
                                                   globalGrid(), stencilWidth,
                                                   nLayers) );
    mIsSolutionKnown = false;
    mFuncSol = NULL;
    mFuncFreeSol = NULL;
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Added." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename PT>
inline void Problem<OWNPRBLM, CT, DIM>::addVariable(std::string const& nameVar,
                                                   PT& prblm,
                                                   int const& stencilWidth,
                                                   void (PT::*fptrSol)(CT&,
                                                           Ntuple<CT, DIM> const&,
                                                           CT const&) )
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Add variable " << nameVar << " to problem..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    std::string nameDatafieldOld = nameVar + "Old";
    std::string nameDatafieldNew = nameVar + "New";
    mVectOld.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldOld, params(),
                                                   globalGrid(), stencilWidth) );
    mVectNew.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldNew, params(),
                                                   globalGrid(), stencilWidth) );
    mIsSolutionKnown = false;
    if (fptrSol) {
        mFuncSol = fptrSol;
        mFuncFreeSol = NULL;
        mIsSolutionKnown = true;
        int bwSol = 0;
        std::string nameDatafieldSol = nameVar + "Sol";
        mVectSol.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldSol, prblm,
                                                       fptrSol, params(),
                                                       globalGrid(), bwSol) );
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Added." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::addDataField(std::string const& nameDf,
                                                 int const& stencilWidth,
                                                 void (*fptrInit)(CT&,
                                                         Ntuple<CT, DIM> const&,
                                                        CT const&),
                                                 bool const& isKnown)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::string strKnown = "unknown";
        if (isKnown) {
            strKnown = "known";
        }
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Add " << strKnown << " data field "
                  << nameDf << " to problem..."
                  << std::endl;
    } 
    mParams.decreaseLevel();
    if (isKnown) {
        mVectKnownDf.push_back( ScaFES::DataField<CT, DIM>(nameDf, fptrInit,
                                                          params(),
                                                   globalGrid(), stencilWidth) );
    }
    else {
        std::string nameDatafieldOld = nameDf + "Old";
        std::string nameDatafieldNew = nameDf + "New";
        mVectOld.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldOld, fptrInit,
                                                       params(),
                                                   globalGrid(), stencilWidth) );
        mVectNew.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldNew, fptrInit,
                                                       params(),
                                                   globalGrid(), stencilWidth) );
        mIsSolutionKnown = false;
        if (fptrInit) {
            mFuncSol = NULL;
            mFuncFreeSol = fptrInit;
            mIsSolutionKnown = true;
            int bwSol = 0;
            std::string nameDatafieldSol = nameDf + "Sol";
            mVectSol.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldSol,
                                                       fptrInit, params(),
                                                       globalGrid(), bwSol) );
        }
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Added." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename PT>
inline void Problem<OWNPRBLM, CT, DIM>::addDataField(std::string const& nameDf,
                                                 int const& stencilWidth,
                                                  PT& prblm,
                                                     void (PT::*fptrInit)(CT&,
                                                         Ntuple<CT, DIM> const&,
                                                        CT const&),
                                                 bool const& isKnown)
{
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::string strKnown = "unknown";
        if (isKnown) {
            strKnown = "known";
        }
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << " * Add " << strKnown << " data field "
                  << nameDf << " to problem..."
                  << std::endl;
    }
    mParams.decreaseLevel();
    if (isKnown) {
        mVectKnownDf.push_back( ScaFES::DataField<CT, DIM>(nameDf, prblm,
                                                           fptrInit, params(),
                                                   globalGrid(), stencilWidth) );
    }
    else {
        std::string nameDatafieldOld = nameDf + "Old";
        std::string nameDatafieldNew = nameDf + "New";
        mVectOld.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldOld,
                                                       prblm, fptrInit, params(),
                                                       globalGrid(), stencilWidth) );
        mVectNew.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldNew,
                                                       prblm, fptrInit, params(),
                                                   globalGrid(), stencilWidth) );
        mIsSolutionKnown = false;
        if (fptrInit) {
            mFuncSol = fptrInit;
            mFuncFreeSol = NULL;
            mIsSolutionKnown = true;
            int bwSol = 0;
            std::string nameDatafieldSol = nameDf + "Sol";
            mVectSol.push_back( ScaFES::DataField<CT, DIM>(nameDatafieldSol,
                                                           prblm,
                                                       fptrInit, params(),
                                                       globalGrid(), bwSol) );
        }
    }
    mParams.increaseLevel();
    if ((0 == myRank()) && (0 < params().indentDepth())) {
        std::cout << getPrefix(params().indentDepth(), myRank())
                  << "   Added." << std::endl;
    }
}

/*******************************************************************************
 * FREE METHODS.
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline void setVectorDf2ArrayMemAll(CT*& array,
                              std::vector<DataField<CT, DIM> > const& vect,
                              std::vector<bool> const& transfComp)
{
    int kk;
    int nNodesTotalPartitionCurr;
    const std::size_t NCOMP = vect.size();
    int gg = 0;
    for (std::size_t jj = 0; jj < NCOMP; ++jj) {
        if (transfComp.at(jj)) {
            kk = 0;
            nNodesTotalPartitionCurr = vect.at(jj).memAll().nNodes().size();
            for (typename Grid<DIM>::iterator it = vect.at(jj).memAll().begin(),
                et = vect.at(jj).memAll().end(); it < et; ++it) {
                array[nNodesTotalPartitionCurr * gg + kk] = vect[jj](it.idxNode());
                ++kk;
            }
            ++gg;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void setArray2VectorDfMemAll(std::vector<DataField<CT, DIM> >& vect,
                              CT* const& array,
                              std::vector<bool> const& transfComp)
{
    int kk;
    int nNodesTotalPartitionCurr;
    const std::size_t NCOMP = vect.size();
    int gg = 0;
    for (std::size_t jj = 0; jj < NCOMP; ++jj) {
        if (transfComp.at(jj)) {
            kk = 0;
            nNodesTotalPartitionCurr = vect.at(jj).memAll().nNodes().size();
            for (typename Grid<DIM>::iterator it = vect.at(jj).memAll().begin(),
                et = vect.at(jj).memAll().end(); it < et; ++it) {
                vect[jj](it.idxNode()) = array[nNodesTotalPartitionCurr * gg + kk];
                ++kk;
            }
            ++gg;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void setVectorDf2ArrayMemNormal(CT*& array,
                              std::vector<DataField<CT, DIM> > const& vect,
                              std::vector<bool> const& transfComp)
{
    int kk;
    int nNodesTotalPartitionCurr;
    const std::size_t NCOMP = vect.size();
    int gg = 0;
    for (std::size_t jj = 0; jj < NCOMP; ++jj) {
        if (transfComp.at(jj)) {
            kk = 0;
            nNodesTotalPartitionCurr = vect.at(jj).memNormal().nNodesSub().size();
            for (typename GridSub<DIM>::iterator it = vect.at(jj).memNormal().begin(),
                et = vect.at(jj).memNormal().end(); it < et; ++it) {
                array[nNodesTotalPartitionCurr * gg + kk] = vect[jj](it.idxNode());
                ++kk;
            }
            ++gg;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
inline void setArray2VectorDfMemNormal(std::vector<DataField<CT, DIM> >& vect,
                              CT* const& array,
                              std::vector<bool> const& transfComp)
{
    int kk;
    int nNodesTotalPartitionCurr;
    const std::size_t NCOMP = vect.size();
    int gg = 0;
    for (std::size_t jj = 0; jj < NCOMP; ++jj) {
        if (transfComp.at(jj)) {
            kk = 0;
            nNodesTotalPartitionCurr = vect.at(jj).memNormal().nNodesSub().size();
            for (typename GridSub<DIM>::iterator it = vect.at(jj).memNormal().begin(),
                et = vect.at(jj).memNormal().end(); it < et; ++it) {
                vect[jj](it.idxNode()) = array[nNodesTotalPartitionCurr * gg + kk];
                ++kk;
            }
            ++gg;
        }
    }
}

} // End of namespace.
#endif

