/** System test for ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef HEATEQNFDMSTENCILWITHONE_HPP_
#define HEATEQNFDMSTENCILWITHONE_HPP_

#include "ScaFES.hpp"

/*******************************************************************************
 ******************************************************************************/
/** This class serves as an example for checking the correctness of the MPI
 *  communication in ScaFES.
 *
 *  \n Aim: Scalar field with border width of 1.
 *  \n Initialisation:
 *  \code$f_{i,j,k}^0 = i+j+k \quad \forall \ i, j, k$\endcode
 *  \n Computation:
 *  \code
    \begin{align*}
       f_{i,j,k}^{n+1} = \frac{2}{7} \left(  f_{i,j,k}^n
        + f_{i-1,j,k}^n + f_{i+1,j,k}^n + f_{i,j+1,k}^n + f_{i,j-1,k}^n
                                + f_{i,j,k-1}^n + f_{i,j,k+1}^n \right)  \\
                                \forall \ i = 1, 2, \dots, N_0-2 \nonumber \\
                                        \ j = 1, 2, \dots, N_1-2 \nonumber \\
                                        \ k = 1, 2, \dots, N_2-2 \nonumber
    \end{align*}
    \endcode
 *  \n Verification:
 *  \code$f_{i,j,k}^n = 2^n(i+j+k)\quad \forall \ n= 0, 1, \dots, T$\endcode
 * 
 * \n Function
 * \code
 * \begin{align}
 * fana(x,y,z,t) &:= x/hx + y/hy + z/hz + 2 t/tau \\
 * ==> fana(x_i,y_j,z_k,t_l) &= x_i/h_0 + y_j/h_1 + z_k/h_2 + 2 t_l/tau \\
 *                         &= i h_0/h_0 + j h_1/h_1 + k h_2/h_2 + 2 l tau/tau\\
 *                         &= i + j + k + 2 l.
 * \end{align}
 * \endcode
 * 
 * \n \code $f(locNode)$\endcode is calculated as the arithmetic average of
 * the values of the six direct neighbour nodes and itself.
 * If the exchange works fine, $f(locNode)$ is constant for all time steps.
 *
 *\n Simplest case at the boundary: Each node needs the values at one neighbour
 * node:
 * \code t = 1: f[i,j,k] = (i+j+k) + (i+1+j+k) - 1 = 2*(i,j,k)\endcode
 * \n More complicated case at the boundary:
 * Each node needs the values at all direct neighbour nodes,
 * corner nodes will need the values at three neighbour nodes.
 */
template<typename CT, std::size_t DIM>
class HeatEqnFdmStencilWidthOne
    : public ScaFES::Problem<HeatEqnFdmStencilWidthOne<CT, DIM>, CT, DIM>
{
    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own constructor. */
        HeatEqnFdmStencilWidthOne(ScaFES::Parameters const& cl,
                                  ScaFES::GridGlobal<DIM> const& gg);

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Initialises initial function 'u_0'. */
        template<typename TT>
        void funcInit(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises exact solution 'u_0'. */
        template<typename TT>
        void funcSol(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises variables at all grid nodes. */
        template<typename TT>
        void initialise(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                        std::vector<TT> const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep);

        /** Computes new iteration at interior grid nodes. */
        template<typename TT>
        void updateInner(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                         std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                         ScaFES::Ntuple<int,DIM> const& idxNode,
                         int const& timestep);

        /** Computes new iteration at boundary grid nodes. */
        template<typename TT>
        void updateBorder(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                          std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep);
}; // End of class. //

/*******************************************************************************
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline HeatEqnFdmStencilWidthOne<CT, DIM>::HeatEqnFdmStencilWidthOne(
                                             ScaFES::Parameters const& cl,
                                             ScaFES::GridGlobal<DIM> const& gg)
    : ScaFES::Problem<HeatEqnFdmStencilWidthOne<CT, DIM>, CT, DIM>(cl, gg)
{
    this->addDataField("F", 1, *this,
                       &HeatEqnFdmStencilWidthOne<CT, DIM>::funcSol<CT>,
                       false);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthOne<CT, DIM>::funcInit(TT& fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    TT sum = 0.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        sum += (x[pp] / this->gridsize(pp));
    }
    fx = sum;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthOne<CT, DIM>::funcSol(TT& fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    TT sum = 0.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        sum += (x[pp] / this->gridsize(pp));
    }
    sum *= pow(2.0, t / this->tau());
    fx = sum;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthOne<CT, DIM>::initialise(
                          std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                          std::vector<TT> const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep)
{
    TT sum = 0.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        sum += idxNode.elem(pp);
    }
    vNew[0](idxNode) = sum;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthOne<CT, DIM>::updateInner(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
    vNew[0](idxNode) = vOld[0](idxNode);
    for (std::size_t pp = 0; pp < 2*DIM; ++pp) {
        vNew[0](idxNode) += vOld[0](this->connect(idxNode,pp));
    }
    vNew[0](idxNode) *= 2.0 / (2 * DIM + 1);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthOne<CT, DIM>::updateBorder(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
    TT correction = pow(2.0, timestep - 1);
    for (std::size_t pp = 0; pp < 2*DIM; ++pp) {
        // At boundary?
        if (ScaFES::Ntuple<int,DIM>(-(pp+1)) == this->connect(idxNode,pp)) {
            // LEFT, FRONT, BOTTOM...
            if (0 == pp % 2) {
                vNew[0](idxNode) = vOld[0](idxNode)
                             + vOld[0](this->connect(idxNode,pp+1))
                             - correction;
            }
            // RIGHT, BACK, TOP...
            else {
                vNew[0](idxNode) = vOld[0](idxNode)
                             + vOld[0](this->connect(idxNode,pp-1))
                             + correction;
            }
        }
    }
}
#endif
