/** System test for ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef HEATEQNFDMSTENCILWIDTHVAR_HPP_
#define HEATEQNFDMSTENCILWIDTHVAR_HPP_

#include "ScaFES.hpp"

/*******************************************************************************
 ******************************************************************************/
/** This class serves as an example for checking the correctness of the MPI
 *  communication in ScaFES.
 *
 *  \n Aim: variable stencil width
 *  \n Initialisation:
 *  \code $\varepsilon_{i,j,k}^0 = i+j+k \quad\forall \ i, j, k$\endcode
 *  \n Computation: Star with witdh "bw" or less if near to boundary.
 *  \n Verification:
 *  \code$\varepsilon_{i,j,k}^n = i+j+k+2^n \quad\forall \ n= 0,1,\dots$\endcode
 *
 *  \n Time step \code $t = 1, 2, \dots$\endcode:
 *  \code
    \begin{flalign}
    v &= \sum\limits_{r=1}^{m_i} \varepsilon_{i-r,j,k}^n
      + \sum\limits_{r=1}^{n_i} \varepsilon_{i+r,j,k}^n
      + \sum\limits_{r=1}^{m_i} \varepsilon_{i,j-r,k}^n
      + \sum\limits_{r=1}^{n_i} \varepsilon_{i,j+r,k}^n
      + \sum\limits_{r=1}^{m_i} \varepsilon_{i,j,k-r}^n
      + \sum\limits_{r=1}^{n_i} \varepsilon_{i,j,k+r}^n \nonumber \\
     \varepsilon_{i,j,k}^{n+1} &= \underbrace{
     \frac{\varepsilon_{i,j,k}^n + v - cc}{num} }_{= \varepsilon_{i,j,k}^n} +
       \ 2.0
   \end{flalign}
   \endcode
 * \code $m_s$\endcode and \code$n_s$\endcode are the numbers of neighboured
 * grid nodes (in one given direction). The values at these neighboured grid
 * nodes will be added.
 * \code
   \begin{flalign*}
     m_s &:= \min(s, epsBw) \mbox{ f\"ur } s =i,j,k \\
     \mbox{and }  n_s &:= min(N_{ss} - s - 1, epsBw)
    \mbox{ for } s = i (ss=0), j (ss=1), k (ss=2)
    \end{flalign*}
    \endcode
 *
 *  \n The value \code$\varepsilon_{i,j,k}^{n+1}$\endcode is the sum of the
 *  values at all neighboured nodes which are in a circle with a radius
 *  smaller than of equal to the given stencil width.
 *  If the distance of a grid node to the global boundary is less than the
 *  given stencil width, then the sum must be corrected by \code$cc$\endcode.
 */
template<typename CT, std::size_t DIM>
class HeatEqnFdmStencilWidthVar
    : public ScaFES::Problem<HeatEqnFdmStencilWidthVar<CT, DIM>, CT, DIM>
{
    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own constructor. */
        HeatEqnFdmStencilWidthVar(ScaFES::Parameters const& cl,
                                           ScaFES::GridGlobal<DIM> const& gg);

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Initialises initial function 'u_0'. */
        template<typename TT>
        void funcInit(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises exact solution 'u_0'. */
        template<typename TT>
        void funcSol(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises variables at all grid nodes. */
        template<typename TT>
        void initialise(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                        std::vector<TT> const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep);

        /** Computes new iteration at interior grid nodes. */
        template<typename TT>
        void updateInner(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                         std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                         ScaFES::Ntuple<int,DIM> const& idxNode,
                         int const& timestep);

        /** Computes new iteration at boundary grid nodes. */
        template<typename TT>
        void updateBorder(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                          std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep);
}; // End of class. //

/*******************************************************************************
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline HeatEqnFdmStencilWidthVar<CT, DIM>::HeatEqnFdmStencilWidthVar(
                                            ScaFES::Parameters const& cl,
                                            ScaFES::GridGlobal<DIM> const& gg)
    : ScaFES::Problem<HeatEqnFdmStencilWidthVar<CT, DIM>, CT, DIM>(cl, gg)
{
    this->addDataField("Eps", 3, *this,
                      &HeatEqnFdmStencilWidthVar<CT, DIM>::funcSol<CT>,
                      false);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthVar<CT, DIM>::funcInit(TT & fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    fx = 0.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        fx += (x[pp] / this->gridsize(pp));
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthVar<CT, DIM>::funcSol(TT & fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    fx = 2.0 * t / this->tau();
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        fx += (x[pp] / this->gridsize(pp));
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthVar<CT, DIM>::initialise(
                          std::vector< ScaFES::DataField<TT, DIM> > & vNew,
                          std::vector<TT> const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep)
{
    vNew[0](idxNode) = 0.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        vNew[0](idxNode) += idxNode[pp];
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthVar<CT, DIM>::updateInner(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
    // How many grid nodes are between the current global node and the boundary?
    ScaFES::Ntuple<int,DIM> diff2Start
        = idxNode - vOld[0].gridGlobal().idxNodeFirst();
    ScaFES::Ntuple<int,DIM> diff2End
        = vOld[0].gridGlobal().idxNodeLast() - idxNode;

    int epsBw = vOld[0].stencilWidth();
    TT sumValuesAdded = 0.0;
    // Correcture term: +1+2...+bw and -1-2-... are not equal at boundary nodes.
    int cc = 0;
    int nValuesAdded = 1;
    int minStartBw;
    int minEndBw;
    ScaFES::Ntuple<int,DIM> tmp;
    ScaFES::Ntuple<int,DIM> tmp2;
    for (std::size_t kk = 0; kk < DIM; ++kk) {
        // To the left, front, bottom...
        minStartBw = std::min(diff2Start[kk], epsBw);
        for (int rr = 1; rr <= minStartBw; ++rr) {
            tmp = idxNode;
            tmp[kk] -= rr;
            sumValuesAdded += vOld[0](tmp);
            cc -= rr;
        }
        // To the right, back, top...
        minEndBw = std::min(diff2End[kk], epsBw);
        for (int rr = 1; rr <= minEndBw; ++rr) {
            tmp = idxNode;
            tmp[kk] += rr;
            sumValuesAdded += vOld[0](tmp);
            cc += rr;
        }
        nValuesAdded += (minStartBw + minEndBw);
    }
    vNew[0](idxNode) = 2.0 + (sumValuesAdded + vOld[0](idxNode) - cc)
                        / nValuesAdded;
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void HeatEqnFdmStencilWidthVar<CT, DIM>::updateBorder(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
   this->updateInner<TT>(vNew,vOld, idxNode, timestep);
}
#endif
