/** System test for ScaFES
 *  Copyright (c) 2011-2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef PROBLEM_PROPAGATIONCOMPLEXU_HPP_
#define PROBLEM_PROPAGATIONCOMPLEXU_HPP_

#include "ScaFES.hpp"

const double EPS = 2.2e-2;
const double DEFAULT_VALUE = -25.0;

/*******************************************************************************
 ******************************************************************************/
/** This class serves as an example for checking the correctness of the MPI
 *  communication in ScaFES.
 *
 * \n Propagation from grid node \code$(0,0,0)$\endcode through the
 *    global computational domain.
 * \n Aim: Scalar field with border width of 1.
 *  \n Initialisation:
    \code\begin{flalign*}
      d &= 15,
      u_{i,j,k}^0  &= d & \forall \ i = 1, 2, \dots, N_0-1, &\\
                    &                            & j = 1, 2, \dots, N_1-1, \\
                    &                            & k = 1, 2, \dots, N_2-1  \\
      \text{und } &u_{0,0,0}^0  = 1.0
    \end{flalign*}\endcode
 *  \n Computation:
 *  \code
    \begin{flalign*}
      u_{i,j,k}^{n+1} = \begin{cases}
                    n+1, & \mbox{if } u_{i-1,j,k} \ne d, \\
                    n+1, & \mbox{if } u_{i,j-1,k} \ne d, \\
                    n+1, & \mbox{if } u_{i,j,k-1} \ne d, \\
                    u_{i,j,k}^{n}  & \mbox{else.}
                    \end{cases}
    \end{flalign*}
    \endcode
 *  \n Verification:
 *  \code$u_{i,j,k}^n = i+j+k\quad \mbox{for } \ i+j+k \le n$\endcode
 */
template<typename CT = double, std::size_t DIM = 1>
class PropagationFromLeftLowerCorner
    : public ScaFES::Problem<PropagationFromLeftLowerCorner<CT, DIM>, CT, DIM>

{
    public:
        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own constructor. */
        PropagationFromLeftLowerCorner(ScaFES::Parameters const& cl,
                                           ScaFES::GridGlobal<DIM> const& gg);

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Initialises initial function. */
        template<typename TT>
        void funcInit(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises exact solution. */
        template<typename TT>
        void funcSol(TT& fx, ScaFES::Ntuple<TT,DIM> const& x, TT const& t);

        /** Initialises variables at all grid nodes. */
        template<typename TT>
        void initialise(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                        std::vector<TT> const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep);

        /** Computes new iteration at interior grid nodes. */
        template<typename TT>
        void updateInner(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                         std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                         ScaFES::Ntuple<int,DIM> const& idxNode,
                         int const& timestep);

        /** Computes new iteration at boundary grid nodes. */
        template<typename TT>
        void updateBorder(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                          std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep);
}; // End of class. //

/*******************************************************************************
 ******************************************************************************/
template<typename CT, std::size_t DIM>
inline PropagationFromLeftLowerCorner<CT, DIM>::PropagationFromLeftLowerCorner(ScaFES::Parameters const& cl,
                                               ScaFES::GridGlobal<DIM> const& gg)
    : ScaFES::Problem<PropagationFromLeftLowerCorner<CT, DIM>, CT, DIM>(cl, gg)
{
    this->addDataField("U", 1, *this,
                       &PropagationFromLeftLowerCorner<CT, DIM>::funcSol<CT>,
                       false);
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void PropagationFromLeftLowerCorner<CT, DIM>::funcInit(TT & fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    if (x < EPS) {
        fx = 0.0;
    } else {
        fx = DEFAULT_VALUE;
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void PropagationFromLeftLowerCorner<CT, DIM>::funcSol(TT & fx,
                                   ScaFES::Ntuple<TT,DIM> const& x, TT const& t)
{
    TT sum = 0.0;
    for (std::size_t kk = 0; kk < DIM; ++kk) {
        sum += (x[kk] / this->gridsize(kk));
    }
    if (sum <= (t / this->tau() + EPS)) {
        fx = sum;
    } else {
        fx = DEFAULT_VALUE;
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void PropagationFromLeftLowerCorner<CT, DIM>::initialise(
                          std::vector< ScaFES::DataField<TT, DIM> > & vNew,
                          std::vector<TT> const& vOld,
                          ScaFES::Ntuple<int,DIM> const& idxNode,
                          int const& timestep)
{
    ScaFES::Ntuple<double, DIM> x = this->coordinates(idxNode);
    if (x < EPS) {
        vNew[0](idxNode) = 0.0;
    } else {
        vNew[0](idxNode) = DEFAULT_VALUE;
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void PropagationFromLeftLowerCorner<CT, DIM>::updateInner(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
    TT defvalCmpx = DEFAULT_VALUE;
    bool updateUnew = false;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        if ( (fabs(defvalCmpx - vOld[0](this->connect(idxNode,2*pp))) >= EPS) ) {
            updateUnew = true;
            break;
        }
    }
    if (updateUnew && (fabs(defvalCmpx - vOld[0](idxNode)) < EPS) ) {
        vNew[0](idxNode) = static_cast<double>(timestep);
    }
    else {
        vNew[0](idxNode) = vOld[0](idxNode);
    }
}
/*----------------------------------------------------------------------------*/
template<typename CT, std::size_t DIM>
template<typename TT>
inline void PropagationFromLeftLowerCorner<CT, DIM>::updateBorder(
                      std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT, DIM> > const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep)
{
    TT defvalCmpx = DEFAULT_VALUE;
    TT cmplx = defvalCmpx;
    bool updateUnew = false;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        if (vOld[0].gridGlobal().idxNodeFirst(pp) < idxNode.elem(pp)) {
            cmplx = vOld[0](this->connect(idxNode, 2*pp));
        }
        else {
            cmplx = defvalCmpx;
        }
        if ( (fabs(defvalCmpx - cmplx) >= EPS) ) {
            updateUnew = true;
            break;
        }
    }
    if (updateUnew && (fabs(defvalCmpx - vOld[0](idxNode)) < EPS) ) {
        vNew[0](idxNode) = static_cast<double>(timestep);
    }
    else {
        vNew[0](idxNode) = vOld[0](idxNode);
    }
}
#endif